library(dplyr)
library(readr)
library(rstudioapi)
library(tidyr)
library(stringr)
library(forcats)
library(extrafont)
library(ggplot2)
library(flextable)
library(officer)
library(extrafont)
base_size <- 40
theme_set(theme_light(base_size = base_size))
theme_update(text = element_text(size=base_size))
theme_update(axis.text.x = element_text(family = "TeX Gyre Pagella", angle = 45, vjust = 1.0, hjust=1, color = "black"),
             panel.grid.major.x = element_blank(),
             panel.border = element_blank(),
             panel.spacing = unit(2, "lines"),
             strip.background=element_rect(fill="white"),
             strip.text = element_text(colour = 'black'),
             plot.title = element_text(hjust = 0.5),
             axis.ticks.y  = element_blank(),
             axis.ticks.x = element_line(color = "black"))
canonical_order <- c("face_mask", "deface", "quickshear", "mri_watershed", "cp-gan")
setwd(dirname(getActiveDocumentContext()$path))
dataset.name <- "OASIS-3"
oasis3.path <- "~/remote/anonymized_server/media/datastore1/anonymized_user/Data/OASIS-3"
adni.path <- "~/remote/anonymized_server/home/anonymized_user/Data/ADNI"
resolution <- "128"
oasis3.volume_path <- str_c(oasis3.path, "/", "OASIS-3_NIFTI_PT_", resolution, "_FslSienax_VOLUMES.csv")
oasis3.dice_path <- str_c(oasis3.path, "/","OASIS-3_NIFTI_PT_", resolution, "_FslSienax_DICE.csv")
adni.volume_path <- str_c(adni.path, "/", "ADNI_NIFTI_PT_", resolution, "_FslSienax_VOLUMES.csv")
adni.dice_path <- str_c(adni.path, "/","ADNI_NIFTI_PT_", resolution, "_FslSienax_DICE.csv")
data.dice.oasis3 <- read_csv(oasis3.dice_path)
data.dice.adni <- read_csv(adni.dice_path)
data.dice <- bind_rows(list(`oasis-3`=data.dice.oasis3, adni=data.dice.adni), .id = "dataset") %>%
  select(-alt) %>%
  mutate(segmentation_type=factor(segmentation_type)) %>%
  mutate(segmentation_type=fct_recode(segmentation_type,
                                      BRAIN="I_stdmask",
                                      WHITE="I_stdmaskbrain_seg_1",
                                      GREY="I_stdmaskbrain_seg_2",
                                      VCSF="I_stdmask_segvent"))
dice.mean <- data.dice %>%
  select(-subject_id) %>%
  group_by(dataset, segmentation_type) %>%
  summarize_all(mean) 
data.dice <- data.dice %>%
  mutate(dataset=str_to_upper(dataset)) %>%
  select(-subject_id) %>%
  pivot_longer(cols = c(-segmentation_type, -dataset), names_to = "data_type", values_to = "dice_score")
data.lims <- data.dice %>%
  group_by(segmentation_type, data_type) %>%
  summarize(lowQ = quantile(dice_score, 0.01)) %>%
  ungroup()
data.filtered <- merge(data.dice, data.lims, by = c("segmentation_type", "data_type")) %>%
  filter(dice_score > lowQ) %>%
  select(-lowQ)
gray.color <- "lightslategray"
plot <- data.filtered %>%
  filter(data_type != "original") %>%
  filter(dataset == dataset.name) %>%
  mutate(data_type = str_to_upper(data_type)) %>%
  mutate(data_type = factor(data_type, levels = str_to_upper(canonical_order))) %>%
  mutate(data_type = fct_relabel(data_type, ~ str_replace(., "_", " "))) %>%
  ggplot(aes(x=data_type, y=dice_score, fill=data_type)) +
  geom_boxplot(show.legend = FALSE, outlier.shape = NA, fatten = NULL, color=gray.color) +
  geom_boxplot(aes(color = data_type),
               fatten = NULL, fill = NA, coef = 0, outlier.shape = NA, 
               show.legend = F) +
  stat_summary(geom = "crossbar", width=0.65, fatten=0, color=gray.color, show.legend = F, 
               fun.data = function(x){c(y=median(x), ymin=median(x), ymax=median(x))}) +
  facet_wrap(~segmentation_type) + 
  xlab(NULL) + 
  ylab(switch(dataset.name == "OASIS-3", "Sørensen–Dice coefficient", "")) + 
  ylim(c(0.25, 1)) +
  ggtitle(dataset.name)
if (dataset.name == "ADNI") {
  plot <- plot + theme(
        axis.text.y=element_text(color = "white"), 
        axis.ticks.y=element_blank())
}
plot
path <- str_c("~/Authoring/ICLR21/figs/experiments/sienax/", dataset.name, ".png")
ggsave(filename = path, plot = plot, width = 14, height = 14)
data.volume.oasis3 <- read_csv(oasis3.volume_path)
data.volume.adni <- read_csv(adni.volume_path)
data.volume <- bind_rows(list(`oasis-3` = data.volume.oasis3, `adni` = data.volume.adni), .id = "dataset") %>%
  select(-`unnormalised-volume`) %>%
  filter(data_type != "alt") %>%
  mutate(data_type = str_to_upper(data_type),
         tissue = str_to_upper(tissue),
         dataset = str_to_upper(dataset)) %>%
  mutate(data_type = factor(data_type),
         tissue = factor(tissue),
         dataset = factor(dataset)) 
data.volume.original <- data.volume %>% 
  filter(data_type == "ORIGINAL") %>%
  select(-data_type)
data.join <- left_join(data.volume, data.volume.original, by = c("dataset", "subject_id", "tissue"), suffix = c(".method", ".original")) %>%
  filter(data_type != "ORIGINAL")
relative.changes <- data.join %>%
  mutate(MethodOriginalDiff = volume.method - volume.original) %>%
  mutate(RelativeChange = 100 * abs(MethodOriginalDiff) / volume.original) %>%
  group_by(dataset, data_type, tissue) %>%
  summarize(MethodOriginalVolumeDiffMean = mean(RelativeChange),
            MethodOriginalVolumeDiffSD = sd(RelativeChange)) %>%
  ungroup() %>%
  group_by(dataset, tissue) %>%
  mutate(isBest = row_number(MethodOriginalVolumeDiffMean) == 1) %>%
  mutate(Error = str_c(format(MethodOriginalVolumeDiffMean, digits = 2), " ± ", format(MethodOriginalVolumeDiffSD, digits = 2))) %>%
  mutate(Error = str_replace_all(Error, " ", "")) %>%
  select(-MethodOriginalVolumeDiffMean, -MethodOriginalVolumeDiffSD) %>%
  arrange(dataset, tissue, desc(data_type)) %>%
  select(dataset, tissue, data_type, Error, isBest) %>%
  rename(Tissue = tissue, Method = data_type) %>%
  ungroup()
relative.changes.oasis3 <- relative.changes %>%
  filter(dataset == "OASIS-3") %>%
  select(-`dataset`)
relative.changes.adni <- relative.changes %>%
  filter(dataset == "ADNI") %>%
  select(-`dataset`)
data.stats <- inner_join(relative.changes.oasis3, relative.changes.adni, by = c("Tissue", "Method"), suffix = c(".OASIS-3", ".ADNI")) %>%
  mutate(`Error.OASIS-3` = if_else(`isBest.OASIS-3`, str_c(`Error.OASIS-3`, " "), `Error.OASIS-3`)) %>%
  mutate(`Error.ADNI` = if_else(`isBest.ADNI`, str_c(`Error.ADNI`, " "), `Error.ADNI`)) %>%
  select(-starts_with("isBest")) %>%
  pivot_wider(names_from = Tissue, values_from =c(`Error.OASIS-3`, `Error.ADNI`))
avg_errors <- relative.changes %>%
  select(-isBest) %>%
  mutate(Error = as.numeric(str_match(Error, "([0-9]+\\.[0-9]+)±.+")[,2])) %>%
  group_by(dataset, Method) %>%
  summarize(AvgError = mean(Error)) %>%
  ungroup() %>%
  mutate(AvgError = format(AvgError, digits = 2)) %>%
  pivot_wider(names_from = dataset, values_from = AvgError, names_glue = "AvgError.{dataset}_Avg.")
data.stats <- inner_join(data.stats, avg_errors, by = "Method") %>%
  relocate(`AvgError.OASIS-3_Avg.`, .after = `Error.OASIS-3_WHITE`)
first_header_row <- if_else(names(data.stats) == "Method", "Method", if_else(str_detect(names(data.stats), "OASIS-3"), "OASIS-3", "ADNI"))
names(first_header_row) <- names(data.stats)
second_header_row <- if_else(names(data.stats) == "Method", "Method", str_match(names(data.stats), "_([A-Za-z\\.]+)")[,2])
path <- "~/Authoring/ICLR21/figs/experiments/sienax/oasis3_adni_volumes.png"
data.stats %>%
  mutate(`Error.OASIS-3_VCSF` = str_sub(`Error.OASIS-3_VCSF`, end=-2)) %>%
  mutate(Method = str_replace(Method, "_", " ")) %>%
  mutate(Method = str_replace(Method, "MRI WATERSHED", "MRI\U202FWATERSHED")) %>%
  flextable() %>%
  set_header_labels(values = as.list(first_header_row)) %>%
  add_header_row(values = second_header_row, top = FALSE) %>%
  merge_v(part = "header") %>%
  merge_h(part = "header") %>%
  theme_booktabs(fontsize = 14) %>%
  bold(part = "header", j = 1) %>%
  bold(part = "header", i = 2) %>%
  bold(part = "body", i=~str_ends(`Error.OASIS-3_BRAIN`, " "), j = "Error.OASIS-3_BRAIN") %>%
  bold(part = "body", i=~str_ends(`Error.OASIS-3_GREY`, " "), j = "Error.OASIS-3_GREY") %>%
  bold(part = "body", i=~str_ends(`Error.OASIS-3_PGREY`, " "), j = "Error.OASIS-3_PGREY") %>%
  bold(part = "body", i=1, j = "Error.OASIS-3_VCSF") %>%
  bold(part = "body", i=~str_ends(`Error.OASIS-3_WHITE`, " "), j = "Error.OASIS-3_WHITE") %>%
  bold(part = "body", i=~str_ends(`Error.ADNI_BRAIN`, " "), j = "Error.ADNI_BRAIN") %>%
  bold(part = "body", i=~str_ends(`Error.ADNI_GREY`, " "), j = "Error.ADNI_GREY") %>%
  bold(part = "body", i=~str_ends(`Error.ADNI_PGREY`, " "), j = "Error.ADNI_PGREY") %>%
  bold(part = "body", i=~str_ends(`Error.ADNI_VCSF`, " "), j = "Error.ADNI_VCSF") %>%
  bold(part = "body", i=~str_ends(`Error.ADNI_WHITE`, " "), j = "Error.ADNI_WHITE") %>%
  bold(part = "body", i=1, j = "AvgError.OASIS-3_Avg.") %>%
  bold(part = "body", i=3, j = "AvgError.ADNI_Avg.") %>%
  align(i = c(1, 2), part = "header", align = "center") %>%
  align(part = "body", align = "center") %>%
  align(part = "body", align = "left", j = 1) %>%
  border(j = 8, border.left = fp_border(color = "black")) %>%
  font(fontname = "TeX Gyre Pagella", j = 1) %>%
  font(fontname = "Helvetica", part = "header", i = 1) %>%
  autofit() %>%
  fix_border_issues() %>%
  save_as_image(path = path)
system(str_c("convert ", path, " -fuzz 1% -trim +repage ", path))
