import os
from argparse import ArgumentParser
import pathlib
from functools import reduce
from operator import iand
import re
import numpy as np
import h5py
from tqdm import tqdm
import nibabel as nib
from utils import get_real_stem, get_file_extension, read_lzma_ndarray
extract_oasis3_subject = re.compile(r"(OAS[0-9]+)")
extract_adni_subject = re.compile(r"ADNI_([0-9]{3}_S_[0-9]{4})")
def relu(arr):
    return np.where(arr < 0, 0, arr)
def main():
    parser = ArgumentParser()
    parser.add_argument("--data_dir",
                        help="Where the modalities resides",
                        required=True)
    parser.add_argument("--dataset",
                        help="Either OASIS-3 or ADNI",
                        choices=("OASIS-3", "ADNI"),
                        required=True)
    parser.add_argument("--modalities",
                        help="From which modalities the images are extracted",
                        default=["cp-gan", "deface", "mri_watershed", "quickshear", "face_mask"])
    args = parser.parse_args()
    args.data_dir = pathlib.Path(args.data_dir)
    args.modality_dir = {modality: args.data_dir.joinpath(args.dataset, f"{args.dataset}_NIFTI_PT_128__{modality}__") for modality in args.modalities}
    args.modality_dir['original'] = args.data_dir.joinpath(args.dataset, f"{args.dataset}_NIFTI_PT_128")
    args.modality_dir['brain_mask'] = args.data_dir.joinpath(args.dataset, f"{args.dataset}_NDARRAY_PT_128_Robex")
    extract_subject = extract_oasis3_subject if args.dataset == "OASIS-3" else extract_adni_subject
    assert args.data_dir.exists(), "Data directory does not exist!"
    assert all(modality_dir.exists() for modality_dir in args.modality_dir.values()), "Not all modalities exist!"
    modality_extensions = {modality: get_file_extension(next(modality_dir.glob("*"))) for modality, modality_dir in args.modality_dir.items()}
    modality_fnames = [set(map(get_real_stem, modality_dir.glob("*"))) for modality_dir in args.modality_dir.values()]
    example_ids = list(reduce(iand, modality_fnames))
    subject_ids = [extract_subject.search(example_id).group() for example_id in example_ids]
    with open(args.data_dir.joinpath(args.dataset, "subject_axis.dat"), "w") as f:
        for subject_id in subject_ids:
            f.write(subject_id + "\n")
    with open(args.data_dir.joinpath(args.dataset, "modality_axis.dat"), "w") as f:
        for modality in args.modality_dir.keys():
            if modality != "brain_mask":
                f.write(modality + "\n")
    output_fname = args.data_dir.joinpath(args.dataset, f"MRI_Ident.h5")
    with h5py.File(output_fname, "w") as fhandle, open(args.data_dir.joinpath(args.dataset, "example_axis.dat"), "w") as f_example:
        fhandle.create_group("original")
        for example_id in tqdm(example_ids):
            brain_mask_arr = read_lzma_ndarray(args.modality_dir["brain_mask"].joinpath(f"{example_id}.npy.xz"))
            for modality in args.modality_dir:
                if modality == "brain_mask":
                    continue
                modality_dir = args.modality_dir[modality]
                file_extension = modality_extensions[modality]
                input_fname = modality_dir.joinpath(example_id + file_extension)
                if file_extension == ".nii.gz":
                    arr = nib.load(input_fname).get_fdata(dtype=np.float32)
                elif file_extension == ".npy":
                    arr = np.load(input_fname)
                else:
                    raise NotImplementedError("Unknown file extension!")
                arr = relu(arr)
                arr = arr / np.max(arr)
                arr = (1 - brain_mask_arr) * arr
                fhandle["original"].create_dataset(name=f"{modality}~{example_id}",
                                                   shape=arr.shape,
                                                   dtype=np.float32,
                                                   data=arr)
                f_example.write(f"{modality}~{example_id}\n")
if __name__ == "__main__":
    main()
