import pathlib
import re
import random
from itertools import groupby
from operator import itemgetter
import torch
from torch.utils.data import Dataset as AbstractDataset
import h5py
extract_oasis3_subject = re.compile(r"(OAS[0-9]+)")
extract_adni_subject = re.compile(r"ADNI_([0-9]{3}_S_[0-9]{4})")
class Dataset(AbstractDataset):
    def __init__(self,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 subject_axis_fname: pathlib.Path):
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.subject_axis_fname = subject_axis_fname
        self.full_example_ix = self.index_fname.read_text().strip().split("\n")
        self.subject_axis = self.subject_axis_fname.read_text().strip().split("\n")
        self.num_examples = len(self.full_example_ix)
        self.num_subjects = len(self.subject_axis)
        self.extract_subject = extract_oasis3_subject if self.subject_axis_fname.parent.stem == "OASIS-3" else extract_adni_subject
        with h5py.File(self.data_fname, "r") as fhandle:
            self.modalities = sorted({example_id.split("~")[0] for example_id in list(fhandle["original"])})
            self.num_modalities = len(self.modalities)
    def __len__(self):
        return self.num_examples
    def __getitem__(self, i):
        with h5py.File(self.data_fname, "r") as fhandle:
            full_example_id = self.full_example_ix[i]
            modality, example_id = full_example_id.split("~")
            subject_id = self.extract_subject.search(example_id).group()
            original = fhandle["original"][full_example_id][()]
            data_dict = {
                "original": original,
                "label": self.subject_axis.index(subject_id),
                "subject_id": subject_id,
                "modality": modality
            }
            return {k: torch.from_numpy(v).transpose(0, 2) if k == "original" else v for k, v in data_dict.items()}
class PairwiseTrainingDataset(AbstractDataset):
    def __init__(self,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 subject_axis_fname: pathlib.Path,
                 example_axis_fname: pathlib.Path):
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.subject_axis_fname = subject_axis_fname
        self.example_axis_fname = example_axis_fname
        self.example_fold_ix = self.index_fname.read_text().strip().split("\n")
        self.example_axis = self.example_axis_fname.read_text().strip().split("\n")
        self.extract_subject = extract_oasis3_subject if  "OASIS-3" in str(self.subject_axis_fname) else extract_adni_subject
        self.subject_ids_to_example_ids = sorted([(self.extract_subject.search(example_id).group(), example_id) for example_id in self.example_fold_ix], key=itemgetter(0))
        self.subject_ids_to_example_ids = {k: [self.example_fold_ix.index(example_id) for subject_id, example_id in g if example_id in self.example_fold_ix] for k, g in groupby(self.subject_ids_to_example_ids, key=itemgetter(0))}
        self.subject_axis = list(self.subject_ids_to_example_ids.keys())
        self.num_fold_examples = len(self.example_fold_ix)
        self.num_subjects = len(self.subject_axis)
        with h5py.File(self.data_fname, "r") as fhandle:
            self.modalities = sorted({example_id.split("~")[0] for example_id in list(fhandle["original"])})
            self.num_modalities = len(self.modalities)
    def __len__(self):
        return len(self.subject_axis)
    def __getitem__(self, i):
        with h5py.File(self.data_fname, "r") as fhandle:
            subject_id_ref = self.subject_axis[i]
            example_id_ref = random.choice(self.subject_ids_to_example_ids[subject_id_ref])
            full_example_id_ref = self.example_fold_ix[example_id_ref]
            full_example_id_ref = "original" + full_example_id_ref[full_example_id_ref.index("~"):]
            modality_ref, example_id_ref = full_example_id_ref.split("~")
            subject_id_pos = self.subject_axis[i]
            example_id_pos = random.choice(self.subject_ids_to_example_ids[subject_id_ref])
            full_example_id_pos = self.example_fold_ix[example_id_pos]
            modality_pos, example_id_pos = full_example_id_pos.split("~")
            example_id_neg = random.choice(list(range(self.num_fold_examples)))
            full_example_id_neg = self.example_fold_ix[example_id_neg]
            modality_neg, example_id_neg = full_example_id_neg.split("~")
            subject_id_neg = self.extract_subject.search(example_id_neg).group()
            arr_ref = fhandle["original"][full_example_id_ref][()]
            arr_pos = fhandle["original"][full_example_id_pos][()]
            arr_neg = fhandle["original"][full_example_id_neg][()]
            data_dict = {
                'subject_id_ref': subject_id_ref,
                'example_id_ref': example_id_ref,
                'modality_ref': modality_ref,
                'arr_ref': arr_ref,
                'subject_id_pos': subject_id_pos,
                'example_id_pos': example_id_pos,
                'modality_pos': modality_pos,
                'arr_pos': arr_pos,
                'subject_id_neg': subject_id_neg,
                'example_id_neg': example_id_neg,
                'modality_neg': modality_neg,
                'arr_neg': arr_neg
            }
            return {k: torch.from_numpy(v).transpose(0, 2) if k.startswith("arr") else v for k, v in data_dict.items()}
class PairwisePredictionDataset(AbstractDataset):
    def __init__(self,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 example_axis_fname: pathlib.Path):
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.example_axis_fname = example_axis_fname
        self.example_scan_ix = list(set(line.split("~")[1] for line in self.index_fname.read_text().strip().split("\n")))
        self.example_axis = self.example_axis_fname.read_text().strip().split("\n")
        self.num_distinct_example_ids = len(self.example_scan_ix)
        with h5py.File(self.data_fname, "r") as fhandle:
            self.modalities = sorted({example_id.split("~")[0] for example_id in list(fhandle["original"])})
            self.num_modalities = len(self.modalities)
    def __len__(self):
        return self.num_distinct_example_ids * self.num_modalities
    def __getitem__(self, i):
        with h5py.File(self.data_fname, "r") as fhandle:
            example_id = self.example_scan_ix[i % self.num_distinct_example_ids]
            modality = random.choice(self.modalities)
            data_dict = dict()
            data_dict['original'] = f"original~{example_id}"
            data_dict['correct_option'] = f"{modality}~{example_id}"
            for i in range(4):
                alternative_example_id = random.choice([j for j in self.example_scan_ix if j != example_id])
                alternative_example_id = f"{modality}~{alternative_example_id}"
                data_dict[f"option_{i}"] = alternative_example_id
            data_dict = {k: fhandle["original"][v][()] for k, v in data_dict.items()}
            data_dict = {k: torch.from_numpy(v).transpose(0, 2) for k, v in data_dict.items()}
            data_dict['modality'] = modality
            return data_dict
class RetrievalDataset(AbstractDataset):
    def __init__(self,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 example_axis_fname: pathlib.Path):
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.example_axis_fname = example_axis_fname
        self.extract_subject = extract_oasis3_subject if  "OASIS-3" in str(self.index_fname) else extract_adni_subject
        self.scan_ids = list(set(line.split("~")[1] for line in self.index_fname.read_text().strip().split("\n")))
        self.num_distinct_scans = len(self.scan_ids)
        self.scan_id_to_subject_id = {scan_id: self.extract_subject.search(scan_id).group() for scan_id in self.scan_ids}
        with h5py.File(self.data_fname, "r") as fhandle:
            self.modalities = sorted({example_id.split("~")[0] for example_id in list(fhandle["original"])})
            self.num_modalities = len(self.modalities)
    def __len__(self):
        return self.num_distinct_scans * self.num_modalities
    def __getitem__(self, i):
        with h5py.File(self.data_fname, "r") as fhandle:
            anchor_scan_id = self.scan_ids[i % self.num_distinct_scans]
            anchor_subject_id = self.scan_id_to_subject_id[anchor_scan_id]
            modality = self.modalities[i // self.num_distinct_scans]
            candidate_scan_ids = [f"original~{anchor_scan_id}", f"{modality}~{anchor_scan_id}"]
            candidate_scan_ids += [modality + "~" + scan_id for scan_id in self.scan_ids if self.scan_id_to_subject_id[scan_id] != anchor_subject_id]
            candidate_scan_arrs = {candidate_scan_id: fhandle["original"][candidate_scan_id][()] for candidate_scan_id in candidate_scan_ids}
            candidate_scan_arrs = {k: torch.from_numpy(v).transpose(0, 2) for k, v in candidate_scan_arrs.items()}
            candidate_scan_arrs['modality'] = modality
            return candidate_scan_arrs
