import torch
import torch.nn as nn
def binary_logarithm(n):
    assert n & (n - 1) == 0, "Input needs to be a power of two!"
    bitstring = bin(n)[2:]
    return len(bitstring) - 1
def swish(x):
    return torch.sigmoid(x) * x
class Layer(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        norm_func = nn.InstanceNorm2d
        conv_func = nn.Conv2d
        dropout_func = nn.Dropout2d
        self.norm_layer = norm_func(num_features=self.out_channels)
        self.conv_layer = conv_func(in_channels=self.in_channels,
                                    out_channels=self.out_channels,
                                    kernel_size=2,
                                    stride=2)
        self.dropout_layer = dropout_func(p=.2)
        self.activation = nn.LeakyReLU()
    def forward(self, x):
        x = self.conv_layer(x)
        x = self.dropout_layer(x)
        x = self.activation(x)
        x = self.norm_layer(x)
        return x
class DeidentificationNetwork(nn.Module):
    def __init__(self,
                 input_resolution: int,
                 channel_offset: int,
                 num_classes: int,
                 num_layers: int):
        super().__init__()
        self.input_resolution = input_resolution
        self.channel_offset = channel_offset
        self.num_classes = num_classes
        self.num_layers = num_layers
        self.log_input_resolution = binary_logarithm(self.input_resolution)
        self.layers = nn.ModuleList([Layer(in_channels=3,
                                           out_channels=2 ** (self.log_input_resolution + self.channel_offset))] + [
                                        Layer(in_channels=2 ** (self.log_input_resolution + self.channel_offset - i),
                                              out_channels=2 ** ( self.log_input_resolution + self.channel_offset - i - 1)) for
                                        i in range(self.num_layers)])
        self.out_channels = 2 ** (self.log_input_resolution + self.channel_offset - self.num_layers)
        self.out_resolution = (2 ** self.log_input_resolution) // (2 ** (self.num_layers + 1))
        self.out_features = self.out_channels * self.out_resolution**2
        self.fc = nn.Linear(in_features=self.out_features, out_features=self.num_classes)
    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        x =  torch.flatten(x, start_dim=1)
        x = self.fc(x)
        return x
class PairwiseDeidentificationNetwork(nn.Module):
    def __init__(self,
                 input_resolution: int,
                 channel_offset: int,
                 num_classes: int,
                 num_layers: int):
        super().__init__()
        self.input_resolution = input_resolution
        self.channel_offset = channel_offset
        self.num_classes = num_classes
        self.num_layers = num_layers
        self.log_input_resolution = binary_logarithm(self.input_resolution)
        self.layers = nn.ModuleList([Layer(in_channels=3,
                                           out_channels=2 ** (self.log_input_resolution + self.channel_offset))] + [
                                        Layer(in_channels=2 ** (self.log_input_resolution + self.channel_offset - i),
                                              out_channels=2 ** ( self.log_input_resolution + self.channel_offset - i - 1)) for
                                        i in range(self.num_layers)])
        self.out_channels = 2 ** (self.log_input_resolution + self.channel_offset - self.num_layers)
        self.out_resolution = (2 ** self.log_input_resolution) // (2 ** (self.num_layers + 1))
        self.out_features = self.out_channels * self.out_resolution**2
        self.pdist = nn.PairwiseDistance(p=2)
    def forward(self, x, y):
        for layer in self.layers:
            x = layer(x)
            y = layer(y)
        x =  torch.flatten(x, start_dim=1)
        y =  torch.flatten(y, start_dim=1)
        return self.pdist(x, y)
