import sys
from argparse import ArgumentParser
import pathlib
from collections import defaultdict
from tqdm import tqdm
import yaml
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from adamp import AdamP
from torch.utils.data import DataLoader
from algorithmic_deidentification.data import Dataset
from algorithmic_deidentification.network import DeidentificationNetwork
torch.manual_seed(5)
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
def main():
    parser = ArgumentParser()
    parser.add_argument("--config_fname",
                        help="The location of the config",
                        required=True)
    parser.add_argument("--data_fname",
                        help="The path to the data",
                        required=True)
    parser.add_argument("--train_index_fname",
                        help="The path to the file containing the training indices",
                        required=True)
    parser.add_argument("--test_index_fname",
                        help="The path to the file containing the test indices",
                        required=True)
    parser.add_argument("--subject_axis_fname",
                        help="The file that contains the overall subjects",
                        required=True)
    args = parser.parse_args()
    args.data_fname = pathlib.Path(args.data_fname)
    args.train_index_fname = pathlib.Path(args.train_index_fname)
    args.test_index_fname = pathlib.Path(args.test_index_fname)
    args.subject_axis_fname = pathlib.Path(args.subject_axis_fname)
    assert args.data_fname.exists(), "Data does not exist!"
    assert args.train_index_fname.exists(), "Index file does not exist!"
    assert args.test_index_fname.exists(), "Index file does not exist!"
    assert args.subject_axis_fname.exists(), "Class axis file does not exist!"
    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        print("CUDA not available!")
        sys.exit(1)
    with open(args.config_fname, "r") as config_f:
        config = yaml.load(config_f, Loader=yaml.FullLoader)
    dataset = Dataset(data_fname=args.data_fname,
                      index_fname=args.train_index_fname,
                      subject_axis_fname=args.subject_axis_fname)
    data = torch.utils.data.DataLoader(dataset,
                                       batch_size=config['batch_size'],
                                       num_workers=4,
                                       shuffle=True,
                                       pin_memory=True)
    net = DeidentificationNetwork(input_resolution=128,
                                  channel_offset=config["channel_offset"],
                                  num_classes=dataset.num_subjects,
                                  num_layers=config["num_layers"]).to(device)
    loss_fct = nn.CrossEntropyLoss()
    optimizer = AdamP(net.parameters(), lr=config["lr"], betas=config["betas"])
    """ Training """
    for epoch_id in range(config["num_epochs"]):
        with tqdm(data) as bar:
            for x in bar:
                x.pop("subject_id")
                x.pop("modality")
                optimizer.zero_grad()
                x = {k: t.to(device, non_blocking=True) for k, t in x.items()}
                y = net(x["original"])
                loss = loss_fct(y, x['label'])
                loss.backward()
                optimizer.step()
                loss_v = loss.detach().cpu().numpy()
                bar.set_description("Epoch {}\tLoss {:.3f}\tProbability {:.3f}%".format(epoch_id, loss_v, 100 * math.exp(-loss_v)))
    """ Prediction """
    print("Prediction..")
    net = net.eval()
    dataset = Dataset(data_fname=args.data_fname,
                      index_fname=args.test_index_fname,
                      subject_axis_fname=args.subject_axis_fname)
    data = torch.utils.data.DataLoader(dataset,
                                       batch_size=1,
                                       num_workers=4,
                                       shuffle=False,
                                       pin_memory=True)
    with torch.no_grad():
        modality_scores = defaultdict(int)
        for n, x in tqdm(enumerate(data)):
            subject_id = x.pop("subject_id")[0]
            modality = x.pop("modality")[0]
            x = {k: t.to(device, non_blocking=True) for k, t in x.items()}
            y = F.softmax(net(x["original"]).flatten(), dim=0)
            correct_class = int(x["label"])
            pred_class = int(y.argmax())
            is_correct = 100. * int(pred_class == correct_class)
            modality_scores[modality] = 1/(n + 1)  * (n * modality_scores[modality] + is_correct)
        for modality, score in modality_scores.items():
            print(f"{modality} => {score:.3f}%", end="\t")
        print()
if __name__ == "__main__":
    main()
