import sys
from argparse import ArgumentParser
import pathlib
from random import sample
from itertools import groupby
from operator import itemgetter
import yaml
import pandas as pd
import numpy as np
import imageio.core.util
def ignore_warnings(*args, **kwargs):
    pass
imageio.core.util._precision_warn = ignore_warnings
import torch
import torch.nn as nn
import torch.nn.functional as F
from adamp import AdamP
from torch.utils.data import DataLoader
from algorithmic_deidentification.data import PairwiseTrainingDataset
from algorithmic_deidentification.data import RetrievalDataset
from algorithmic_deidentification.network import PairwiseDeidentificationNetwork as DeidentificationNetwork
torch.manual_seed(1)
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
def out_of_place_shuffle(ls):
    return sample(ls, len(ls))
def main():
    parser = ArgumentParser()
    parser.add_argument("--config_fname",
                        help="The location of the config",
                        required=True)
    parser.add_argument("--data_fname",
                        help="The path to the data",
                        required=True)
    parser.add_argument("--train_index_fname",
                        help="The path to the file containing the training indices",
                        required=True)
    parser.add_argument("--test_index_fname",
                        help="The path to the file containing the test indices",
                        required=True)
    parser.add_argument("--subject_axis_fname",
                        help="The file that contains the overall subjects",
                        required=True)
    parser.add_argument("--example_axis_fname",
                        help="The file that contains the overall examples",
                        required=True)
    parser.add_argument("--output_fname",
                        help="Where to store the scores",
                        required=True)
    parser.add_argument("--device",
                        help="Which GPU to use",
                        default=0,
                        type=int)
    args = parser.parse_args()
    args.data_fname = pathlib.Path(args.data_fname)
    args.train_index_fname = pathlib.Path(args.train_index_fname)
    args.test_index_fname = pathlib.Path(args.test_index_fname)
    args.subject_axis_fname = pathlib.Path(args.subject_axis_fname)
    args.example_axis_fname = pathlib.Path(args.example_axis_fname)
    args.output_fname = pathlib.Path(args.output_fname)
    assert args.data_fname.exists(), "Data does not exist!"
    assert args.train_index_fname.exists(), "Index file does not exist!"
    assert args.test_index_fname.exists(), "Index file does not exist!"
    assert args.subject_axis_fname.exists(), "Class axis file does not exist!"
    if torch.cuda.is_available():
        device = torch.device(f"cuda:{args.device}")
    else:
        print("CUDA not available!")
        sys.exit(1)
    with open(args.config_fname, "r") as config_f:
        config = yaml.load(config_f, Loader=yaml.FullLoader)
    dataset = PairwiseTrainingDataset(data_fname=args.data_fname,
                                      index_fname=args.train_index_fname,
                                      subject_axis_fname=args.subject_axis_fname,
                                      example_axis_fname=args.example_axis_fname)
    data = torch.utils.data.DataLoader(dataset,
                                       batch_size=config['batch_size'],
                                       num_workers=4,
                                       shuffle=True,
                                       pin_memory=True)
    net = DeidentificationNetwork(input_resolution=128,
                                  channel_offset=config["channel_offset"],
                                  num_classes=dataset.num_subjects,
                                  num_layers=config["num_layers"]).to(device)
    loss_fct = nn.BCEWithLogitsLoss()
    optimizer = AdamP(net.parameters(), lr=config["lr"], betas=config["betas"])
    """ Training """
    margin = 5.0
    stop_training = False
    for epoch_id in range(config["num_epochs"]):
        if stop_training:
            break
        for candidate_scan_arrs in data:
            candidate_scan_arrs.pop("subject_id_ref"); candidate_scan_arrs.pop("subject_id_pos"); candidate_scan_arrs.pop("subject_id_neg")
            candidate_scan_arrs.pop("example_id_ref"); candidate_scan_arrs.pop("example_id_pos"); candidate_scan_arrs.pop("example_id_neg")
            candidate_scan_arrs.pop("modality_ref"); candidate_scan_arrs.pop("modality_pos"); candidate_scan_arrs.pop("modality_neg")
            optimizer.zero_grad()
            candidate_scan_arrs = {k: t.to(device, non_blocking=True) for k, t in candidate_scan_arrs.items()}
            y_pos = net(candidate_scan_arrs["arr_ref"], candidate_scan_arrs["arr_pos"])
            y_neg = net(candidate_scan_arrs["arr_ref"], candidate_scan_arrs["arr_neg"])
            loss = torch.relu(margin + y_pos - y_neg).mean()
            loss.backward()
            optimizer.step()
    """ Prediction """
    net = net.eval()
    dataset = RetrievalDataset(data_fname=args.data_fname,
                               index_fname=args.test_index_fname,
                               example_axis_fname=args.example_axis_fname)
    data = torch.utils.data.DataLoader(dataset,
                                       batch_size=1,
                                       num_workers=4,
                                       shuffle=False,
                                       pin_memory=True)
    tmp_dir = pathlib.Path.home().joinpath("tmp")
    tmp_dir.mkdir(exist_ok=True)
    full_df = []
    with torch.no_grad():
        for candidate_scan_arrs in data:
            modality = candidate_scan_arrs.pop('modality')[0]
            candidate_scan_arrs = {k: t.to(device, non_blocking=True) for k, t in candidate_scan_arrs.items()}
            original_scan_id = next(scan_id for scan_id in candidate_scan_arrs if scan_id.startswith("original"))
            candidate_scores = {candidate_scan_id: net(candidate_scan_arrs[original_scan_id], tensor).item() for candidate_scan_id, tensor in candidate_scan_arrs.items()}
            if modality != "original":
                del candidate_scores[original_scan_id]
            candidate_scores = sorted(candidate_scores.items(), key=itemgetter(1))
            candidate_scores = [x for score, candidate_scan_ids in groupby(candidate_scores, key=itemgetter(1)) for x in out_of_place_shuffle(list(candidate_scan_ids))]
            candidate_scores_with_ranks = [(rank,) + candidate_score for rank, candidate_score in enumerate(candidate_scores, 1)]
            df = pd.DataFrame.from_records(candidate_scores_with_ranks, columns=['Rank', 'CandidateScanId', 'Score'])
            df['AnchorId'] = original_scan_id
            df['Modality'] = modality
            df['CorrectScanId'] = modality + "~" + original_scan_id.split("~")[1]
            full_df.append(df)
    full_df = pd.concat(full_df, axis=0).reset_index(drop=True)
    full_df.to_csv(args.output_fname, index=False)
if __name__ == "__main__":
    main()
