import sys
from argparse import ArgumentParser
import pathlib
from collections import defaultdict
import random
from tqdm import tqdm
import yaml
import math
import numpy as np
from skimage.io import imsave
from tempfile import TemporaryDirectory
import imageio.core.util
def ignore_warnings(*args, **kwargs):
    pass
imageio.core.util._precision_warn = ignore_warnings
import torch
import torch.nn as nn
import torch.nn.functional as F
from adamp import AdamP
from torch.utils.data import DataLoader
from algorithmic_deidentification.data import PairwiseTrainingDataset
from algorithmic_deidentification.data import PairwisePredictionDataset
from algorithmic_deidentification.network import PairwiseDeidentificationNetwork as DeidentificationNetwork
torch.manual_seed(1)
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
def main():
    parser = ArgumentParser()
    parser.add_argument("--config_fname",
                        help="The location of the config",
                        required=True)
    parser.add_argument("--data_fname",
                        help="The path to the data",
                        required=True)
    parser.add_argument("--train_index_fname",
                        help="The path to the file containing the training indices",
                        required=True)
    parser.add_argument("--test_index_fname",
                        help="The path to the file containing the test indices",
                        required=True)
    parser.add_argument("--subject_axis_fname",
                        help="The file that contains the overall subjects",
                        required=True)
    parser.add_argument("--example_axis_fname",
                        help="The file that contains the overall examples",
                        required=True)
    parser.add_argument("--output_fname",
                        help="Where to store the scores",
                        required=True)
    parser.add_argument("--device",
                        help="Which GPU to use",
                        default=0,
                        type=int)
    args = parser.parse_args()
    args.data_fname = pathlib.Path(args.data_fname)
    args.train_index_fname = pathlib.Path(args.train_index_fname)
    args.test_index_fname = pathlib.Path(args.test_index_fname)
    args.subject_axis_fname = pathlib.Path(args.subject_axis_fname)
    args.example_axis_fname = pathlib.Path(args.example_axis_fname)
    args.output_fname = pathlib.Path(args.output_fname)
    assert args.data_fname.exists(), "Data does not exist!"
    assert args.train_index_fname.exists(), "Index file does not exist!"
    assert args.test_index_fname.exists(), "Index file does not exist!"
    assert args.subject_axis_fname.exists(), "Class axis file does not exist!"
    if torch.cuda.is_available():
        device = torch.device(f"cuda:{args.device}")
    else:
        print("CUDA not available!")
        sys.exit(1)
    with open(args.config_fname, "r") as config_f:
        config = yaml.load(config_f, Loader=yaml.FullLoader)
    dataset = PairwiseTrainingDataset(data_fname=args.data_fname,
                                      index_fname=args.train_index_fname,
                                      subject_axis_fname=args.subject_axis_fname,
                                      example_axis_fname=args.example_axis_fname)
    data = torch.utils.data.DataLoader(dataset,
                                       batch_size=config['batch_size'],
                                       num_workers=4,
                                       shuffle=True,
                                       pin_memory=True)
    net = DeidentificationNetwork(input_resolution=128,
                                  channel_offset=config["channel_offset"],
                                  num_classes=dataset.num_subjects,
                                  num_layers=config["num_layers"]).to(device)
    loss_fct = nn.BCEWithLogitsLoss()
    optimizer = AdamP(net.parameters(), lr=config["lr"], betas=config["betas"])
    """ Training """
    margin = 5.0
    stop_training = False
    for epoch_id in range(config["num_epochs"]):
        if stop_training:
            break
        for x in data:
            x.pop("subject_id_ref" ); x.pop("subject_id_pos"); x.pop("subject_id_neg")
            x.pop("example_id_ref"); x.pop("example_id_pos"); x.pop("example_id_neg")
            x.pop("modality_ref"); x.pop("modality_pos"); x.pop("modality_neg")
            optimizer.zero_grad()
            x = {k: t.to(device, non_blocking=True) for k, t in x.items()}
            y_pos = net(x["arr_ref"], x["arr_pos"])
            y_neg = net(x["arr_ref"], x["arr_neg"])
            loss = torch.relu(margin + y_pos - y_neg).mean()
            loss.backward()
            optimizer.step()
    """ Prediction """
    net = net.eval()
    dataset = PairwisePredictionDataset(data_fname=args.data_fname,
                                        index_fname=args.test_index_fname,
                                        example_axis_fname=args.example_axis_fname)
    data = torch.utils.data.DataLoader(dataset,
                                       batch_size=1,
                                       num_workers=4,
                                       shuffle=False,
                                       pin_memory=True)
    tmp_dir = pathlib.Path.home().joinpath("tmp")
    tmp_dir.mkdir(exist_ok=True)
    with torch.no_grad():
        modality_scores = defaultdict(float)
        modality_counts = defaultdict(int)
        for x in data:
            modality = x.pop('modality')[0]
            n = modality_counts[modality]
            x = {k: t.to(device, non_blocking=True) for k, t in x.items()}
            x_scores = {option: torch.sigmoid(net(x["original"], tensor)).item() for option, tensor in x.items() if option != "original"}
            x_scores = sorted(x_scores.items(), key=lambda t: t[1])
            min_score = min(tup[1] for tup in x_scores)
            winners = [tup[0] for tup in x_scores if tup[1] == min_score]
            winner = random.choice(winners)
            guessed_correct = winner == 'correct_option'
            modality_scores[modality] = 1/(n + 1)  * (n * modality_scores[modality] + 100 * float(guessed_correct))
            modality_counts[modality] += 1
        modality_scores = sorted(modality_scores.items(), key=lambda t: t[0])
        results = ",".join(["=".join(map(str, tup)) for tup in modality_scores])
        print(results)
        with open(args.output_fname, "w") as f:
            f.write(results)
if __name__ == "__main__":
    main()
