from argparse import ArgumentParser
import pathlib
from typing import Tuple
from functools import partial
import numpy as np
from scipy.spatial import ConvexHull
from joblib import delayed, Parallel
from tqdm import tqdm
from utils import save_lzma_ndarray_archive, read_ndarray, glob_ndarrays
def hessian_normal_form(p1: np.ndarray, p2: np.ndarray, p3: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    g2 = p2 - p1
    g3 = p3 - p1
    n = np.cross(g2, g3)
    n = n/np.linalg.norm(n, axis=1, ord=2, keepdims=True)
    d = np.sum(np.multiply(n, p1), axis=1)
    return n, d
def convex_hull_integral_points(tri_coords: np.ndarray,
                                points: np.ndarray,
                                max_num_triangles: int = 0):
    num_points = len(points)
    num_triangles = len(tri_coords) if max_num_triangles == 0 else min(len(tri_coords), max_num_triangles)
    np.random.shuffle(tri_coords)
    p1, p2, p3 = tri_coords[:, 0], tri_coords[:, 1], tri_coords[:, 2]
    n, d = hessian_normal_form(p1, p2, p3)
    centroid = np.mean(tri_coords, axis=(0, 1))
    centroid_membership = np.sign(np.dot(n, centroid) - d)
    n = np.multiply(n, centroid_membership[:, np.newaxis])
    d = np.multiply(d, centroid_membership)
    membership = np.ones(shape=[num_points], dtype=np.bool)
    for simplex_id in range(num_triangles):
        points_d = np.dot(points, n[simplex_id])
        half_space_membership = points_d >= d[simplex_id]
        membership = np.logical_and(membership, half_space_membership)
    integral_points = np.compress(membership, points, axis=0)
    return integral_points
def convex_hull_lattice(img: np.ndarray, max_num_triangles: int):
    coords = np.transpose(np.nonzero(img))
    convex_hull = ConvexHull(coords)
    tri_ix = convex_hull.simplices
    tri_coords = coords[tri_ix].astype(np.float32)
    lattice_points = np.stack(np.meshgrid(np.arange(img.shape[0]), np.arange(img.shape[1]), np.arange(img.shape[2]),
                                          indexing="ij"), axis=3).astype(np.int32)
    lattice_points = np.reshape(lattice_points, [-1, 3])
    lattice_points = convex_hull_integral_points(tri_coords=tri_coords,
                                                 points=lattice_points,
                                                 max_num_triangles=max_num_triangles)
    img_ch = np.zeros_like(img, dtype=np.bool)
    img_ch[lattice_points[:, 0], lattice_points[:, 1], lattice_points[:, 2]] = True
    return img_ch
def combine(input_original_fname: pathlib.Path,
            input_face_fname: pathlib.Path,
            input_brain_fname: pathlib.Path,
            max_num_triangles: int,
            output_fname: pathlib.Path = None,
            return_as_arrays: bool = False):
    if not input_original_fname.exists():
        print(f"File {input_original_fname} does not exist!")
        return
    if not input_face_fname.exists():
        print(f"File {input_face_fname} does not exist!")
        return
    if not input_brain_fname.exists():
        print(f"File {input_brain_fname} does not exist!")
        return
    if output_fname is not None and output_fname.exists():
        print(f"Skipping {output_fname} as it already exists")
        return
    mri_original = read_ndarray(input_original_fname).astype(np.float32)
    mri_face = read_ndarray(input_face_fname).astype(np.float32)
    mri_brain = read_ndarray(input_brain_fname).astype(np.float32)
    mri_face = mri_face > np.random.rand(*mri_face.shape).astype(np.float32)
    convex_hull = convex_hull_lattice(mri_face, max_num_triangles=max_num_triangles)
    output = {
        "original": mri_original,
        "convex_hull": convex_hull,
        "brain_mask": mri_brain > 0
    }
    if return_as_arrays:
        return output
    else:
        save_lzma_ndarray_archive(output_fname=output_fname, **output)
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_original_dir",
                        help="Where the original representations reside",
                        required=True)
    parser.add_argument("--input_face_dir",
                        help="Where the face representations reside",
                        required=True)
    parser.add_argument("--input_brain_dir",
                        help="Where the brain representations reside",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the resulting files are being stored",
                        required=True)
    parser.add_argument("--max_num_triangles",
                        type=int,
                        default=250)
    parser.add_argument("--num_jobs",
                        help="How many jobs to spawn",
                        type=int,
                        default=1)
    args = parser.parse_args()
    args.input_original_dir = pathlib.Path(args.input_original_dir)
    args.input_face_dir = pathlib.Path(args.input_face_dir)
    args.input_brain_dir = pathlib.Path(args.input_brain_dir)
    args.output_dir = pathlib.Path(args.output_dir)
    assert args.input_original_dir.exists(), "Original input directory does not exist!"
    assert args.input_face_dir.exists(), "Face input directory does not exist!"
    assert args.input_brain_dir.exists(), "Brain input directory does not exist!"
    args.output_dir.mkdir(exist_ok=True)
    fnames = [(original_fname,
               args.input_face_dir.joinpath(original_fname.name).with_suffix(".npy.xz"),
               args.input_brain_dir.joinpath(original_fname.name).with_suffix(".npy.xz"),
               args.output_dir.joinpath(original_fname.stem).with_suffix(".npy.tar.xz")) for original_fname in glob_ndarrays(args.input_original_dir)]
    func = partial(combine,
                   max_num_triangles=args.max_num_triangles)
    Parallel(n_jobs=args.num_jobs)(delayed(func)(*args) for args in tqdm(fnames))
if __name__ == "__main__":
    main()
