from argparse import ArgumentParser
import pathlib
import os
import numpy as np
import h5py
from tqdm import tqdm
from skimage.util.shape import view_as_blocks
from utils import read_lzma_ndarray_archive, get_real_stem
OUTPUT_SIZE = (128, 128, 128)
def pad_to_output_size(arr):
    def _halve_number(n):
        return -(-n // 2), n//2
    pad_width = [_halve_number(s1 - s2) for s1, s2 in zip(OUTPUT_SIZE, arr.shape)]
    return np.pad(arr,
                  pad_width=pad_width,
                  mode="constant",
                  constant_values=False if arr.dtype == np.bool else 0.)
def resize_to_target_size(arr, tgt_size):
    block_shape = tuple([s1 // s2 for s1, s2 in zip(arr.shape, tgt_size)])
    arr_blocks = view_as_blocks(arr, block_shape=block_shape)
    if arr.dtype == np.bool:
        return np.mean(arr_blocks, axis=(3, 4, 5)) >= .5
    else:
        return np.mean(arr_blocks, axis=(3, 4, 5))
def main():
    parser = ArgumentParser()
    parser.add_argument("--paired_dataset",
                        help="The directory in which the .npy.tar.xz files reside",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the resulting HDF5 files will be stored",
                        required=True)
    parser.add_argument("--sizes",
                        help="Which sizes to use for the resizing",
                        default=[128])
    args = parser.parse_args()
    args.paired_dataset = pathlib.Path(args.paired_dataset)
    args.output_dir = pathlib.Path(args.output_dir)
    args.output_dir.mkdir(exist_ok=True)
    args.sizes = [(size,) * 3 for size in args.sizes]
    assert args.paired_dataset.exists(), "Input dataset does not exist!"
    fhandles = {size: h5py.File(os.path.join(args.output_dir, "MRI_{}.h5".format(size[0])), "w") for size in args.sizes}
    for size, fhandle in fhandles.items():
        fhandle.create_group("original")
        fhandle.create_group("convex_hull")
        fhandle.create_group("brain_mask")
    fnames = list(args.paired_dataset.glob("*.npy.tar.xz"))
    for i, fname in enumerate(tqdm(fnames)):
        archive = read_lzma_ndarray_archive(fname)
        original = archive["original"]  
        convex_hull = archive["convex_hull"]    
        brain_mask = archive["brain_mask"] if "brain_mask" in archive else ~convex_hull  
        for size in args.sizes:
            original_resized = original
            convex_hull_resized = convex_hull
            brain_mask_resized = brain_mask
            name = get_real_stem(fname)
            fhandles[size]["original"].create_dataset(name, data=original_resized,
                                                      shape=size + (3,), dtype=np.float32, compression="gzip")
            fhandles[size]["convex_hull"].create_dataset(name, data=convex_hull_resized,
                                                         shape=size, dtype=np.bool, compression="gzip")
            fhandles[size]["brain_mask"].create_dataset(name, data=brain_mask_resized,
                                                        shape=size, dtype=np.bool, compression="gzip")
    for fhandle in fhandles.values():
        fhandle.close()
if __name__ == "__main__":
    main()
