from argparse import ArgumentParser
import pathlib
from typing import Tuple, Union, List
import numpy as np
from scipy.linalg import norm
from joblib import Parallel, delayed
import itertools
from utils import glob_ndarrays, read_ndarray, save_ndarray
from utils.interpolation import fast_trilinear_interpolation
HISTOGRAM_STEP_SIZE = .01
def halve_number(n: int) -> Tuple[int, int]:
    return n // 2, -(-n // 2)
def get_bounding_box(coords: np.ndarray) -> np.ndarray:
    """
    Computes the bounding box for a set of coordinates
    :param coords: An np.ndarray of shape [n, d]
    :return: An np.ndarray of shape [d, 2] otherwise
    """
    num_dims = coords.shape[1]
    return np.array([(np.min(coords[:, i]), np.max(coords[:, i])) for i in range(num_dims)])    
def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))
def uniform_3d_rotation() -> Tuple[np.ndarray, float]:
    normal_variate = np.random.normal(size=4)
    spherical_variate = normal_variate/norm(normal_variate, ord=2)
    rot_angle = 2. * np.arccos(spherical_variate[3])
    rot_axis = spherical_variate[:3]/np.sin(rot_angle/2.)
    return rot_axis, rot_angle
def apply_3d_rotation(arr, rot_axis: np.ndarray, rot_angle: float):
    a = np.cos(rot_angle/2.)
    b, c, d = np.sin(rot_angle/2.) * rot_axis
    rot_mat = np.array([
        [a**2 + b**2 - c**2 - d**2, 2 * (b * c - a * d), 2 * (b * d + a * c)],
        [2 * (b * c + a * d), a**2 + c**2 - b**2 - d**2, 2 * (c * d - a * b)],
        [2 * (b * d - a * c), 2 * (c * d + a * b), a**2 + d**2 - b**2 - c**2]
    ], dtype=np.float32)
    if arr.ndim == 2:
        return np.dot(arr, rot_mat.T)
    width, height, depth = arr.shape
    points = [(np.arange(dim) - dim/2).astype(np.float32) for dim in (height, width, depth)]
    grid = np.stack(np.meshgrid(*points, indexing="xy"), axis=3)    
    rot_grid = np.dot(grid, rot_mat.T)    
    x, y, z = np.hsplit(np.reshape(rot_grid, newshape=(-1, 3)), 3)    
    x = np.squeeze(x + height/2.)    
    y = np.squeeze(y + width/2.)   
    z = np.squeeze(z + depth/2.)    
    rot_arr = fast_trilinear_interpolation(img=arr, x=x, y=y, z=z)
    rot_arr = np.reshape(rot_arr, [width, height, depth])     
    return rot_arr
def first_hits(img: np.ndarray, threshold: float, reduce=True) -> Union[np.ndarray, List[np.ndarray]]:
    dims = img.shape
    ndim = img.ndim
    img_binary = img >= threshold
    def _argmax(arr, axis, direction, keepdims=False):
        if direction == -1:
            arr = np.flip(arr, axis=axis)
        result = np.argmax(arr, axis=axis)
        if direction == -1:
            result = arr.shape[axis] - result - 2
        if keepdims:
            result = np.expand_dims(result, axis=axis)
        return result
    def _first_hit_from_direction(axis: int, direction: int) -> np.ndarray:
        """
        :param axis: Satisfies 0 <= axis < ndim
        :param direction: Either +1 (forward) or -1 (backward)
        """
        pad_width = [(0, 0) if dim != axis else (0, 1)[::direction] for dim in range(ndim)]
        img_binary_padded = np.pad(img_binary,
                                   pad_width=pad_width,
                                   mode="constant",
                                   constant_values=True)
        ix = _argmax(img_binary_padded, axis=axis, direction=direction, keepdims=True)
        newshape = (1,) * axis + (dims[axis],) + (1,) * (ndim - axis - 1)
        reps = dims[:axis] + (1,) + dims[axis+1:]
        axis_range = np.tile(np.reshape(np.arange(dims[axis]), newshape=newshape), reps=reps)
        return np.equal(ix, axis_range).astype(img.dtype)
    imgs = [_first_hit_from_direction(axis=axis, direction=direction) for axis, direction in
            itertools.product(range(ndim), [-1, +1])]
    return np.mean(imgs, axis=0) if reduce else imgs
def unpad(x, pad_width):
    """ c.f. https://stackoverflow.com/a/57956349 """
    slices = []
    for c in pad_width:
        e = None if c[1] == 0 else -c[1]
        slices.append(slice(c[0], e))
    return x[tuple(slices)]
def extract_surface(input_fname: pathlib.Path,
                    num_rotations: int,
                    threshold: float,
                    output_dir: pathlib.Path,
                    compress_ndarrays: bool):
    output_fname = output_dir.joinpath(input_fname.name)
    mri_scan = read_ndarray(input_fname).astype(np.float32)
    width, height, depth = mri_scan.shape
    bins = np.arange(mri_scan.min(), mri_scan.max(), HISTOGRAM_STEP_SIZE)
    hist, bin_edges = np.histogram(mri_scan, bins=bins, density=True)
    hist_cum = np.cumsum(hist * HISTOGRAM_STEP_SIZE)
    ix = np.argmax(hist_cum > threshold)
    threshold = bins[ix]
    domain = np.array([[-height / 2, +height / 2],
                       [-width / 2, +width / 2],
                       [-depth / 2, +depth / 2]], dtype=np.float32)     
    bbox_domain = np.array(list(itertools.product(*domain)))   
    result = np.zeros_like(mri_scan)
    for i in range(num_rotations):
        rot_axis, rot_angle = uniform_3d_rotation()
        rot_domain = apply_3d_rotation(bbox_domain, rot_axis=rot_axis, rot_angle=rot_angle)     
        bbox_rot_domain = get_bounding_box(rot_domain)     
        pad_width = np.ceil(np.abs(bbox_rot_domain - domain)).astype(np.int32)  
        mri_scan_padded = np.pad(mri_scan,
                                 pad_width=pad_width,
                                 mode="constant",
                                 constant_values=mri_scan.min())
        rot_img = apply_3d_rotation(mri_scan_padded, rot_axis=rot_axis, rot_angle=rot_angle)
        first_hit_img = first_hits(rot_img, threshold, reduce=True)
        back_rot_img = apply_3d_rotation(first_hit_img, rot_axis=rot_axis, rot_angle=-rot_angle)
        back_rot_img = unpad(back_rot_img, pad_width=pad_width)
        result = 1/(i + 1) * (i * result + back_rot_img)
    save_ndarray(arr=result, output_fname=output_fname, compress=compress_ndarrays)
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="Path of the directory containing the MRI scans",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the transformed numpy arrays should be stored",
                        required=False)
    parser.add_argument("--num_rotations",
                        help="How many random rotations should be applied to infer the surface",
                        default=6,
                        type=int)
    parser.add_argument("--threshold",
                        help="When to classify a voxel as belonging to MRI and not to the surrounding",
                        default=0.75,
                        type=float)
    parser.add_argument("--num_jobs",
                        default=1,
                        type=int)
    parser.add_argument("--compress_ndarrays",
                        help="Whether to LZMA compress each saved ndarray",
                        action="store_true",
                        default=False)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.output_dir = pathlib.Path(args.output_dir)
    args.output_dir.mkdir(exist_ok=True)
    assert args.input_dir.exists(), "Input directory does not exist!"
    input_fnames = glob_ndarrays(args.input_dir)
    Parallel(n_jobs=args.num_jobs)(delayed(extract_surface)(input_fname=input_fname,
                                                            num_rotations=args.num_rotations,
                                                            threshold=args.threshold,
                                                            output_dir=args.output_dir,
                                                            compress_ndarrays=args.compress_ndarrays) for input_fname in
                                   input_fnames)
if __name__ == "__main__":
    main()
