import torch
import torch.nn as nn
class DenseNetBlock3D(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 padding: int):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.in_layer = nn.InstanceNorm3d(num_features=in_channels)
        self.cnn_layer = nn.Conv3d(in_channels=self.in_channels,
                                   out_channels=self.out_channels,
                                   kernel_size=self.kernel_size,
                                   padding=self.padding)
    @staticmethod
    def swish(x):
        return x * torch.sigmoid(x)
    def forward(self, x):
        return self.swish(self.cnn_layer(self.in_layer(x)))
class DenseNetDiscriminatorBlock3D(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 padding: int):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.in_layer = nn.InstanceNorm3d(num_features=in_channels)
        self.cnn_layer = nn.Conv3d(in_channels=self.in_channels,
                                   out_channels=2 * self.out_channels,
                                   kernel_size=self.kernel_size,
                                   padding=self.padding)
    @staticmethod
    def swish(x):
        return x * torch.sigmoid(x)
    def glu(self, x):
        return self.swish(x[:, :self.out_channels, ...]) * x[:, self.out_channels:, ...]
    def forward(self, x):
        return self.glu(self.cnn_layer(self.in_layer(x)))
class DenseNet3D(nn.Module):
    def __init__(self,
                 num_layers: int,
                 num_channels: int,
                 kernel_size: int,
                 is_discriminator=False):
        super().__init__()
        assert num_layers > 0, "Number of layers has to be non-negative!"
        assert num_channels > 0, "Number of channels has to be non-negative!"
        assert kernel_size % 2 == 1, "Kernel size has to be odd!"
        self.num_layers = num_layers
        self.num_channels = num_channels
        self.kernel_size = kernel_size
        self.is_discriminator = is_discriminator
        padding = (kernel_size - 1) // 2
        if is_discriminator:
            self.blocks = nn.ModuleList(DenseNetDiscriminatorBlock3D(in_channels=2 + num_channels * layer_id,
                                                                     out_channels=num_channels,
                                                                     kernel_size=kernel_size,
                                                                     padding=padding) for layer_id in range(num_layers))
        else:
            self.blocks = nn.ModuleList(DenseNetBlock3D(in_channels=1 + num_channels * layer_id,
                                                        out_channels=num_channels,
                                                        kernel_size=kernel_size,
                                                        padding=padding) for layer_id in range(num_layers))
    def forward(self, x, y=None):
        x = torch.cat([x, y], dim=1)
        intermediate_features = []
        for block in self.blocks:
            intermediate_features.append(x)
            x = block(torch.cat(intermediate_features, dim=1))
        x = torch.mean(x, dim=1, keepdim=True)
        if self.is_discriminator:
            print()
        return x
