from argparse import ArgumentParser
import pathlib
import yaml
from typing import Tuple
import numpy as np
import h5py
from scipy.ndimage import zoom
from skimage.morphology import binary_closing, disk
from remodeling.generative.CPGAN.orchestrator import strip_distributed_markers, build_resolutions, GAN
from remodeling.generative.CPGAN.network_no_bottleneck import Generator as GeneratorNoBottleneck
from remodeling.generative.CPGAN.network_bottleneck import Generator as GeneratorBottleneck, binary_logarithm
from remodeling.generative.CPGAN.custom_layers import UpsamplingBlock3d
import torch
from torch.utils.data import DataLoader, Dataset as AbstractDataset
from utils import save_lzma_ndarray_archive
from tqdm import tqdm
def extract_global_step(path: pathlib.Path):
    return int(path.stem.split("_")[1])
def upsampling(tensor: torch.FloatTensor, scale_factor: int):
    block_upsampling = UpsamplingBlock3d()
    for _ in range(binary_logarithm(scale_factor)):
        tensor = block_upsampling(tensor)
    return tensor
class Dataset(AbstractDataset):
    def __init__(self,
                 dimensionality: int,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 min_max_slice: Tuple[int, int],
                 num_samples: int,
                 zoom_factor: float):
        super().__init__()
        self.dimensionality = dimensionality
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.min_max_slice = min_max_slice
        self.zoom_factor = zoom_factor
        self.num_samples = num_samples
        self.resolution = None
        self.ix = [line.strip() for line in open(self.index_fname, "r")]
        self.ix = self.ix[:self.num_samples]
    def __len__(self):
        return len(self.ix)
    def center_image(self, arr):
        if arr.shape[0] > self.resolution[0]:
            diff_shape = [u - v for u, v in zip(arr.shape, self.resolution)]
            slice_def = [slice(s // 2, u - (s + 1) // 2) for s, u in zip(diff_shape, arr.shape)]
            return arr[slice_def]
        elif arr.shape[0] < self.resolution[0]:
            diff_shape = [v - u for u, v in zip(arr.shape, self.resolution)]
            pad_def = [(s // 2, s - s // 2) for s in diff_shape]
            return np.pad(arr,
                          pad_width=pad_def,
                          mode="constant",
                          constant_values=np.min(arr))
        return arr
    def __getitem__(self, i):
        with h5py.File(self.data_fname, "r") as fhandle:
            example_id = self.ix[i]
            original = fhandle["original"][example_id][()]
            convex_hull = fhandle["convex_hull"][example_id][()]
            brain_mask = fhandle["brain_mask"][example_id][()]
            if self.resolution is None:
                self.resolution = original.shape
            original = self.center_image(zoom(original, zoom=self.zoom_factor, order=1))
            convex_hull = self.center_image(zoom(convex_hull, zoom=self.zoom_factor, order=1))
            brain_mask = self.center_image(zoom(brain_mask, zoom=self.zoom_factor, order=1))
            assert all(a == b for a, b in zip(original.shape, self.resolution)), "Shape mismatch!"
            selem = disk(radius=3)
            selem_3d = selem[np.newaxis, :, :] * selem[:, np.newaxis, :] * selem[:, :, np.newaxis]
            brain_mask = binary_closing(brain_mask, selem=selem_3d)
            convex_hull = binary_closing(convex_hull, selem=selem_3d)
            conditional = convex_hull.astype(np.float32)
            brain = np.where(brain_mask, original, np.full_like(original, fill_value=np.min(original)))
            brain_mask = brain_mask.astype(np.float32)
            slice_index = np.full_like(original, fill_value=0)
            data_dict = {
                "example_id": example_id,
                "conditional": conditional,
                "original": original,
                "brain": brain,
                "brain_mask": brain_mask,
                "slice_index": slice_index
            }
            return {k: torch.from_numpy(v).unsqueeze(dim=0) if k != "example_id" else v for k, v in data_dict.items()}
def main():
    parser = ArgumentParser()
    parser.add_argument("--model_dir",
                        help="Where the model directory is located",
                        required=True)
    parser.add_argument("--data_hdf5_fname",
                        required=True)
    parser.add_argument("--index_fname",
                        help="A YAML file that contains indices of the samples",
                        required=True)
    parser.add_argument("--zoom_factors",
                        help="The various zoom factors",
                        nargs="+",
                        default=(0.85, 1.15))
    parser.add_argument("--zoom_step",
                        type=float,
                        default=.05)
    parser.add_argument("--num_samples",
                        help="How many samples per zoom level",
                        type=int,
                        default=8)
    args = parser.parse_args()
    args.model_dir = pathlib.Path(args.model_dir)
    args.data_hdf5_fname = pathlib.Path(args.data_hdf5_fname)
    args.index_fname = pathlib.Path(args.index_fname)
    args.output_dir = args.data_hdf5_fname.parent.parent.joinpath("RESCALE")
    args.output_dir.mkdir(exist_ok=True)
    args.zoom_factors = np.arange(start=args.zoom_factors[0], stop=args.zoom_factors[1] + args.zoom_step, step=args.zoom_step)
    assert args.model_dir.exists(), "Model directory does not exist!"
    assert args.data_hdf5_fname.exists(), "Data does not exist!"
    assert args.index_fname.exists(), "Index does not exist!"
    global_steps = sorted(map(extract_global_step, args.model_dir.joinpath("checkpoints").glob("*.pt")))
    args.global_step = global_steps[-1]
    args.checkpoint_fname = args.model_dir.joinpath("checkpoints", f"checkpoint_{args.global_step}.pt")
    assert args.checkpoint_fname.exists(), "Checkpoint does not exist"
    config_fname = args.model_dir.joinpath("config.yml")
    with open(config_fname, "r") as config_f:
        config = yaml.load(config_f, Loader=yaml.FullLoader)
    if config['architecture'] == "bottleneck":
        generator_factory = GeneratorBottleneck
    elif config['architecture'] == "no_bottleneck":
        generator_factory = GeneratorNoBottleneck
    else:
        raise NotImplementedError()
    config['generator'].pop('use_ema')
    config['generator'].pop('ema_decay')
    config['generator'].pop('optimizer')
    num_blocks = binary_logarithm(config['max_size']) - binary_logarithm(config['min_size'])
    generator = generator_factory(dimensionality=config['dimensionality'],
                                  min_size=config['min_size'],
                                  num_blocks=num_blocks,
                                  **config['generator'])
    device = torch.device("cuda:0")
    ckpt = strip_distributed_markers(torch.load(args.checkpoint_fname, map_location=device))
    generator.load_state_dict(ckpt['generator'])
    generator = generator.to(device)
    with torch.no_grad():
        for zoom_factor in args.zoom_factors:
            zoom_output_dir = args.output_dir.joinpath(f"zoom_{zoom_factor}")
            zoom_output_dir.mkdir(exist_ok=True)
            data = Dataset(dimensionality=3,
                           data_fname=args.data_hdf5_fname,
                           index_fname=args.index_fname,
                           min_max_slice=config['min_max_slice'],
                           num_samples=args.num_samples,
                           zoom_factor=zoom_factor)
            data = DataLoader(data,
                              batch_size=1,
                              shuffle=False,
                              pin_memory=True)
            for example in tqdm(data):
                example_id = example.pop('example_id')[0]
                batch = example
                batch = {k: t.to(device, non_blocking=True) for k, t in batch.items()}
                batch = {k: [t.contiguous() for t in build_resolutions(t, dimensionality=3, n=num_blocks - 1)] for k, t in batch.items()}
                original = batch.pop('original')
                fake_gray = generator(**batch)
                convex_hull = batch["conditional"]
                brain_mask = batch["brain_mask"]
                original_min_max = [(original_t.min(), original_t.max()) for original_t in original]
                original = [torch.div(original_t - original_min, original_max - original_min) for
                            original_t, (original_min, original_max) in
                            zip(original, original_min_max)]
                fake_gray = [torch.clamp(torch.div(fake_gray_t - original_min, original_max - original_min), 0, 1) for
                             fake_gray_t, (original_min, original_max) in zip(fake_gray, original_min_max)]
                scale_factors = [config['max_size'] // original_t.shape[-1] for original_t in original]
                resolutions = [original_t.shape[-1] for original_t in original]
                original = [upsampling(tensor=original_t,
                                       scale_factor=scale_factor) for original_t, scale_factor in
                            zip(original, scale_factors)]
                convex_hull = [torch.bernoulli(upsampling(tensor=convex_hull_t.type(torch.float32),
                                                          scale_factor=scale_factor)).type(torch.bool) for
                               convex_hull_t, scale_factor in zip(convex_hull, scale_factors)]
                brain_mask = [torch.bernoulli(upsampling(tensor=brain_mask_t.type(torch.float32),
                                                         scale_factor=scale_factor)).type(torch.bool) for
                              brain_mask_t, scale_factor in zip(brain_mask, scale_factors)]
                fake_gray = [upsampling(tensor=fake_gray_t,
                                        scale_factor=scale_factor) for fake_gray_t, scale_factor in
                             zip(fake_gray, scale_factors)]
                fake_gray = GAN.starmap(GAN.copy_brain_into_tensor, zip(fake_gray, original, brain_mask))
                original = [original_arr.cpu().numpy() for original_arr in original]
                convex_hull = [convex_hull_arr.cpu().numpy() for convex_hull_arr in convex_hull]
                brain_mask = [brain_mask_arr.cpu().numpy() for brain_mask_arr in brain_mask]
                fake_gray = [fake_gray_arr.cpu().numpy() for fake_gray_arr in fake_gray]
                original = {f"original_{r}": np.squeeze(v) for r, v in zip(resolutions, original)}
                convex_hull = {f"convex_hull_{r}": np.squeeze(v) for r, v in zip(resolutions, convex_hull)}
                brain_mask = {f"brain_mask_{r}": np.squeeze(v) for r, v in zip(resolutions, brain_mask)}
                fake_gray = {f"fake_gray_{r}": np.squeeze(v) for r, v in zip(resolutions, fake_gray)}
                example_data = {**original, **brain_mask, **fake_gray, **convex_hull}
                save_lzma_ndarray_archive(output_fname=zoom_output_dir.joinpath(f"{example_id}.npy.tar.xz"), **example_data)
if __name__ == "__main__":
    main()
