from argparse import ArgumentParser
from joblib import delayed, Parallel
import pathlib
import numpy as np
import nibabel as nib
from tqdm import tqdm
from utils import save_lzma_ndarray
def get_stem(fname: pathlib.Path):
    return fname.name.split(".")[0]
def convert_to_ndarray(input_fname: pathlib.Path,
                       output_dir: pathlib.Path,
                       compress_ndarrays: bool):
    output_fname = output_dir.joinpath(get_stem(input_fname)).with_suffix(".npy")
    if output_fname.exists():
        return
    try:
        img = nib.load(str(input_fname))
        arr = img.get_fdata(dtype=np.float32)
        if compress_ndarrays:
            save_lzma_ndarray(arr=arr, output_fname=output_fname)
        else:
            np.save(file=output_fname, arr=arr)
    except nib.filebasedimages.ImageFileError:
        print(f"Error with file {input_fname}")
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="Where the NifTi files are stored",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the NdArray files will be stored",
                        required=True)
    parser.add_argument("--num_jobs",
                        type=int,
                        default=1)
    parser.add_argument("--compress_ndarrays",
                        help="Whether to LZMA compress each saved ndarray",
                        action="store_true",
                        default=False)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.output_dir = pathlib.Path(args.output_dir)
    args.output_dir.mkdir(exist_ok=True)
    assert args.input_dir.exists(), "Input directory does not exist!"
    Parallel(n_jobs=args.num_jobs)(delayed(convert_to_ndarray)(input_fname=input_fname,
                                                               output_dir=args.output_dir,
                                                               compress_ndarrays=args.compress_ndarrays) for input_fname in tqdm(args.input_dir.glob("*.nii.gz")))
if __name__ == "__main__":
    main()
