from argparse import ArgumentParser
import pathlib
import numpy as np
import h5py
from tqdm import tqdm
import png
from utils import save_lzma_ndarray
def zscore(arr):
    if arr.dtype == np.bool:
        arr = arr.astype(np.float32)
    else:
        arr = np.divide(arr - arr.min(), arr.max() - arr.min())
    arr *= 255
    return arr.astype(np.uint8)
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_fname",
                        help="The HDF5 input file",
                        required=True)
    parser.add_argument("--slice_dimension",
                        type=int,
                        default=2,
                        choices=(0, 1, 2))
    parser.add_argument("--output_format",
                        help="Which output format to use",
                        choices=('png', 'ndarray'),
                        default='png')
    parser.add_argument("--num_slices_around_center",
                        help="Number of slices to take around the center",
                        type=int,
                        default=30)
    args = parser.parse_args()
    args.input_fname = pathlib.Path(args.input_fname)
    args.output_dir = args.input_fname.parent.with_name(args.input_fname.parent.name.replace("HDF5", "NDARRAY" if args.output_format == 'ndarray' else 'PNG') + "_Slices")
    args.output_dir.mkdir(exist_ok=True)
    assert args.input_fname.exists(), "Input file does not exists!"
    with h5py.File(args.input_fname, mode="r") as f:
        original = f["original"]
        convex_hull = f["convex_hull"]
        brain_mask = f["brain_mask"]
        ix = list(original)
        for ident in tqdm(ix):
            original_arr = original[ident][()]
            convex_hull_arr = convex_hull[ident][()]
            brain_mask_arr = brain_mask[ident][()]
            if args.output_format == "png":
                original_arr = zscore(original_arr)
                convex_hull_arr = zscore(convex_hull_arr)
                brain_mask_arr = zscore(brain_mask_arr)
            num_slices = original_arr.shape[args.slice_dimension]
            center_slice_id = num_slices // 2
            for slice_id in range(center_slice_id - args.num_slices_around_center, center_slice_id + args.num_slices_around_center):
                original_slice = np.take(original_arr, indices=slice_id, axis=args.slice_dimension)
                convex_hull_slice = np.take(convex_hull_arr, indices=slice_id, axis=args.slice_dimension)
                brain_mask_slice = np.take(brain_mask_arr, indices=slice_id, axis=args.slice_dimension)
                if args.output_format == "png":
                    png.from_array(original_slice, 'L').save(args.output_dir.joinpath(f"{ident}_original_{slice_id}.png"))
                    png.from_array(convex_hull_slice, 'L').save(args.output_dir.joinpath(f"{ident}_convex_hull_{slice_id}.png"))
                    png.from_array(brain_mask_slice, 'L').save(args.output_dir.joinpath(f"{ident}_brain_mask_{slice_id}.png"))
                else:
                    save_lzma_ndarray(original_slice, output_fname=args.output_dir.joinpath(f"{ident}_original_{slice_id}.npy.xz"))
                    save_lzma_ndarray(convex_hull_slice, output_fname=args.output_dir.joinpath(f"{ident}_convex_hull_{slice_id}.npy.xz"))
                    save_lzma_ndarray(brain_mask_slice, output_fname=args.output_dir.joinpath(f"{ident}_brain_mask_{slice_id}.npy.xz"))
if __name__ == "__main__":
    main()
