from argparse import ArgumentParser
import pathlib
import numpy as np
from typing import List, Tuple, Dict
from tqdm import tqdm
from joblib import delayed, Parallel
import nibabel as nib
from nilearn.image import resample_img
from anonymization.prepare_dl_dataset import pad_to_output_size
from utils import get_real_stem
""" c.f. https://www.kaggle.com/mechaman/resizing-reshaping-and-resampling-nifti-files """
def downscale_and_save_nifti(input_fname: pathlib.Path,
                             resolutions: List[Tuple[int, int, int]],
                             output_dirs: Dict[int, pathlib.Path]):
    try:
        img = nib.load(input_fname)
        img_arr = img.get_fdata(dtype=np.float32)
        img_arr_padded = pad_to_output_size(img_arr)
        img_padded = nib.Nifti1Image(img_arr_padded, affine=img.affine, header=img.header)
        for resolution in resolutions:
                target_affine = img_padded.affine.copy()
                target_affine[:3, :3] = 256/resolution[0] * target_affine[:3, :3]
                img_downscaled = resample_img(img=img_padded,
                                              target_affine=target_affine,
                                              target_shape=resolution,
                                              interpolation="nearest")
                nib.save(img_downscaled, output_dirs[resolution[0]].joinpath(input_fname.name))
    except Exception:
        print(f"Error with file {input_fname.name}")
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="The location of the directory that contains the NifTi files to be resized",
                        required=True)
    parser.add_argument("--resolutions",
                        nargs="+",
                        type=int,
                        default=[128])
    parser.add_argument("--num_jobs",
                        type=int,
                        default=1)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.resolutions = [(resolution,) * 3 for resolution in args.resolutions]
    args.output_dirs = {resolution[0]: args.input_dir.parent.joinpath(f"{args.input_dir.name}_{resolution[0]}") for resolution in args.resolutions}
    for output_dir in args.output_dirs.values():
        output_dir.mkdir(exist_ok=True)
    args.input_dir.exists(), "Input directory does not exist!"
    Parallel(n_jobs=args.num_jobs)(delayed(downscale_and_save_nifti)(input_fname=input_fname,
                                                                     resolutions=args.resolutions,
                                                                     output_dirs=args.output_dirs) for input_fname in tqdm(args.input_dir.glob("*.nii.gz")))
if __name__ == "__main__":
    main()
