import os
from argparse import ArgumentParser
from pathlib import Path
import shlex
import subprocess
import numpy as np
import tempfile
import shutil
import re
from typing import List, Tuple
from abc import ABC, abstractmethod
from joblib import delayed, Parallel
from tqdm import tqdm
import nibabel as nib
from nibabel.orientations import ornt_transform, axcodes2ornt, inv_ornt_aff, apply_orientation, io_orientation
import conversion.ants_registration as ants_registration
from utils import get_file_extension
""" Executables """
HOME_DIR = os.path.join(os.path.expanduser("~"))
os.environ["PATH"] += os.pathsep + os.pathsep.join([os.path.join(HOME_DIR, "Build/dpat-tools/LIBRARY/"),
                                                    os.path.join(HOME_DIR, "FSL/bin/")])
os.environ["FSLWISH"] = os.path.join(HOME_DIR, "FSL/bin/fslwish")
os.environ["FSLDIR"] = os.path.join(HOME_DIR, "FSL")
os.environ["FSLMACHINELIST"] = ""
os.environ["FSLTCLSH"] = os.path.join(HOME_DIR, "FSL/bin/fsltclsh")
os.environ["FSLREMOTECALL"] = ""
os.environ["FSLLOCKDIR"] = ""
os.environ["FSLGECUDAQ"] = "cuda.q"
os.environ["FSLOUTPUTTYPE"] = "NIFTI_GZ"
os.environ["FSLMULTIFILEQUIT"] = "TRUE"
os.environ["ANTSPATH"] = os.path.join(HOME_DIR, "Build/antsbin/")
class NifTiTransformation(ABC):
    """ File Transformation Baseclass """
    def __init__(self):
        pass
    @abstractmethod
    def __call__(self, input_fname: Path, output_fname: Path) -> None:
        pass
class FixOrientation(NifTiTransformation):
    """ Reorient NifTi to TARGET_ORIENTATION """
    def __init__(self,
                 target_orientation: Tuple[str, str, str] = ('L', 'A', 'S')):
        super().__init__()
        self.target_orientation = target_orientation
    def __call__(self, input_fname: Path, output_fname: Path) -> None:
        img = nib.load(input_fname)
        new_ornt = axcodes2ornt(self.target_orientation)
        vox_array = img.get_fdata()
        affine = img.affine
        orig_ornt = io_orientation(img.affine)
        ornt_trans = ornt_transform(orig_ornt, new_ornt)
        orig_shape = vox_array.shape
        new_vox_array = apply_orientation(vox_array, ornt_trans)
        aff_trans = inv_ornt_aff(ornt_trans, orig_shape)
        new_affine = np.dot(affine, aff_trans)
        reoriented_img = nib.Nifti1Image(new_vox_array, new_affine, img.header)
        nib.save(reoriented_img, output_fname)
class CropNeck(NifTiTransformation):
    """ Apply FSL's robustfov  """
    def __init__(self,
                 brain_size: int = 170):
        super().__init__()
        self.brain_size = brain_size
        self.extract_fov = re.compile(r"(\d+).\d+")
    def __call__(self, input_fname: Path, output_fname: Path) -> None:
        fov_data = subprocess.check_output(shlex.split(f"robustfov -b {self.brain_size} -i {input_fname} --debug"))
        fov_data = self.extract_fov.findall(fov_data.decode("utf-8").split(os.linesep)[0])
        """
        fov_data[0] = '0'
        fov_data[1] = '-1'
        fov_data[2] = '0'
        fov_data[3] = '-1'
        fov_data[4] = '0'
        fov_data[5] = '-1'
        """
        subprocess.check_output(shlex.split(f"fslroi {input_fname} {output_fname} " + " ".join(fov_data)))
class BiasFieldCorrection(NifTiTransformation):
    """ Apply N4BiasFieldCorrection """
    def __init__(self):
        super().__init__()
    def __call__(self, input_fname: Path, output_fname: Path) -> None:
        subprocess.check_output(shlex.split(
            f"{os.environ['ANTSPATH']}/ANTS-build/Examples/N4BiasFieldCorrection -i {input_fname} -o {output_fname} -d 3"))
class CoRegistration(NifTiTransformation):
    """ Apply ANTs co-registration  """
    def __init__(self, fixed_fname: Path):
        super().__init__()
        self.fixed_fname = fixed_fname
    def __call__(self, input_fname: Path, output_fname: Path) -> None:
        ants_registration.ants_registration(input_fname=input_fname, fixed_fname=self.fixed_fname, output_fname=output_fname)
class FileProcessor:
    """ Applies chain of FileTransformations on MRI image  """
    def __init__(self,
                 dataset: str,
                 nifti_transformations: List[NifTiTransformation],
                 output_dir: Path,
                 debug: bool):
        self.dataset = dataset
        self.file_transformations = nifti_transformations
        self.output_dir = output_dir
        self.debug = debug
        if self.debug:
            self.intermediate_dirs = dict()
            for file_transformation in self.file_transformations:
                class_name = file_transformation.__class__.__name__
                intermediate_dir = output_dir.joinpath(class_name)
                intermediate_dir.mkdir(exist_ok=True)
                self.intermediate_dirs[class_name] = intermediate_dir
    def __call__(self, input_fname: Path):
        output_fname = self.output_dir.joinpath(input_fname.name)
        if output_fname.exists():
            return
        file_extension = get_file_extension(input_fname)
        assert file_extension.lower() in (".nii", ".nii.gz"), f"Incompatible file type {file_extension} for file {input_fname}"
        try:
            with tempfile.NamedTemporaryFile(suffix=file_extension) as tmp_f,\
                    tempfile.NamedTemporaryFile(suffix=file_extension) as next_tmp_f:
                tmp_fname, next_tmp_fname = Path(tmp_f.name), Path(next_tmp_f.name)
                shutil.copy2(src=input_fname, dst=tmp_fname)
                for file_transformation in self.file_transformations:
                    file_transformation(input_fname=tmp_fname, output_fname=next_tmp_fname)
                    if self.debug:
                        class_name = file_transformation.__class__.__name__
                        intermediate_dir = self.intermediate_dirs[class_name]
                        shutil.copy2(src=next_tmp_fname, dst=intermediate_dir.joinpath(input_fname.name))
                    shutil.copy2(src=next_tmp_fname, dst=tmp_fname)
                shutil.copy2(src=tmp_fname, dst=output_fname)
        except subprocess.CalledProcessError:
            print(f"File {input_fname} is corrupted!")
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        required=True)
    parser.add_argument("--output_dir",
                        required=True)
    parser.add_argument("--dataset",
                        choices=["ADNI", "OASIS-3", "IXI"],
                        required=True)
    parser.add_argument("--debug",
                        help="Don't delete temporary files",
                        default=False,
                        action="store_true")
    parser.add_argument("--registration",
                        help="Whether to do registration",
                        default=False,
                        action="store_true")
    parser.add_argument("--template",
                        help="Registration template",
                        required=True)
    parser.add_argument("--num_jobs",
                        help="How many jobs to spawn",
                        type=int,
                        default=1)
    args = parser.parse_args()
    args.input_dir = Path(args.input_dir)
    args.output_dir = Path(args.output_dir)
    args.template = Path(args.template)
    assert args.input_dir.exists(), "Input directory does not exist!"
    assert args.template.exists(), "Template does not exist!"
    args.output_dir.mkdir(exist_ok=True)
    input_fnames = args.input_dir.glob("*.nii.gz")
    nifti_transformations = [FixOrientation(),
                             BiasFieldCorrection()]
    if args.registration:
        nifti_transformations.append(CoRegistration(args.template))
    file_processor = FileProcessor(dataset=args.dataset,
                                   nifti_transformations=nifti_transformations,
                                   output_dir=args.output_dir,
                                   debug=args.debug)
    Parallel(n_jobs=args.num_jobs)(delayed(file_processor.__call__)(input_fname=input_fname) for input_fname in tqdm(input_fnames))
if __name__ == "__main__":
    main()
