import torch
import torch.nn as nn

INPUT_SIZE = 128
KERNEL_SIZE = 3
BRANCH_FACTOR = 2
K = 16

def binary_logarithm(n):
    assert n & (n - 1) == 0, "Input needs to be a power of two!"
    bitstring = bin(n)[2:]
    return len(bitstring) - 1

class Layer(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size

        self.conv = nn.Conv3d(in_channels=in_channels,
                              out_channels=out_channels,
                              kernel_size=kernel_size,
                              stride=2,
                              padding=(kernel_size - 1) // 2)
        self.norm = nn.BatchNorm3d(num_features=out_channels)
        self.activation = nn.SiLU()

    def forward(self, x):
        return self.norm(self.activation(self.conv(x)))

class Network(nn.Module):
    def __init__(self):
        super().__init__()

        self.num_layers = binary_logarithm(INPUT_SIZE)
        self.layers = nn.ModuleList([Layer(in_channels=1 if i == 0 else K * BRANCH_FACTOR**i,
                                           out_channels=K * BRANCH_FACTOR ** (i + 1),
                                           kernel_size=KERNEL_SIZE) for i in range(self.num_layers)])
        self.fc = nn.Linear(in_features=K * BRANCH_FACTOR**self.num_layers, out_features=1)


    def forward(self, x):
        for layer in self.layers:
           x = layer(x)
        x = x.view(-1, K * BRANCH_FACTOR**self.num_layers)
        return self.fc(x).flatten()