import sys
from argparse import ArgumentParser
import pathlib
from tqdm import trange, tqdm
from tempfile import gettempdir
import string
import random
from torch.utils.tensorboard import SummaryWriter

from dl_age_estimation.data import Dataset
from dl_age_estimation.network import Network

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

import pandas as pd

device = torch.device("cuda:0")

def main():
    # Instantiate parser
    parser = ArgumentParser()

    parser.add_argument("--base_dir",
                        required=True)
    parser.add_argument("--reference_dir",
                        required=True)

    parser.add_argument("--train_ids_fname",
                        required=True)

    parser.add_argument("--test_ids_fname",
                        required=True)

    parser.add_argument("--age_fname",
                        required=True)

    parser.add_argument("--comparison_dirs",
                        nargs="+",
                        required=True)

    parser.add_argument("--num_epochs",
                        type=int,
                        required=True)

    # Parse
    args = parser.parse_args()

    # Input conversions
    args.base_dir = pathlib.Path(args.base_dir)
    args.reference_dir = args.base_dir.joinpath(args.reference_dir)
    args.train_ids_fname = args.base_dir.joinpath(args.train_ids_fname)
    args.test_ids_fname = args.base_dir.joinpath(args.test_ids_fname)
    args.age_fname = args.base_dir.joinpath(args.age_fname)
    args.comparison_dirs = list(map(args.base_dir.joinpath, args.comparison_dirs))

    # Input assertions
    assert args.base_dir.exists(), "Base directory does not exist!"
    assert args.reference_dir.exists(), "Reference directory does not exist!"
    assert args.train_ids_fname.exists(), "Train ids do not exist!"
    assert args.test_ids_fname.exists(), "Test ids do not exist!"
    assert args.age_fname.exists(), "Age metadata does not exist!"
    assert all(comparison_dir.exists() for comparison_dir in args.comparison_dirs), "Not all comparison directories exist!"

    # Temporary directory
    tmp_dir = pathlib.Path(gettempdir()).joinpath("AGE_NET_" + ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(16)))
    print(tmp_dir)
    tmp_dir.mkdir()

    # Tensorboard
    summary_writer = SummaryWriter(log_dir=tmp_dir)

    # Data
    data = Dataset(data_dir=args.reference_dir,
                   index_fname=args.train_ids_fname,
                   age_fname=args.age_fname)
    data = DataLoader(data,
                      batch_size=32,
                      shuffle=True)

    # Architecture
    net = Network().to(device=device)
    net = net.train()

    # Loss
    loss_fct = nn.L1Loss()
    optimizer = optim.Adam(net.parameters(), lr=0.001)

    global_step = 0
    for epoch in trange(args.num_epochs):
        for x in data:
            x.pop('example_id')
            optimizer.zero_grad()
            x = {k: t.to(device, non_blocking=True) for k, t in x.items()}
            y = net(x['arr'])
            loss = loss_fct(y, x['age'])
            loss.backward()
            optimizer.step()

            summary_writer.add_scalar(tag="loss", scalar_value=loss, global_step=global_step)
            global_step += 1

    # Evaluation
    net = net.eval()

    with torch.no_grad():
        results = dict()
        for comparison_dir in args.comparison_dirs:
            sub_results = dict()
            data = Dataset(data_dir=comparison_dir,
                           index_fname=args.test_ids_fname,
                           age_fname=args.age_fname)
            data = DataLoader(data, batch_size=1, shuffle=False)
            for x in tqdm(data):
                example_id = x.pop('example_id')[0]
                x = {k: t.to(device, non_blocking=True) for k, t in x.items()}
                age_gt = x['age'].item()
                age_pred = net(x['arr']).item()
                sub_results[example_id] = age_pred

            results[comparison_dir.stem] = sub_results

        results = pd.DataFrame(results)
        output_fname = args.base_dir.joinpath("age_estimation_nouveau.csv")
        results.to_csv(output_fname)


if __name__ == "__main__":
    main()