from argparse import ArgumentParser
import pathlib
from utils import get_real_stem
import random
import torch
from joblib import delayed, Parallel
import numpy as np
from functools import wraps, partial
from time import perf_counter, process_time
import tempfile
def measure_elapsed_time(f):
    @wraps(f)
    def wrap(*args, **kw):
        time_start = perf_counter()
        f(*args, **kw)
        time_end = perf_counter()
        return time_end - time_start
    return wrap
from skull_stripping import Robex
@measure_elapsed_time
def execute_robex(input_fname: pathlib.Path, output_fname: pathlib.Path = None):
    with tempfile.TemporaryDirectory() as tmp_dir:
        if output_fname is None:
            output_fname = pathlib.Path(tmp_dir).joinpath(input_fname.name)
        robex = Robex()
        robex.strip_skull(input_fname=input_fname, output_fname=output_fname)
""" QUICKSHEAR """
from traditional_anonymization_methods.quickshear import process_image_pair as quickshear
@measure_elapsed_time
def execute_quickshear(input_fname: pathlib.Path):
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        brain_fname = tmp_dir.joinpath(f"{get_real_stem(input_fname)}_mask.nii.gz")
        execute_robex(input_fname=input_fname, output_fname=brain_fname)
        output_fname = tmp_dir.joinpath(input_fname.name)
        quickshear(input_fname=input_fname, mask_fname=brain_fname, output_fname=output_fname)
""" MRI_WATERSHED """
try:
    from traditional_anonymization_methods import mri_watershed
except AssertionError:
    print("MRI Watershed could not be loaded!")
@measure_elapsed_time
def execute_mri_watershed(input_fname: pathlib.Path):
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        mri_watershed.execute_watershed(input_fname=input_fname, output_dir=tmp_dir)
""" FACE_MASK """
from conversion.convert_to_analyze import convert_to_analyze
from traditional_anonymization_methods import face_mask
@measure_elapsed_time
def execute_face_mask(input_fname: pathlib.Path):
    with tempfile.TemporaryDirectory() as tmp_in_dir, tempfile.TemporaryDirectory() as tmp_out_dir:
        tmp_in_dir = pathlib.Path(tmp_in_dir)
        tmp_out_dir = pathlib.Path(tmp_out_dir)
        convert_to_analyze(nifti_fname=input_fname, output_dir=tmp_in_dir)
        face_mask.orchestrate_docker(ids=[get_real_stem(input_fname)], input_dir=tmp_in_dir, output_dir=tmp_out_dir)
""" DEFACE """
from traditional_anonymization_methods.deface import execute_deface as deface
@measure_elapsed_time
def execute_deface(input_fname: pathlib.Path,
                   mri_deface_talairach_path: pathlib.Path,
                   mri_deface_face_path: pathlib.Path):
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        deface(nifti_fname=input_fname,
               output_dir=tmp_dir,
               mri_deface_talairach_path=mri_deface_talairach_path,
               mri_deface_face_path=mri_deface_face_path)
""" CP-GAN """
from conversion.convert_to_ndarrays import convert_to_ndarray
from normalization.normalize import min_max_normalization
from anonymization.surface_extraction import extract_surface
from anonymization.create_paired_dataset import combine
from remodeling.generative.CPGAN.orchestrator import build_resolutions
COMPRESS_NDARRAYS = False
@measure_elapsed_time
def execute_c_anon_gan(input_fname: pathlib.Path,
                       generator,
                       device: torch.device,
                       num_blocks: int):
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        brain_dir = tmp_dir.joinpath("brain"); brain_dir.mkdir()
        normalized_dir = tmp_dir.joinpath("normalized"); normalized_dir.mkdir()
        face_dir = tmp_dir.joinpath("face"); face_dir.mkdir()
        pair_dir = tmp_dir.joinpath("pair"); pair_dir.mkdir()
        brain_fname = tmp_dir.joinpath("brain", input_fname.name)
        execute_robex(input_fname=input_fname, output_fname=brain_fname)
        convert_to_ndarray(input_fname=brain_fname,
                           output_dir=brain_dir,
                           compress_ndarrays=COMPRESS_NDARRAYS)
        min_max_normalization(input_fname=input_fname,
                              brain_fname=brain_fname,
                              output_dir=normalized_dir,
                              compress_ndarrays=COMPRESS_NDARRAYS)
        input_normalized_fname = next(normalized_dir.iterdir())
        extract_surface(input_fname=input_normalized_fname,
                        num_rotations=3,
                        threshold=.75,
                        output_dir=face_dir,
                        compress_ndarrays=COMPRESS_NDARRAYS)
        input_face_fname = face_dir.joinpath(input_normalized_fname.name)
        input_brain_fname = brain_dir.joinpath(input_normalized_fname.name)
        paired_arrays = combine(input_original_fname=input_normalized_fname,
                                input_face_fname=input_face_fname,
                                input_brain_fname=input_brain_fname,
                                max_num_triangles=100,
                                return_as_arrays=True)
        with torch.no_grad():
            original = paired_arrays['original']
            convex_hull = paired_arrays['convex_hull']
            brain_mask = paired_arrays['brain_mask']
            conditional = convex_hull.astype(np.float32)
            brain = np.where(brain_mask, original, np.full_like(original, fill_value=np.min(original)))
            brain_mask = brain_mask.astype(np.float32)
            slice_index = np.full_like(original, fill_value=0)
            data_dict = {
                "conditional": conditional,
                "original": original,
                "brain": brain,
                "brain_mask": brain_mask,
                "slice_index": slice_index
            }
            batch = {k: torch.from_numpy(v).unsqueeze(dim=0).unsqueeze(dim=0)  for k, v in data_dict.items()}
            batch = {k: t.to(torch.device("cuda:0"), non_blocking=True) for k, t in batch.items()}
            batch = {k: [t.contiguous() for t in build_resolutions(t, dimensionality=3, n=num_blocks - 1)] for k, t in
                     batch.items()}
            batch.pop('original')
            generator(**batch)
METHODS = {
    'quickshear': execute_quickshear,
    'mri_watershed': execute_mri_watershed,
    'face_mask': execute_face_mask,
    'robex': execute_robex
}
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="A directory containing NifTi files",
                        required=True)
    parser.add_argument("--methods",
                        help="Which methods to compare",
                        nargs="+",
                        default=['quickshear', 'mri_watershed', 'face_mask',
                                 'deface', 'robex', 'cp-gan'])
    parser.add_argument("--index_fname",
                        help="On which indices should the methods be run",
                        required=True)
    parser.add_argument("--c_anon_gan_model_dir",
                        help="Where the model definition is located")
    parser.add_argument("--single_process",
                        action="store_true",
                        default=False)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.methods = sorted(args.methods)
    args.index_fname = pathlib.Path(args.index_fname)
    if 'cp-gan' in args.methods:
        assert args.c_anon_gan_model_dir is not None, "Model directory not specified!"
        args.c_anon_gan_model_dir = pathlib.Path(args.c_anon_gan_model_dir)
        assert args.c_anon_gan_model_dir.exists(), "Model directory does not exist!"
    assert args.input_dir.exists(), "Input directory does not exist!"
    assert set(args.methods) <= {'quickshear', 'mri_watershed', 'face_mask',
                                 'deface', 'robex', 'cp-gan'}, "Some methods are not supported!"
    assert args.index_fname.exists(), "Index does not exist!"
    example_ids = args.index_fname.read_text().strip().split("\n")
    if 'deface' in args.methods:
        from traditional_anonymization_methods.deface import MRI_DEFACE_FACE_PATH, MRI_DEFACE_TALAIRACH_PATH
        from traditional_anonymization_methods.deface import gunzip_and_move_to_tmp_dir
        tmp_dir = pathlib.Path(tempfile.gettempdir())
        mri_deface_face_path = gunzip_and_move_to_tmp_dir(fname=MRI_DEFACE_FACE_PATH, tmp_dir=tmp_dir)
        mri_deface_talairach_path = gunzip_and_move_to_tmp_dir(fname=MRI_DEFACE_TALAIRACH_PATH, tmp_dir=tmp_dir)
        METHODS['deface'] = partial(execute_deface,
                                    mri_deface_talairach_path=mri_deface_talairach_path,
                                    mri_deface_face_path=mri_deface_face_path)
    if 'cp-gan' in args.methods:
        from remodeling.evaluation.evaluate_canon_gan import extract_global_step
        from remodeling.generative.CPGAN.network_bottleneck import Generator as GeneratorBottleneck
        from remodeling.generative.CPGAN.network_no_bottleneck import Generator as GeneratorNoBottleneck
        from remodeling.generative.CPGAN.orchestrator import binary_logarithm, strip_distributed_markers
        import yaml
        global_steps = sorted(map(extract_global_step, args.c_anon_gan_model_dir.joinpath("checkpoints").glob("*.pt")))
        global_step = global_steps[-1]
        checkpoint_fname = args.c_anon_gan_model_dir.joinpath("checkpoints", f"checkpoint_{global_step}.pt")
        assert checkpoint_fname.exists(), "Checkpoint does not exist"
        config_fname = args.c_anon_gan_model_dir.joinpath("config.yml")
        with open(config_fname, "r") as config_f:
            config = yaml.load(config_f, Loader=yaml.FullLoader)
        if config['architecture'] == "bottleneck":
            generator_factory = GeneratorBottleneck
        elif config['architecture'] == "no_bottleneck":
            generator_factory = GeneratorNoBottleneck
        else:
            raise NotImplementedError()
        config['generator'].pop('use_ema')
        config['generator'].pop('ema_decay')
        config['generator'].pop('optimizer')
        num_blocks = binary_logarithm(config['max_size']) - binary_logarithm(config['min_size'])
        generator = generator_factory(dimensionality=config['dimensionality'],
                                      min_size=config['min_size'],
                                      num_blocks=num_blocks,
                                      **config['generator'])
        device = torch.device("cuda:0")
        ckpt = strip_distributed_markers(torch.load(checkpoint_fname, map_location=device))
        generator.load_state_dict(ckpt['generator'])
        generator = generator.to(device)
        METHODS['cp-gan'] = partial(execute_c_anon_gan,
                                        generator=generator,
                                        device=device,
                                        num_blocks=num_blocks)
    n_jobs = 1 if args.single_process else len(args.methods)
    output_fname = args.index_fname.parent.joinpath("execution_time_measurement.csv")
    add_header = not output_fname.exists()
    with open(output_fname, "a") as f_out:
        if add_header:
            f_out.write(",".join(args.methods) + "\n")
        for n, example_id in enumerate(example_ids):
            input_fname = args.input_dir.joinpath(f"{example_id}.nii.gz")
            assert input_fname.exists(), "Input file does not exist!"
            def _launch_method(method):
                torch.cuda.empty_cache()
                return method, METHODS[method](input_fname=input_fname)
            example_measurement = dict(Parallel(n_jobs=n_jobs,
                                                backend="loky",
                                                pre_dispatch=n_jobs,
                                                batch_size=1)(delayed(_launch_method)(method=method)
                                                               for method in
                                                               random.sample(args.methods, k=len(args.methods))))
            f_out.write(",".join(["{:.3f}".format(example_measurement[method]) for method in args.methods]))
            f_out.write("\n")
            f_out.flush()
if __name__ == "__main__":
    main()
