from argparse import ArgumentParser
import pathlib
import numpy as np
import yaml
import nibabel as nib
from joblib import delayed, Parallel
from tqdm import tqdm
from utils import save_lzma_ndarray
def min_max_normalization(input_fname: pathlib.Path,
                          brain_fname: pathlib.Path,
                          output_dir: pathlib.Path,
                          compress_ndarrays: bool):
    try:
        output_fname = output_dir.joinpath(input_fname.name.replace(".nii.gz", ".npy"))
        img = nib.load(input_fname)
        img_brain = nib.load(brain_fname)
        arr = img.get_fdata(dtype=np.float64)
        arr_brain = img_brain.get_fdata(dtype=np.float64)
        mask = arr_brain > 0.
        """
        brain_voxels = np.compress(mask.reshape(-1), arr.reshape(-1))
        mu_brain, std_brain = np.mean(brain_voxels), np.std(brain_voxels)
        arr_rescaled = (arr - mu_brain)/std_brain
        """
        mu, std = np.mean(arr), np.std(arr)
        arr_rescaled = (arr - mu)/std
        if compress_ndarrays:
            save_lzma_ndarray(arr=arr_rescaled.astype(np.float32), output_fname=output_fname)
        else:
            np.save(file=str(output_fname).replace(".nii.gz", ".npy"), arr=arr_rescaled)
        return {
            "id": input_fname.name.split(".")[0],
            "mu": str(mu),
            "std": str(std)
        }
    except ValueError:
        pass
    except FileNotFoundError:
        pass
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="The directory where the (unnormalized) NifTi files reside",
                        required=True)
    parser.add_argument("--brain_dir",
                        help="The directory where the (unnormalized) brain crops reside",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the normalized NdArray files will be stored",
                        required=True)
    parser.add_argument("--num_jobs",
                        type=int,
                        default=1)
    parser.add_argument("--compress_ndarrays",
                        help="Whether to LZMA compress each saved ndarray",
                        action="store_true",
                        default=False)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.brain_dir = pathlib.Path(args.brain_dir)
    args.output_dir = pathlib.Path(args.output_dir)
    args.input_dir.exists(), "Input directory does not exist!"
    args.brain_dir.exists(), "Brain crop directory does not exist!"
    fnames = [{"input_fname": input_fname,
               "brain_fname": args.brain_dir.joinpath(input_fname.name)} for input_fname in args.input_dir.glob("*.nii.gz")]
    args.output_dir.mkdir(exist_ok=True)
    metadata = Parallel(n_jobs=args.num_jobs)(delayed(min_max_normalization)(**fname,
                                                                             output_dir=args.output_dir,
                                                                             compress_ndarrays=args.compress_ndarrays) for fname in tqdm(fnames))
    with open(args.output_dir.joinpath("metadata.yml"), "w") as f:
        yaml.dump(data=metadata, stream=f, default_flow_style=False)
if __name__ == "__main__":
    main()
