from argparse import ArgumentParser
import pathlib
import yaml
from tqdm import tqdm
from remodeling.generative.CPGAN.orchestrator import strip_distributed_markers, build_resolutions, GAN
from remodeling.generative.CPGAN.network_no_bottleneck import Generator as GeneratorNoBottleneck
from remodeling.generative.CPGAN.network_bottleneck import Generator as GeneratorBottleneck, binary_logarithm
from remodeling.generative.CPGAN.custom_layers import UpsamplingBlock3d
from remodeling.generative.CPGAN.data import Dataset
from utils import save_lzma_ndarray_archive
import numpy as np
import torch
from torch.utils.data import DataLoader
def extract_global_step(path: pathlib.Path):
    return int(path.stem.split("_")[1])
def upsampling(tensor: torch.FloatTensor, scale_factor: int):
    block_upsampling = UpsamplingBlock3d()
    for _ in range(binary_logarithm(scale_factor)):
        tensor = block_upsampling(tensor)
    return tensor
def main():
    parser = ArgumentParser()
    parser.add_argument("--model_dir",
                        help="Where the model directory resides",
                        required=True)
    parser.add_argument("--data_hdf5_fname",
                        required=True)
    parser.add_argument("--index_fname",
                        help="A YAML file that contains indices of the samples",
                        required=True)
    parser.add_argument("--global_step",
                        help="Global step of which to extract the checkpoint",
                        type=int)
    parser.add_argument("--batch_size",
                        help="Batch size to use for prediction",
                        default=4,
                        type=int)
    parser.add_argument("--scale_to_unit_interval",
                        default=False,
                        action="store_true")
    args = parser.parse_args()
    if args.scale_to_unit_interval:
        print("Scaling to unit interval")
    else:
        print("NOT scaling to unit interval")
    args.model_dir = pathlib.Path(args.model_dir)
    args.data_hdf5_fname = pathlib.Path(args.data_hdf5_fname)
    args.index_fname = pathlib.Path(args.index_fname)
    args.config_fname = args.model_dir.joinpath("config.yml")
    args.output_dir = args.model_dir.joinpath("output")
    args.output_dir.mkdir(exist_ok=True)
    assert args.model_dir.exists(), "Model directory does not exist!"
    assert args.data_hdf5_fname.exists(), "Data does not exist!"
    assert args.index_fname.exists(), "Index does not exist!"
    assert args.config_fname.exists(), "Config does not exist!"
    if args.global_step is None:
        global_steps = sorted(map(extract_global_step, args.model_dir.joinpath("checkpoints").glob("*.pt")))
        args.global_step = global_steps[-1]
    args.checkpoint_fname = args.model_dir.joinpath("checkpoints", f"checkpoint_{args.global_step}.pt")
    assert args.checkpoint_fname.exists(), "Checkpoint does not exist"
    with open(args.config_fname, "r") as config_f:
        config = yaml.load(config_f, Loader=yaml.FullLoader)
    if config['architecture'] == "bottleneck":
        generator_factory = GeneratorBottleneck
    elif config['architecture'] == "no_bottleneck":
        generator_factory = GeneratorNoBottleneck
    else:
        raise NotImplementedError()
    config['generator'].pop('use_ema')
    config['generator'].pop('ema_decay')
    config['generator'].pop('optimizer')
    num_blocks = binary_logarithm(config['max_size']) - binary_logarithm(config['min_size'])
    generator = generator_factory(dimensionality=config['dimensionality'],
                                  min_size=config['min_size'],
                                  num_blocks=num_blocks,
                                  **config['generator'])
    device = torch.device("cuda:0")
    ckpt = strip_distributed_markers(torch.load(args.checkpoint_fname, map_location=device))
    generator.load_state_dict(ckpt['generator'])
    generator = generator.to(device)
    data = Dataset(dimensionality=3,
                   data_fname=args.data_hdf5_fname,
                   index_fname=args.index_fname,
                   use_augmentations=False,
                   min_max_slice=config['min_max_slice'])
    data = DataLoader(data,
                      batch_size=args.batch_size,
                      shuffle=False,
                      pin_memory=True)
    with torch.no_grad():
        batch_example_ids = []
        for i, example in tqdm(enumerate(data)):
            example_id = example.pop('example_id')
            batch = example
            batch = {k: t.to(device, non_blocking=True) for k, t in batch.items()}
            batch = {k: [t.contiguous() for t in build_resolutions(t, dimensionality=3, n=num_blocks - 1)] for k, t in batch.items()}
            original = batch.pop('original')
            fake_gray = generator(**batch)
            brain_mask = batch["brain_mask"]
            original_min_max = [(original_t.min(), original_t.max()) for original_t in original]
            if args.scale_to_unit_interval:
                original = [torch.div(original_t - original_min, original_max - original_min) for
                            original_t, (original_min, original_max) in
                            zip(original, original_min_max)]
                fake_gray = [torch.clamp(torch.div(fake_gray_t - original_min, original_max - original_min), 0, 1) for
                             fake_gray_t, (original_min, original_max) in zip(fake_gray, original_min_max)]
            else:
                fake_gray = [torch.clamp(fake_gray_t, original_min, original_max) for fake_gray_t, (original_min, original_max) in zip(fake_gray, original_min_max)]
            scale_factors = [config['max_size'] // original_t.shape[-1] for original_t in original]
            resolutions = [original_t.shape[-1] for original_t in original]
            original = [upsampling(tensor=original_t,
                                   scale_factor=scale_factor) for original_t, scale_factor in
                        zip(original, scale_factors)]
            brain_mask = [torch.bernoulli(upsampling(tensor=brain_mask_t.type(torch.float32),
                                                     scale_factor=scale_factor)).type(torch.bool) for
                          brain_mask_t, scale_factor in zip(brain_mask, scale_factors)]
            fake_gray = [upsampling(tensor=fake_gray_t,
                                    scale_factor=scale_factor) for fake_gray_t, scale_factor in
                         zip(fake_gray, scale_factors)]
            fake_gray = GAN.starmap(GAN.copy_brain_into_tensor, zip(fake_gray, original, brain_mask))
            original = [original_arr.cpu().numpy() for original_arr in original]
            brain_mask = [brain_mask_arr.cpu().numpy() for brain_mask_arr in brain_mask]
            fake_gray = [fake_gray_arr.cpu().numpy() for fake_gray_arr in fake_gray]
            original = {f"original_{r}": v for r, v in zip(resolutions, original)}
            brain_mask = {f"brain_mask_{r}": v for r, v in zip(resolutions, brain_mask)}
            fake_gray = {f"fake_gray_{r}": v for r, v in zip(resolutions, fake_gray)}
            example_data = {**original, **brain_mask, **fake_gray}
            batch_example_ids.append(example_id)
            save_lzma_ndarray_archive(output_fname=args.output_dir.joinpath(f"batch-id-{i}.npy.tar.xz"), **example_data)
            with open(args.output_dir.joinpath("example-ids.yml"), "a") as f:
                f.write(",".join(example_id) + "\n")
if __name__ == "__main__":
    main()
