from argparse import ArgumentParser
import pathlib
from typing import List
import yaml
from cytoolz.dicttoolz import dissoc
from tqdm import tqdm
import nibabel as nib
import h5py
from joblib import delayed, Parallel
import numpy as np
from utils import read_lzma_ndarray_archive, get_real_stem
def store_nifti(batch_fname: pathlib.Path,
                example_ids: List[str],
                reference_dir: pathlib.Path,
                dl_dir: pathlib.Path,
                metadata,
                resolution: int,
                output_dir: pathlib.Path,
                alt_dir: pathlib.Path):
    batch_arr = read_lzma_ndarray_archive(batch_fname)
    assert all({v.shape[0] == len(example_ids) for v in batch_arr.values()})
    with h5py.File(dl_dir.joinpath(f"MRI_{resolution}.h5"), "r") as fhandle:
        for i, example_id in enumerate(example_ids):
            arr = {k: v[i, 0] for k, v in batch_arr.items()}
            original_alt = arr[f"original_{resolution}"]
            fake_gray = arr[f"fake_gray_{resolution}"]
            original_nifti = nib.load(reference_dir.joinpath(f"{example_id}.nii.gz"))
            original_alt = original_alt.astype(np.float64)
            fake_gray = fake_gray.astype(np.float64)
            original_alt = original_alt * metadata[example_id]['std_brain'] + metadata[example_id]['mu_brain']
            fake_gray = fake_gray * metadata[example_id]['std_brain'] + metadata[example_id]['mu_brain']
            original_alt = original_alt.astype(np.float32)
            fake_gray = fake_gray.astype(np.float32)
            convex_hull = fhandle['convex_hull'][example_id][()]
            fake_gray[~convex_hull] = 0
            original_nifti_alt = nib.Nifti1Image(original_alt, header=original_nifti.header, affine=original_nifti.affine)
            fake_gray_nifti = nib.Nifti1Image(fake_gray, header=original_nifti.header, affine=original_nifti.affine)
            nib.save(img=original_nifti_alt, filename=alt_dir.joinpath(f"{example_id}.nii.gz"))
            nib.save(img=fake_gray_nifti, filename=output_dir.joinpath(f"{example_id}.nii.gz"))
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="Where the LZMA compressed predictions are located",
                        required=True)
    parser.add_argument("--reference_dir",
                        help="A directory that contains NifTi files supplying header data",
                        required=True)
    parser.add_argument("--dl_dir",
                        help="The directory in which the HDF5 files can be found",
                        required=True)
    parser.add_argument("--metadata_fname",
                        help="Where the metadafile resides that contains the scale information",
                        required=True)
    parser.add_argument("--resolution",
                        default=128,
                        type=int)
    parser.add_argument("--num_jobs",
                        type=int,
                        default=10)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.reference_dir = pathlib.Path(args.reference_dir)
    args.dl_dir = pathlib.Path(args.dl_dir)
    args.metadata_fname = pathlib.Path(args.metadata_fname)
    args.input_dir.exists(), "Input directory does not exist!"
    args.reference_dir.exists(), "Reference directory does not exist!"
    args.dl_dir.exists(), "DL directory does not exist!"
    args.metadata_fname.exists(), "Metadata file does not exist!"
    args.output_dir = args.reference_dir.parent.joinpath(args.reference_dir.name + "__cp-gan__")
    args.output_dir.mkdir(exist_ok=True)
    args.alt_dir = args.reference_dir.parent.joinpath(args.reference_dir.name + "__alt__")
    args.alt_dir.mkdir(exist_ok=True)
    with open(args.metadata_fname) as f_metadata, open(args.input_dir.joinpath("example-ids.yml")) as f_example_ids:
        metadata = yaml.load(f_metadata, Loader=yaml.FullLoader)
        metadata = [_ for _ in metadata if _ is not None]
        metadata = {subject_dict['id']: dissoc(subject_dict, 'id') for subject_dict in metadata}
        batch_fnames = [batch_fname for batch_fname in args.input_dir.glob("*.npy.tar.xz")]
        def _extract_batch_id(batch_fname):
            return int(get_real_stem(batch_fname).split("-")[-1])
        batch_fnames.sort(key=_extract_batch_id)
        batch_fnames = [(batch_fname, line.strip().split(",")) for batch_fname, line in zip(batch_fnames, f_example_ids)]
        Parallel(n_jobs=args.num_jobs)(delayed(store_nifti)(batch_fname=batch_fname,
                                                            example_ids=example_ids,
                                                            reference_dir=args.reference_dir,
                                                            dl_dir=args.dl_dir,
                                                            metadata=metadata,
                                                            resolution=args.resolution,
                                                            output_dir=args.output_dir,
                                                            alt_dir=args.alt_dir) for batch_fname, example_ids in tqdm(batch_fnames))
if __name__ == "__main__":
    main()
