import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
def custom_forward(module: nn.Module):
    def f(*args, **kwargs):
        return module(*args, **kwargs)
    return f
class DownsamplingLinear(nn.Module):
    def __init__(self, dimensionality: int):
        super().__init__()
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.op = self.bilinear_downsampling
        elif self.dimensionality == 3:
            self.op = self.trilinear_downsampling
        else:
            raise NotImplementedError()
    def bilinear_downsampling(self, x):
        return F.interpolate(input=x,
                             scale_factor=1/2,
                             align_corners=True,
                             mode="bilinear")
    def trilinear_downsampling(self, x):
        return F.interpolate(input=x,
                             scale_factor=1/2,
                             align_corners=True,
                             mode="trilinear")
    def forward(self, x):
        return self.op(x)
class UpsamplingLinear(nn.Module):
    def __init__(self, dimensionality: int):
        super().__init__()
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.op = self.bilinear_upsampling
        elif self.dimensionality == 3:
            self.op = self.trilinear_upsampling
        else:
            raise NotImplementedError()
    def bilinear_upsampling(self, x):
        return F.interpolate(input=x,
                             scale_factor=2,
                             align_corners=True,
                             mode="bilinear")
    def trilinear_upsampling(self, x):
        return F.interpolate(input=x,
                             scale_factor=2,
                             align_corners=True,
                             mode="trilinear")
    def forward(self, x):
        return self.op(x)
class DownsamplingBlock(nn.Module):
    def __init__(self, dimensionality: int, *args, **kwargs):
        super().__init__()
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.op = DownsamplingBlock2d(*args, **kwargs)
        elif self.dimensionality == 3:
            self.op = DownsamplingBlock3d(*args, **kwargs)
        else:
            raise NotImplementedError()
    def forward(self, x):
        return self.op(x)
class DownsamplingBlock2d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
    def forward(self, x):
        dtype = x.dtype
        if dtype == torch.bool:
            x = x.type(torch.float32)
        x = x \
            .unfold(2, 2, 2) \
            .unfold(3, 2, 2)
        x = x.mean(dim=[-1, -2])
        if dtype == torch.bool:
            x = x >= .5
        return x
class DownsamplingBlock3d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
    def forward(self, x):
        dtype = x.dtype
        if dtype == torch.bool:
            x = x.type(torch.float32)
        x = x \
            .unfold(2, 2, 2) \
            .unfold(3, 2, 2) \
            .unfold(4, 2, 2)
        x = x.mean(dim=[-1, -2, -3])
        if dtype == torch.bool:
            x = x >= .5
        return x
class UpsamplingBlock(nn.Module):
    def __init__(self, dimensionality: int, *args, **kwargs):
        super().__init__()
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.op = UpsamplingBlock2d(*args, **kwargs)
        elif self.dimensionality == 3:
            self.op = UpsamplingBlock3d(*args, **kwargs)
        else:
            raise NotImplementedError()
    def forward(self, x):
        return self.op(x)
class UpsamplingBlock2d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
    def forward(self, x):
        s = x.size()
        x = x.view(s[0], s[1],
                   s[2], 1,
                   s[3], 1)
        x = x.expand(-1, -1,
                     -1, 2,
                     -1, 2)
        x = x.contiguous().view(s[0], s[1], s[2] * 2, s[3] * 2)
        return x
class UpsamplingBlock3d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
    def forward(self, x):
        s = x.size()
        x = x.view(s[0], s[1],
                   s[2], 1,
                   s[3], 1,
                   s[4], 1)
        x = x.expand(-1, -1,
                     -1, 2,
                     -1, 2,
                     -1, 2)
        x = x.contiguous().view(s[0], s[1], s[2] * 2, s[3] * 2, s[4] * 2)
        return x
class PixelNorm(nn.Module):
    def __init__(self, num_features, eps=1E-12, *args, **kwargs):
        super().__init__()
        assert num_features > 1, "Too few dimensions to perform PixelNorm!"
        self.num_features = num_features
        self.eps = eps
    def forward(self, x):
        l2_norm = torch.rsqrt(torch.sum(x * x, dim=1, keepdim=True) + self.eps)    
        return x * l2_norm
class MinibatchDiscriminationLayer(nn.Module):
    def __init__(self):
        super().__init__()
    def forward(self, x, eps=1e-12):
       mu = x.mean(dim=0, keepdim=True)     
       std = torch.sqrt(torch.mean(torch.mul(x - mu, x - mu), dim=0, keepdim=True) + eps)      
       global_std = torch.mean(std).view(*(1,) * x.ndim)   
       global_std = global_std.repeat(x.shape[0], 1, *x.shape[2:])  
       y = torch.cat([x, global_std], dim=1)
       return y
NORM_FACTORY = PixelNorm
class EqualizedConv(nn.Module):
    def __init__(self, dimensionality: int, *args, **kwargs):
        super().__init__()
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.op = EqualizedConv2d(*args, **kwargs)
        elif self.dimensionality == 3:
            self.op = EqualizedConv3d(*args, **kwargs)
        else:
            raise NotImplementedError()
    def forward(self, x):
        return self.op(x)
class EqualizedConv2d(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 use_spectral_norm: bool,
                 use_equalized_convolutions: bool,
                 normalized: bool,
                 padding: int = 0,
                 groups:int = 1):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.groups = groups
        self.use_spectral_norm = use_spectral_norm
        self.use_equalized_convolutions = use_equalized_convolutions
        self.normalized = normalized
        self.weight = nn.Parameter(data=torch.randn([self.out_channels, self.in_channels // self.groups,
                                                     self.kernel_size, self.kernel_size]),
                                   requires_grad=True)
        self.bias = nn.Parameter(data=torch.zeros(self.out_channels),
                                 requires_grad=True)
        if self.use_equalized_convolutions:
            fan_in = np.prod(self.weight.shape[1:])
            self.weight_factor = np.sqrt(2)/np.sqrt(fan_in)
        else:
            self.weight_factor = 1
        if self.use_spectral_norm:
            raise NotImplementedError()
        self.maybe_norm = NORM_FACTORY(num_features=self.in_channels, affine=False) if self.normalized else nn.Identity()
    def forward(self, x):
        return F.conv2d(input=self.maybe_norm(x),
                        weight=self.weight_factor * self.weight,
                        bias=self.bias,
                        padding=self.padding,
                        groups=self.groups)
class EqualizedConv3d(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 use_spectral_norm: bool,
                 use_equalized_convolutions: bool,
                 normalized: bool,
                 padding: int = 0,
                 groups:int = 1):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.groups = groups
        self.use_spectral_norm = use_spectral_norm
        self.use_equalized_convolutions = use_equalized_convolutions
        self.normalized = normalized
        self.weight = nn.Parameter(data=torch.randn([self.out_channels, self.in_channels // self.groups,
                                                     self.kernel_size, self.kernel_size, self.kernel_size]),
                                   requires_grad=True)
        self.bias = nn.Parameter(data=torch.zeros(self.out_channels),
                                 requires_grad=True)
        if self.use_equalized_convolutions:
            fan_in = np.prod(self.weight.shape[1:])
            self.weight_factor = np.sqrt(2)/np.sqrt(fan_in)
        else:
            self.weight_factor = 1
        if self.use_spectral_norm:
            raise NotImplementedError()
        self.maybe_norm = NORM_FACTORY(num_features=self.in_channels, affine=False) if self.normalized else nn.Identity()
    def forward(self, x):
        return F.conv3d(input=self.maybe_norm(x),
                        weight=self.weight_factor * self.weight,
                        bias=self.bias,
                        padding=self.padding,
                        groups=self.groups)
class EqualizedLinear(nn.Module):
    def __init__(self,
                 in_features: int,
                 out_features: int,
                 use_spectral_norm: bool,
                 use_equalized_convolutions: bool,
                 normalized: bool):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.use_spectral_norm = use_spectral_norm
        self.use_equalized_convolutions = use_equalized_convolutions
        self.normalized = normalized
        self.weight = nn.Parameter(data=torch.randn(self.out_features, self.in_features),
                                   requires_grad=True)
        self.bias = nn.Parameter(data=torch.zeros(self.out_features),
                                 requires_grad=True)
        if self.use_equalized_convolutions:
            fan_in = self.in_features
            self.weight_factor = np.sqrt(2) / np.sqrt(fan_in)
        else:
            self.weight_factor = 1
        if self.use_spectral_norm:
            raise NotImplementedError()
        self.maybe_norm = nn.LayerNorm(normalized_shape=self.in_features, elementwise_affine=False) if self.normalized else nn.Identity()
    def forward(self, x):
        return F.linear(input=self.maybe_norm(x),
                        weight=self.weight_factor * self.weight,
                        bias=self.bias)
class MobileNetConv(nn.Module):  
    def __init__(self, dimensionality: int, *args, **kwargs):  
        super().__init__()  
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.op = MobileNetConv2d(*args, **kwargs)
        elif self.dimensionality == 3:
            self.op = MobileNetConv3d(*args, **kwargs)
        else:
            raise NotImplementedError()
    def forward(self, x):
        return self.op(x)
class MobileNetConv2d(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 use_spectral_norm: bool,
                 use_equalized_convolutions: bool,
                 padding: int = 0):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.use_spectral_norm = use_spectral_norm
        self.use_equalized_convolutions = use_equalized_convolutions
        self.depthwise_layer = EqualizedConv2d(in_channels=self.in_channels,
                                               out_channels=self.in_channels,
                                               kernel_size=self.kernel_size,
                                               padding=self.padding,
                                               groups=self.in_channels,
                                               use_spectral_norm=self.use_spectral_norm,
                                               use_equalized_convolutions=self.use_equalized_convolutions,
                                               normalized=True)
        self.pointwise_layer = EqualizedConv2d(in_channels=self.in_channels,
                                               out_channels=self.out_channels,
                                               kernel_size=1,
                                               use_spectral_norm=self.use_spectral_norm,
                                               use_equalized_convolutions=self.use_equalized_convolutions,
                                               normalized=False)
    def forward(self, x):
        x = self.depthwise_layer(x)
        x = self.pointwise_layer(x)
        return x
class MobileNetConv3d(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 use_spectral_norm: bool,
                 use_equalized_convolutions: bool,
                 padding: int = 0):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.use_spectral_norm = use_spectral_norm
        self.use_equalized_convolutions = use_equalized_convolutions
        self.depthwise_layer = EqualizedConv3d(in_channels=self.in_channels,
                                               out_channels=self.in_channels,
                                               kernel_size=self.kernel_size,
                                               padding=self.padding,
                                               groups=self.in_channels,
                                               use_spectral_norm=self.use_spectral_norm,
                                               use_equalized_convolutions=self.use_equalized_convolutions,
                                               normalized=True)
        self.pointwise_layer = EqualizedConv3d(in_channels=self.in_channels,
                                               out_channels=self.out_channels,
                                               kernel_size=1,
                                               use_spectral_norm=self.use_spectral_norm,
                                               use_equalized_convolutions=self.use_equalized_convolutions,
                                               normalized=False)
    def forward(self, x):
        x = self.depthwise_layer(x)
        x = self.pointwise_layer(x)
        return x
""" Taken from https://github.com/akanimax/BMSG-GAN/blob/master/sourcecode/MSG_GAN/CustomLayers.py """
def update_average(model_tgt, model_src, beta):
    """
    update the model_target using exponential moving averages
    :param model_tgt: target model
    :param model_src: source model
    :param beta: value of decay beta
    :return: None (updates the target model)
    """
    def toggle_grad(model, requires_grad):
        for p in model.parameters():
            p.requires_grad_(requires_grad)
    toggle_grad(model_tgt, False)
    toggle_grad(model_src, False)
    param_dict_src = dict(model_src.named_parameters())
    for p_name, p_tgt in model_tgt.named_parameters():
        p_src = param_dict_src[p_name]
        assert (p_src is not p_tgt)
        p_tgt.copy_(beta * p_tgt + (1. - beta) * p_src)
    toggle_grad(model_tgt, True)
    toggle_grad(model_src, True)
