import pathlib
from typing import Tuple
import numpy as np
import h5py
import torch
import torch.nn as nn
import albumentations as A
import cv2
from torch.utils.data import Dataset as AbstractDataset
class Augmentor:
    def __init__(self, dimensionality: int):
        self.dimensionality = dimensionality
        if self.dimensionality == 2:
            self.transform = A.Compose(
                [
                    A.IAAAdditiveGaussianNoise(scale=(0.0, 0.2), p=1.0),
                    A.ShiftScaleRotate(rotate_limit=(-45, +45),
                                       scale_limit=(-.05, +.05),
                                       shift_limit=(-.025, +.025),
                                       border_mode=cv2.BORDER_WRAP, p=1.0)
                ], additional_targets={'conditional': 'image',
                                       'brain': 'image',
                                       'brain_mask': 'mask'}
            )
        elif self.dimensionality == 3:
            self.transform = None
        else:
            raise NotImplementedError()
    def __call__(self,
                 example_id: int,
                 conditional: np.ndarray,
                 original: np.ndarray,
                 brain: np.ndarray,
                 brain_mask: np.ndarray,
                 slice_index: np.ndarray):
        if self.dimensionality == 2:
            batch = self.transform(image=original,
                                   conditional=conditional,
                                   brain=brain,
                                   brain_mask=brain_mask)
            original = batch['image'].copy()
            conditional = batch['conditional'].copy()
            brain = batch['brain'].copy()
            brain_mask = batch['brain_mask'].copy()
        elif self.dimensionality == 3:
            raise NotImplementedError()
        else:
            raise NotImplementedError()
        original_t = torch.from_numpy(original).unsqueeze(dim=0)
        conditional_t = torch.from_numpy(conditional).unsqueeze(dim=0)
        brain_t = torch.from_numpy(brain).unsqueeze(dim=0)
        brain_mask_t = torch.from_numpy(brain_mask).unsqueeze(dim=0)
        slice_index_t = torch.from_numpy(slice_index).unsqueeze(dim=0)
        return {
            "example_id": example_id,
            "conditional": conditional_t,
            "original": original_t,
            "brain": brain_t,
            "brain_mask": brain_mask_t,
            "slice_index": slice_index_t
        }
class Dataset(AbstractDataset):
    def __init__(self,
                 dimensionality: int,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 use_augmentations: bool,
                 min_max_slice: Tuple[int, int]):
        super().__init__()
        self.dimensionality = dimensionality
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.use_augmentations = use_augmentations
        self.min_max_slice = min_max_slice
        self.ix = [line.strip() for line in open(self.index_fname, "r")]
        self.augmentor = Augmentor(dimensionality=self.dimensionality)
    def __len__(self):
        return len(self.ix)
    def __getitem__(self, i):
        with h5py.File(self.data_fname, "r") as fhandle:
            example_id = self.ix[i]
            original = fhandle["original"][example_id][()]
            convex_hull = fhandle["convex_hull"][example_id][()]
            brain_mask = fhandle["brain_mask"][example_id][()]
            k = 0
            if self.dimensionality == 2:
                low = 0 if self.min_max_slice[0] is None else self.min_max_slice[0]
                high = original.shape[-1] if self.min_max_slice[1] is None else self.min_max_slice[1]
                k = np.random.randint(low=low, high=high)
                original = original[..., k]
                convex_hull = convex_hull[..., k]
                brain_mask = brain_mask[..., k]
                k = (k - low)/(high - low - 1)
            conditional = convex_hull.astype(np.float32)
            brain = np.where(brain_mask, original, np.full_like(original, fill_value=np.min(original)))
            brain_mask = brain_mask.astype(np.float32)
            slice_index = np.full_like(original, fill_value=k)
            data_dict = {
                "example_id": example_id,
                "conditional": conditional,
                "original": original,
                "brain": brain,
                "brain_mask": brain_mask,
                "slice_index": slice_index
            }
            if self.use_augmentations:
                return self.augmentor(**data_dict)
            else:
                return {k: torch.from_numpy(v).unsqueeze(dim=0) if k != "example_id" else v for k, v in data_dict.items()}
