from typing import List
import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint
from remodeling.generative.CPGAN.custom_layers import DownsamplingBlock as Downsampling, UpsamplingBlock as Upsampling,\
    MobileNetConv, EqualizedConv, EqualizedLinear, custom_forward
from remodeling.generative.CPGAN.activation_functions import Activation
def binary_logarithm(n):
    assert n & (n - 1) == 0, "Input needs to be a power of two!"
    bitstring = bin(n)[2:]
    return len(bitstring) - 1
class Generator(nn.Module):
    def __init__(self,
                 dimensionality: int,
                 use_skips: bool,
                 use_checkpointing: bool,
                 use_trainable_constants: bool,
                 min_size: int,
                 num_blocks: int,
                 num_channels_per_block: List[int],
                 num_layers_per_block: int,
                 kernel_size: int,
                 activation_type: str,
                 use_equalized_convolutions: bool):
        super().__init__()
        assert len(num_channels_per_block) == num_blocks, "Inconsistent channel specification!"
        self.dimensionality = dimensionality
        self.use_skips = use_skips
        self.use_checkpointing = use_checkpointing
        self.use_trainable_constants = use_trainable_constants
        self.min_size = min_size
        self.num_blocks = num_blocks
        self.num_channels_per_block = num_channels_per_block
        self.num_layers_per_block = num_layers_per_block
        self.kernel_size = kernel_size
        self.activation_type = activation_type
        self.use_equalized_convolutions = use_equalized_convolutions
        self.input_size = 2**self.num_blocks * min_size
        self.generator_blocks = nn.ModuleList(Generator.GeneratorBlock(
            dimensionality=self.dimensionality,
            input_resolution=2 ** (binary_logarithm(self.min_size) + i + 1),
            use_skips=self.use_skips,
            use_trainable_constants=self.use_trainable_constants,
            num_layers=self.num_layers_per_block,
            num_channels=self.num_channels_per_block[i],
            kernel_size=self.kernel_size,
            activation_type=self.activation_type,
            use_equalized_convolutions=self.use_equalized_convolutions
        ) for i in range(self.num_blocks))
        if self.use_trainable_constants:
            init_shape = (1, 1) + (self.min_size,) * self.dimensionality
            self.trainable_constant = nn.Parameter(data=torch.randn(*init_shape, dtype=torch.float32),
                                                   requires_grad=True)
    def forward(self,
                conditional: List[torch.FloatTensor],
                brain: List[torch.FloatTensor],
                brain_mask: List[torch.FloatTensor],
                slice_index: List[torch.FloatTensor]) -> List[torch.FloatTensor]:
        batch_size = conditional[0].shape[0]
        device = conditional[0].device
        init_shape = (batch_size, 1) + (self.min_size,) * self.dimensionality
        if self.use_trainable_constants:
            x = self.trainable_constant.expand(*init_shape)
        else:
            x = torch.randn(*init_shape, dtype=torch.float32, device=device)
        output = []
        for conditional_t, brain_t, brain_mask_t, slice_index_t, generator_block in zip(conditional,
                                                                                        brain,
                                                                                        brain_mask,
                                                                                        slice_index,
                                                                                        self.generator_blocks):
            input_t = torch.cat([conditional_t, brain_t, brain_mask_t, slice_index_t], dim=1)
            x = checkpoint(custom_forward(generator_block), x, input_t) if self.use_checkpointing else generator_block(x=x, input_t=input_t)
            output.append(x)
        return output
    class GeneratorBlock(nn.Module):
        def __init__(self,
                     dimensionality: int,
                     input_resolution: int,
                     use_skips: bool,
                     use_trainable_constants: bool,
                     num_layers: int,
                     num_channels: int,
                     kernel_size: int,
                     activation_type: str,
                     use_equalized_convolutions: bool):
            super().__init__()
            assert kernel_size % 2 == 1, "Kernel size needs to be odd!"
            self.dimensionality = dimensionality
            self.input_resolution = input_resolution
            self.use_skips = use_skips
            self.use_trainable_constants = use_trainable_constants
            self.num_layers = num_layers
            self.num_channels = num_channels
            self.kernel_size = kernel_size
            self.activation_type = activation_type
            self.use_equalized_convolutions = use_equalized_convolutions
            self.activation = Activation(self.activation_type)
            self.upsampling = Upsampling(dimensionality=self.dimensionality)
            self.init_cnn = EqualizedConv(dimensionality=self.dimensionality,
                                          in_channels=1 + 2 * 4,
                                          out_channels=self.num_channels,
                                          kernel_size=1,
                                          use_spectral_norm=False,
                                          use_equalized_convolutions=self.use_equalized_convolutions,
                                          normalized=True)
            self.cnn_layers = nn.ModuleList(MobileNetConv(dimensionality=self.dimensionality,
                                                          in_channels=self.num_channels,
                                                          out_channels=self.num_channels,
                                                          kernel_size=self.kernel_size,
                                                          padding=(self.kernel_size - 1) // 2,
                                                          use_spectral_norm=False,
                                                          use_equalized_convolutions=self.use_equalized_convolutions)
                                            for _ in range(self.num_layers))
            self.post_cnn = EqualizedConv(dimensionality=self.dimensionality,
                                          in_channels=self.num_channels,
                                          out_channels=1,
                                          kernel_size=1,
                                          use_spectral_norm=False,
                                          use_equalized_convolutions=self.use_equalized_convolutions,
                                          normalized=False)
            if self.use_trainable_constants:
                init_shape = (1, 4) + (self.input_resolution,) * self.dimensionality
                self.trainable_constant = nn.Parameter(data=torch.randn(*init_shape),
                                                       requires_grad=True)
        def forward(self,
                    x: torch.FloatTensor,
                    input_t: torch.FloatTensor) -> torch.FloatTensor:
            init_shape = input_t.shape
            if self.use_trainable_constants:
                aux = self.trainable_constant.expand(*init_shape)
            else:
                aux = torch.randn(*init_shape, dtype=torch.float32, device=x.device)
            z = self.upsampling(x)
            x = torch.cat([z, input_t, aux], dim=1)
            x = self.init_cnn(x)
            for cnn_layer in self.cnn_layers:
                x = self.activation(cnn_layer(x))
            x = self.post_cnn(x)
            if self.use_skips:
                x = x + z
            return x
class Discriminator(nn.Module):
    def __init__(self,
                 dimensionality: int,
                 use_skips: bool,
                 use_checkpointing: bool,
                 use_trainable_constants: bool,
                 min_size: int,
                 num_blocks: int,
                 num_channels_per_block: List[int],
                 num_layers_per_block: int,
                 kernel_size: int,
                 activation_type: str,
                 use_spectral_norm: bool,
                 use_equalized_convolutions: bool):
        super().__init__()
        assert len(num_channels_per_block) == num_blocks, "Inconsistent channel specification!"
        self.dimensionality = dimensionality
        self.use_skips = use_skips
        self.use_checkpointing = use_checkpointing
        self.use_trainable_constants = use_trainable_constants
        self.min_size = min_size
        self.num_blocks = num_blocks
        self.num_channels_per_block = num_channels_per_block
        self.num_layers_per_block = num_layers_per_block
        self.kernel_size = kernel_size
        self.activation_type = activation_type
        self.use_spectral_norm = use_spectral_norm
        self.use_equalized_convolutions = use_equalized_convolutions
        self.input_size = 2**self.num_blocks * min_size
        self.discriminator_blocks = nn.ModuleList(Discriminator.DiscriminatorBlock(
            dimensionality=self.dimensionality,
            use_skips=use_skips,
            is_first_block=(i == 0),
            num_layers=self.num_layers_per_block,
            num_channels=self.num_channels_per_block[i],
            kernel_size=self.kernel_size,
            activation_type=self.activation_type,
            use_spectral_norm=self.use_spectral_norm,
            use_equalized_convolutions=self.use_equalized_convolutions,
        ) for i in range(self.num_blocks))
        self.fc = EqualizedLinear(in_features=(2 * self.min_size)**self.dimensionality,
                                  out_features=1,
                                  use_spectral_norm=self.use_spectral_norm,
                                  use_equalized_convolutions=self.use_equalized_convolutions,
                                  normalized=True)
    def forward(self,
                probe: List[torch.FloatTensor],
                conditional: List[torch.FloatTensor],
                brain: List[torch.FloatTensor],
                brain_mask: List[torch.FloatTensor],
                slice_index: List[torch.FloatTensor]) -> List[torch.FloatTensor]:
        x = None
        for probe_t, conditional_t, brain_t, brain_mask_t, slice_index_t, discriminator_block in zip(reversed(probe),
                                                                                                     reversed(conditional),
                                                                                                     reversed(brain),
                                                                                                     reversed(brain_mask),
                                                                                                     reversed(slice_index),
                                                                                                     self.discriminator_blocks):
            input_t = torch.cat([conditional_t, brain_t, brain_mask_t, slice_index_t], dim=1)
            x = checkpoint(custom_forward(discriminator_block), x, probe_t, input_t) if self.use_checkpointing else discriminator_block(x=x, probe=probe_t, input_t=input_t)
        x = x.flatten(start_dim=1)
        x = self.fc(x)
        x = x.squeeze(dim=1)
        return x
    class DiscriminatorBlock(nn.Module):
        def __init__(self,
                     dimensionality: int,
                     use_skips: bool,
                     is_first_block: bool,
                     num_layers: int,
                     num_channels: int,
                     kernel_size: int,
                     activation_type: str,
                     use_spectral_norm: bool,
                     use_equalized_convolutions: bool):
            super().__init__()
            assert kernel_size % 2 == 1, "Kernel size needs to be odd!"
            self.dimensionality = dimensionality
            self.use_skips = use_skips
            self.is_first_block = is_first_block
            self.num_layers = num_layers
            self.num_channels = num_channels
            self.kernel_size = kernel_size
            self.activation_type = activation_type
            self.use_spectral_norm = use_spectral_norm
            self.use_equalized_convolutions = use_equalized_convolutions
            self.activation = Activation(self.activation_type)
            self.downsampling = Downsampling(dimensionality=self.dimensionality)
            self.init_cnn = EqualizedConv(dimensionality=self.dimensionality,
                                          in_channels=5 if self.is_first_block else 6,
                                          out_channels=self.num_channels,
                                          kernel_size=1,
                                          use_spectral_norm=self.use_spectral_norm,
                                          use_equalized_convolutions=self.use_equalized_convolutions,
                                          normalized=True)
            self.cnn_layers = nn.ModuleList(MobileNetConv(
                dimensionality=self.dimensionality,
                in_channels=self.num_channels,
                out_channels=self.num_channels,
                kernel_size=self.kernel_size,
                padding=(self.kernel_size - 1) // 2,
                use_spectral_norm=self.use_spectral_norm,
                use_equalized_convolutions=self.use_equalized_convolutions) for _ in
                                            range(self.num_layers))
            self.post_cnn = EqualizedConv(dimensionality=self.dimensionality,
                                          in_channels=self.num_channels,
                                          out_channels=1,
                                          kernel_size=1,
                                          use_spectral_norm=self.use_spectral_norm,
                                          use_equalized_convolutions=self.use_equalized_convolutions,
                                          normalized=True)
        def forward(self,
                    x: torch.FloatTensor,
                    probe: torch.FloatTensor,
                    input_t: torch.FloatTensor) -> torch.FloatTensor:
            input_t = torch.cat([probe, input_t], dim=1)
            if not self.is_first_block:
                z = self.downsampling(x)
                input_t = torch.cat([z, input_t], dim=1)
            x = self.init_cnn(input_t)
            for cnn_layer in self.cnn_layers:
                x = self.activation(cnn_layer(x))
            x = self.post_cnn(x)
            if not self.is_first_block and self.use_skips:
                x = x + z
            return x
