from typing import List, Dict, Callable, Iterable, Optional
import pathlib
import itertools
import copy
from remodeling.generative.CPGAN.data import Dataset
from remodeling.generative.CPGAN.network_bottleneck import Generator as GeneratorBottleneck, Discriminator as DiscriminatorBottleneck
from remodeling.generative.CPGAN.network_no_bottleneck import Generator as GeneratorNoBottleneck, Discriminator as DiscriminatorNoBottleneck, binary_logarithm
from remodeling.generative.CPGAN.custom_layers import DownsamplingBlock as Downsampling, update_average
from cytoolz.dicttoolz import dissoc
from moviepy.editor import ImageSequenceClip
import math
import numpy as np
import torch
import torch.nn.functional as F
from torch.nn.parallel import DistributedDataParallel
import torchvision
import importlib
import inspect
OPTIMIZERS = {
    "adam": ('torch.optim', 'Adam'),
    "adamp": ('adamp', 'AdamP'),
    "madam": ('remodeling.generative.CPGAN.optimizer', 'Madam')
}
from skimage.io import imsave
def maybe_allreduce_mean(tensor: torch.FloatTensor, distributed: bool) -> torch.FloatTensor:
    if distributed and torch.distributed.is_initialized():
        batch_size = tensor.shape[0]
        tensor_sum = tensor.clone().sum()
        num_el = torch.full([], fill_value=batch_size, device=tensor.device, dtype=torch.float32)
        torch.distributed.all_reduce_multigpu([tensor_sum], op=torch.distributed.ReduceOp.SUM)
        torch.distributed.all_reduce_multigpu([num_el], op=torch.distributed.ReduceOp.SUM)
        tensor_mean = tensor_sum / num_el
        return tensor_mean
    else:
        return torch.mean(tensor)
def ls_gan_loss(x: torch.FloatTensor,
                y: Optional[torch.FloatTensor],
                distributed: bool = True):
    if y is None:
        return (x - 1) ** 2 / 2
    else:
        return ((x - 1) ** 2 + (y + 1) ** 2) / 4
def ra_ls_gan_loss(x: torch.FloatTensor,
                   y: torch.FloatTensor,
                   distributed: bool = True):
    xy_diff = x - y
    yx_diff = y - x
    return ((xy_diff - 1) ** 2 + (yx_diff + 1) ** 2) / 2
def ra_hinge_gan_loss(x: torch.FloatTensor,
                      y: torch.FloatTensor,
                      distributed: bool = True):
    xy_diff = x - y
    yx_diff = y - x
    return (F.relu(1 - xy_diff) + F.relu(1 + yx_diff)) / 2
INPUT_SIZE = 128
def strip_distributed_markers(ckpt):
    from collections import OrderedDict
    module_stripped_ckpt = dict()
    for module, weight_dict in ckpt.items():
        module_stripped_dict = OrderedDict()
        for weight_name in weight_dict:
            module_stripped_dict[weight_name.replace('module.', '')] = weight_dict[weight_name]
        module_stripped_ckpt[module] = module_stripped_dict
    return module_stripped_ckpt
def build_resolutions(t: torch.tensor,
                      dimensionality: int,
                      n: int) -> List[torch.tensor]:
    ts = [t]
    downsampling = Downsampling(dimensionality=dimensionality)
    for _ in range(n):
        ts.append(downsampling(ts[-1]))
    return ts[::-1]
class GAN:
    def __init__(self,
                 mode: str,
                 model_dir: pathlib.Path,
                 data_fname: pathlib.Path,
                 index_fname: pathlib.Path,
                 config: Dict,
                 local_rank: int,
                 world_size: int,
                 device: torch.device,
                 secondary_device: torch.device):
        self.mode = mode
        self.model_dir = model_dir
        self.data_fname = data_fname
        self.index_fname = index_fname
        self.config = config
        self.local_rank = local_rank
        self.world_size = world_size
        self.device = device
        self.secondary_device = secondary_device
        self.architecture = config["architecture"]
        self.use_ema = config["generator"].pop("use_ema")
        self.ema_decay = config["generator"].pop("ema_decay")
        self.use_augmentations = config['use_augmentations']
        self.min_max_slice = config.get("min_max_slice", (None, None))
        self.generator_config, self.discriminator_config = config["generator"], config["discriminator"]
        self.dimensionality = config['dimensionality']
        self.min_size = config['min_size']
        self.batch_size = self.config['batch_size']
        self.num_epochs = self.config['num_epochs']
        self.wgan_gp_alpha = self.config['wgan_gp_alpha']
        self.video_frequency = self.config['video_frequency']
        self.checkpoint_frequency = self.config['checkpoint_frequency']
        assert self.min_size & (self.min_size - 1) == 0, "Min size is not a power of two!"
        self.data, self.sampler = self.__init_data__()
        if self.architecture == "bottleneck":
            generator_factory, discriminator_factory = GeneratorBottleneck, DiscriminatorBottleneck
        elif self.architecture == "no_bottleneck":
            generator_factory, discriminator_factory = GeneratorNoBottleneck, DiscriminatorNoBottleneck
        else:
            raise NotImplementedError()
        self.num_blocks = binary_logarithm(INPUT_SIZE) - binary_logarithm(self.min_size)
        self.generator = generator_factory(dimensionality=self.dimensionality,
                                           min_size=self.min_size,
                                           num_blocks=self.num_blocks,
                                           **dissoc(self.generator_config, 'optimizer'))
        self.discriminator = discriminator_factory(dimensionality=self.dimensionality,
                                                   min_size=self.min_size,
                                                   num_blocks=self.num_blocks,
                                                   **dissoc(self.discriminator_config, 'optimizer'))
        self.generator = self.generator.to(self.device)
        self.discriminator = self.discriminator.to(self.secondary_device)
        self.loss_fct = config['loss_type']
        if self.local_rank == 0:
            print("---")
            print("Architecture..")
            for i, (generator_block, discriminator_block) in enumerate(zip(self.generator.generator_blocks,
                                                                           reversed(self.discriminator.discriminator_blocks))):
                resolution = 2 ** (self.min_size + i - 1)
                g_num_channels = generator_block.out_channels if hasattr(generator_block, 'out_channels') else generator_block.num_channels
                d_num_channels = discriminator_block.out_channels if hasattr(discriminator_block, 'out_channels') else discriminator_block.num_channels
                print('{:>12}:  {:>12}  {:>12}'.format(resolution, g_num_channels, d_num_channels))
            print("---")
        if self.world_size > 1:
            self.generator = DistributedDataParallel(self.generator,
                                                     device_ids=[self.device],
                                                     output_device=self.device)
            self.discriminator = DistributedDataParallel(self.discriminator,
                                                         device_ids=[self.secondary_device],
                                                         output_device=self.secondary_device)
        if self.use_ema:
            self.generator_shadow = copy.deepcopy(self.generator)
            update_average(model_tgt=self.generator_shadow, model_src=self.generator, beta=0.)
        self.generator_opt, self.discriminator_opt = self.__init_optimizers__()
    def __init_optimizers__(self):
        def _instantiate_optimizer(optim_type, optim_args, network_params):
            optim_module, optim_name = OPTIMIZERS[optim_type]
            optim_factory = getattr(importlib.import_module(optim_module), optim_name)
            optim_supported_args = inspect.getfullargspec(optim_factory).args[2:]
            optim_args = {k: v for k, v in optim_args.items() if k in optim_supported_args}
            return optim_factory(filter(lambda p: p.requires_grad, network_params), **optim_args)
        self.generator_opt = _instantiate_optimizer(optim_type=self.generator_config['optimizer'].pop('type'),
                                                    optim_args=self.generator_config['optimizer'],
                                                    network_params=self.generator.parameters())
        self.discriminator_opt = _instantiate_optimizer(optim_type=self.discriminator_config['optimizer'].pop('type'),
                                                        optim_args=self.discriminator_config['optimizer'],
                                                        network_params=self.discriminator.parameters())
        return self.generator_opt, self.discriminator_opt
    def __init_data__(self):
        data = Dataset(data_fname=self.data_fname,
                       index_fname=self.index_fname,
                       dimensionality=self.dimensionality,
                       use_augmentations=self.use_augmentations,
                       min_max_slice=self.min_max_slice)
        sampler = torch.utils.data.distributed.DistributedSampler(
            data,
            shuffle=True,
            num_replicas=self.world_size,
            rank=self.local_rank)
        data = torch.utils.data.DataLoader(data,
                                           batch_size=self.config['batch_size'] if self.mode == "train" else 1,
                                           sampler=sampler,
                                           num_workers=8,
                                           pin_memory=True)
        return data, sampler
    def restore_checkpoint(self) -> int:
        import operator
        def _extract_global_step(path: pathlib.Path):
            return path, int(path.stem.split("_")[1])
        ckpts = list(map(_extract_global_step, self.model_dir.joinpath("checkpoints").glob("checkpoint_*.pt")))
        ckpts = sorted(ckpts, key=operator.itemgetter(1))
        newest_ckpt_path, global_step = ckpts[-1]
        ckpt = torch.load(newest_ckpt_path, map_location=self.device)
        if self.world_size == 1:
            ckpt = strip_distributed_markers(ckpt)
        self.generator.load_state_dict(ckpt["generator"])
        self.discriminator.load_state_dict(ckpt["discriminator"])
        self.generator_opt.load_state_dict(ckpt["generator_optimizer"])
        self.discriminator_opt.load_state_dict(ckpt["discriminator_optimizer"])
        return global_step + 1
    @staticmethod
    def copy_brain_into_tensor(tensor: torch.tensor,
                               original: torch.tensor,
                               brain_mask: torch.tensor) -> torch.tensor:
        return torch.where(brain_mask, original, tensor)
    @staticmethod
    def starmap(func: Callable, iterable: Iterable):
        return list(itertools.starmap(func, iterable))
    @staticmethod
    def detach(tensors: List[torch.tensor]) -> List[torch.tensor]:
        return [t.detach() for t in tensors]
    @staticmethod
    def move_to_device(tensors: List[torch.Tensor], device= torch.device) -> List[torch.Tensor]:
        return [t.to(device) for t in tensors]
    def generator_step(self,
                       original: List[torch.FloatTensor],
                       conditional: List[torch.FloatTensor],
                       brain: List[torch.FloatTensor],
                       brain_mask: List[torch.FloatTensor],
                       slice_index: List[torch.Tensor]):
        self.generator_opt.zero_grad()
        fake_gray: List[torch.tensor] = self.generator(conditional=conditional,
                                                       brain_mask=brain_mask,
                                                       brain=brain,
                                                       slice_index=slice_index)
        brain_mask_binary = [torch.bernoulli(brain_mask_t).type(torch.bool) for brain_mask_t in brain_mask]
        fake_gray: List[torch.tensor] = self.starmap(self.copy_brain_into_tensor, zip(fake_gray, original, brain_mask_binary))
        discriminator_on_fake = self.discriminator(probe=self.move_to_device(fake_gray, device=self.secondary_device),
                                                   conditional=self.move_to_device(conditional, device=self.secondary_device),
                                                   brain=self.move_to_device(brain, device=self.secondary_device),
                                                   brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
                                                   slice_index=self.move_to_device(slice_index, device=self.secondary_device))
        if self.loss_fct == 'WGAN_GP':
            loss = -torch.mean(discriminator_on_fake)
        elif self.loss_fct == 'LS_GAN':
            loss = ls_gan_loss(x=discriminator_on_fake, y=None).mean()
        elif self.loss_fct in ('RA_LS_GAN', 'RA_HINGE_GAN'):
            discriminator_on_original = self.discriminator(probe=self.move_to_device(original, device=self.secondary_device),
                                                           conditional=self.move_to_device(conditional, device=self.secondary_device),
                                                           brain=self.move_to_device(brain, device=self.secondary_device),
                                                           brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
                                                           slice_index=self.move_to_device(slice_index, device=self.secondary_device))
            func = globals()[self.loss_fct.lower() + "_loss"]
            loss = func(discriminator_on_fake, discriminator_on_original).mean()
        else:
            raise NotImplementedError("Unknown loss function!")
        loss.backward()
        self.generator_opt.step()
        if self.use_ema:
            update_average(model_tgt=self.generator_shadow, model_src=self.generator, beta=self.ema_decay)
    def discriminator_step(self,
                           original: List[torch.FloatTensor],
                           conditional: List[torch.FloatTensor],
                           brain: List[torch.FloatTensor],
                           brain_mask: List[torch.FloatTensor],
                           slice_index: List[torch.FloatTensor]):
        self.discriminator_opt.zero_grad()
        fake_gray: List[torch.FloatTensor] = self.detach(self.generator(conditional=conditional,
                                                                        brain=brain,
                                                                        brain_mask=brain_mask,
                                                                        slice_index=slice_index))
        brain_mask_binary = [torch.bernoulli(brain_mask_t).type(torch.bool) for brain_mask_t in brain_mask]
        fake_gray: List[torch.FloatTensor] = self.starmap(self.copy_brain_into_tensor, zip(fake_gray, original, brain_mask_binary))
        discriminator_on_fake = self.discriminator(probe=self.move_to_device(self.detach(fake_gray), device=self.secondary_device),
                                                   conditional=self.move_to_device(conditional, device=self.secondary_device),
                                                   brain=self.move_to_device(brain, device=self.secondary_device),
                                                   brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
                                                   slice_index=self.move_to_device(slice_index, device=self.secondary_device))
        discriminator_on_original = self.discriminator(
            probe=self.move_to_device(original, device=self.secondary_device),
            conditional=self.move_to_device(conditional, device=self.secondary_device),
            brain=self.move_to_device(brain, device=self.secondary_device),
            brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
            slice_index=self.move_to_device(slice_index, device=self.secondary_device))
        if self.loss_fct == 'WGAN_GP':
            batch_size = original[0].shape[0]
            epsilon = torch.rand(batch_size, device=self.secondary_device)
            blends = []
            for f, o in zip(self.move_to_device(fake_gray, device=self.secondary_device),
                            self.move_to_device(original, device=self.secondary_device)):
                e = epsilon.view(batch_size, *(1,) * (f.ndim - 1))
                blend = (1 - e) * f + e * o
                blend.requires_grad_(True)
                blends.append(blend)
            discriminator_on_blends = self.discriminator(
                probe=blends,
                conditional=self.move_to_device(conditional, device=self.secondary_device),
                brain=self.move_to_device(brain, device=self.secondary_device),
                brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
                slice_index=self.move_to_device(slice_index, device=self.secondary_device))
            grads = torch.autograd.grad(outputs=discriminator_on_blends,
                                        inputs=blends,
                                        grad_outputs=torch.ones_like(discriminator_on_blends),
                                        create_graph=True,
                                        retain_graph=True,
                                        only_inputs=True)
            grads = [grad.flatten(start_dim=1) for grad in grads]
            grads_norm = [torch.mul(grad, grad).sum(dim=1).add(1E-12).sqrt() for grad in grads]
            gradient_penalty = torch.mean((grads_norm[-1] - 1)**2)
            print("Gradient penalty:", gradient_penalty.detach().cpu().numpy())
            loss = torch.mean(discriminator_on_fake - discriminator_on_original) + self.wgan_gp_alpha * gradient_penalty
        elif self.loss_fct == 'LS_GAN':
            loss = ls_gan_loss(x=discriminator_on_original, y=discriminator_on_fake).mean()
        elif self.loss_fct in ('RA_LS_GAN', 'RA_HINGE_GAN'):
            func = globals()[self.loss_fct.lower() + "_loss"]
            loss = func(discriminator_on_original, discriminator_on_fake).mean()
        else:
            raise NotImplementedError("Unknown loss function")
        loss.backward()
        self.discriminator_opt.step()
    def save_image(self,
                   global_step: int,
                   conditional: List[torch.Tensor],
                   brain: List[torch.FloatTensor],
                   brain_mask: List[torch.FloatTensor],
                   slice_index: List[torch.FloatTensor],
                   original: List[torch.Tensor],
                   fake_gray: List[torch.Tensor]):
        batch_size = original[0].shape[0]
        highest_resolution = original[-1].shape[2:]
        ref_min = original[-1].flatten(start_dim=1).min(dim=1).values.view(batch_size, 1, 1, 1)
        ref_max = original[-1].flatten(start_dim=1).max(dim=1).values.view(batch_size, 1, 1, 1)
        def _prepare_single_resolution(tensor: torch.tensor, rescale_mode: str = False) -> torch.tensor:
            tensor = F.upsample(input=tensor,
                                size=highest_resolution,
                                mode="bilinear",
                                align_corners=True)
            if rescale_mode == "original":
                tensor = (tensor - ref_min)/(ref_max - ref_min)
            tensor = torch.clamp(255 * tensor, min=0, max=255) 
            n = int(math.sqrt(batch_size))
            tensor = tensor[:n**2]     
            grid = torchvision.utils.make_grid(tensor, padding=0, nrow=n)
            grid = grid.transpose(0, 2)
            return grid
        conditional = [_prepare_single_resolution(t).cpu().numpy() for t in conditional]
        brain = [_prepare_single_resolution(t, rescale_mode='original').cpu().numpy() for t in brain]
        brain_mask = [_prepare_single_resolution(t).cpu().numpy() for t in brain_mask]
        slice_index = [_prepare_single_resolution(t).cpu().numpy() for t in slice_index]
        original = [_prepare_single_resolution(t, rescale_mode='original').cpu().numpy() for t in original]
        fake_gray = [_prepare_single_resolution(t, rescale_mode='original').cpu().numpy() for t in fake_gray]
        overview = np.concatenate([np.concatenate(tup, axis=1) for tup in zip(conditional,
                                                                              brain,
                                                                              brain_mask,
                                                                              slice_index,
                                                                              original,
                                                                              fake_gray)], axis=0)
        sample_pngs_dir = self.model_dir.joinpath("sample_pngs")
        output_fname = sample_pngs_dir.joinpath(f"{global_step}.png")
        imsave(output_fname, arr=overview.astype(np.uint8))
    def save_video(self,
                   global_step: int,
                   conditional: List[torch.FloatTensor],
                   brain: List[torch.FloatTensor],
                   brain_mask: List[torch.FloatTensor],
                   slice_index: List[torch.FloatTensor],
                   original: List[torch.FloatTensor],
                   fake_gray: List[torch.FloatTensor]):
        batch_size = original[0].shape[0]
        highest_resolution = original[-1].shape[-1]
        ref_min = original[-1].flatten(start_dim=1).min(dim=1).values.view(batch_size, 1, 1, 1, 1)
        ref_max = original[-1].flatten(start_dim=1).max(dim=1).values.view(batch_size, 1, 1, 1, 1)
        def _prepare_single_resolution(tensor: torch.FloatTensor, rescale_mode: str = False) -> torch.FloatTensor:
            batch_size = tensor.shape[0]
            tensor = F.upsample(input=tensor,
                                size=(highest_resolution,) * 3,
                                mode="trilinear",
                                align_corners=True)
            if rescale_mode == "original":
                tensor = (tensor - ref_min)/(ref_max - ref_min)
            tensor = torch.clamp(255 * tensor, min=0, max=255)
            n = int(math.sqrt(batch_size))
            tensor = tensor[:n**2]     
            grid = torch.stack([torchvision.utils.make_grid(tensor[..., i],
                                                  nrow=n,
                                                  padding=0) for i in range(highest_resolution)], dim=3)
            grid = grid.transpose(0, 3)
            return grid
        conditional = [_prepare_single_resolution(t).cpu().numpy() for t in conditional]
        brain = [_prepare_single_resolution(t, rescale_mode='original').cpu().numpy() for t in brain]
        brain_mask = [_prepare_single_resolution(t).cpu().numpy() for t in brain_mask]
        original = [_prepare_single_resolution(t, rescale_mode='original').cpu().numpy() for t in original]
        fake_gray = [_prepare_single_resolution(t, rescale_mode='original').cpu().numpy() for t in fake_gray]
        overview = np.concatenate([np.concatenate(tup, axis=2) for tup in zip(conditional,
                                                                              brain,
                                                                              brain_mask,
                                                                              original,
                                                                              fake_gray)], axis=1)
        sample_gifs_dir = self.model_dir.joinpath("sample_gifs")
        output_fname = sample_gifs_dir.joinpath(f"{global_step}.gif")
        video = ImageSequenceClip(list(overview), fps=10)
        video.write_gif(filename=output_fname, fps=10, verbose=False, logger=None)
    def reporting_step(self,
                       global_step: int,
                       save_video: bool,
                       save_checkpoint: bool,
                       original: List[torch.FloatTensor],
                       conditional: List[torch.FloatTensor],
                       brain: List[torch.FloatTensor],
                       brain_mask: List[torch.FloatTensor],
                       slice_index: List[torch.FloatTensor]):
        if self.local_rank != 0:
            return
        with torch.no_grad():
            if self.use_ema:
                fake_gray: List[torch.FloatTensor] = self.generator_shadow(conditional=conditional,
                                                                           brain=brain,
                                                                           brain_mask=brain_mask,
                                                                           slice_index=slice_index)
            else:
                fake_gray: List[torch.FloatTensor] = self.generator(conditional=conditional,
                                                                    brain=brain,
                                                                    brain_mask=brain_mask,
                                                                    slice_index=slice_index)
            brain_mask_binary = [torch.bernoulli(brain_mask_t).type(torch.bool) for brain_mask_t in brain_mask]
            fake_gray: List[torch.FloatTensor] = self.starmap(self.copy_brain_into_tensor,
                                                              zip(fake_gray, original, brain_mask_binary))
            discriminator_on_fake = self.discriminator(probe=self.move_to_device(fake_gray, device=self.secondary_device),
                                                       conditional=self.move_to_device(conditional, device=self.secondary_device),
                                                       brain=self.move_to_device(brain, device=self.secondary_device),
                                                       brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
                                                       slice_index=self.move_to_device(slice_index, device=self.secondary_device))
            discriminator_on_original = self.discriminator(probe=self.move_to_device(original, device=self.secondary_device),
                                                           conditional=self.move_to_device(conditional, device=self.secondary_device),
                                                           brain=self.move_to_device(brain, device=self.secondary_device),
                                                           brain_mask=self.move_to_device(brain_mask, device=self.secondary_device),
                                                           slice_index=self.move_to_device(slice_index, device=self.secondary_device))
            if self.loss_fct == "WGAN_GP":
                generator_loss = (-discriminator_on_fake).cpu().numpy()
                discriminator_loss = (discriminator_on_fake - discriminator_on_original).cpu().numpy()
            elif self.loss_fct == 'LS_GAN':
                generator_loss = ls_gan_loss(x=discriminator_on_fake, y=None).cpu().numpy()
                discriminator_loss = ls_gan_loss(x=discriminator_on_original, y=discriminator_on_fake).cpu().numpy()
            elif self.loss_fct in ('RA_LS_GAN', 'RA_HINGE_GAN'):
                func = globals()[self.loss_fct.lower() + "_loss"]
                generator_loss = func(discriminator_on_fake, discriminator_on_original, distributed=False).cpu().numpy()
                discriminator_loss = func(discriminator_on_original, discriminator_on_fake, distributed=False).cpu().numpy()
            print(
                "{} | G = {:.4f} | D = {:.4f} | D(fake) = {:.4f} | D(original) = {:.4f} | D(original) - D(fake) = {:.4f}".format(
                    global_step,
                    generator_loss.mean(),
                    discriminator_loss.mean(),
                    discriminator_on_fake.mean(),
                    discriminator_on_original.mean(),
                    (discriminator_on_original - discriminator_on_fake).mean()))
            with open(self.model_dir.joinpath("generator_loss.txt"), "a") as f_g, open(self.model_dir.joinpath("discriminator_loss.txt"), "a") as f_d:
                def _array_as_str(arr: np.ndarray) -> str:
                    return "\t".join(map(str, arr.tolist()))
                f_g.write(_array_as_str(generator_loss) + "\n")
                f_d.write(_array_as_str(discriminator_loss) + "\n")
            tensors = {
                "conditional": conditional,
                "brain": brain,
                "brain_mask": brain,
                "slice_index": slice_index,
                "original": original,
                "fake_gray": fake_gray
            }
            if save_video:
                if self.dimensionality == 2:
                    self.save_image(global_step=global_step, **tensors)
                elif self.dimensionality == 3:
                    self.save_video(global_step=global_step, **tensors)
            if save_checkpoint:
                torch.save({
                    "generator": self.generator_shadow.state_dict() if self.use_ema else self.generator.state_dict(),
                    "discriminator": self.discriminator.state_dict(),
                    "generator_optimizer": self.generator_opt.state_dict(),
                    "discriminator_optimizer": self.discriminator_opt.state_dict()
                }, self.model_dir.joinpath("checkpoints").joinpath(f"checkpoint_{global_step}.pt"))
                def _split_resolutions(tensors: List[torch.Tensor], name: str) -> Dict[str, np.ndarray]:
                    return {f"{name}_{t.shape[-1]}": t.cpu().numpy() for t in tensors}
                np.savez_compressed(file=self.model_dir.joinpath("sample_arrays", f"{global_step}.npz"),
                                    **_split_resolutions(original, name="original"),
                                    **_split_resolutions(fake_gray, name="fake_gray"),
                                    **_split_resolutions(brain_mask, name="brain_mask"))
    def train(self, global_step):
        for epoch_id in range(self.num_epochs):
            self.sampler.set_epoch(epoch_id)
            for batch in self.data:
                del batch['example_id']
                batch = {k: t.to(self.device, non_blocking=True) for k, t in batch.items()}
                batch = {k: build_resolutions(t,
                                              dimensionality=self.dimensionality,
                                              n=self.num_blocks - 1) for k, t in batch.items()}
                if global_step % 3 == 0:
                    self.generator_step(**batch)
                elif global_step % 3 == 1:
                    self.discriminator_step(**batch)
                else:
                    save_video = (global_step - 2) % (3 * self.video_frequency) == 0
                    save_checkpoint = (global_step - 2) % (3 * self.checkpoint_frequency) == 0
                    self.reporting_step(global_step=global_step,
                                        save_video=save_video,
                                        save_checkpoint=save_checkpoint,
                                        **batch)
                global_step += 1
