import os
from argparse import ArgumentParser
import pathlib
import tempfile
from datetime import datetime
import shutil
import yaml
import torch
from remodeling.generative.CPGAN.orchestrator import GAN
torch.manual_seed(5)
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
def main():
    parser = ArgumentParser()
    parser.add_argument("--mode",
                        choices=('train', 'test'),
                        default='train')
    parser.add_argument("--model_dir",
                        help="Where a previously trained model is stored")
    parser.add_argument("--tmp_dir",
                        help="Where temporary data should be stored",
                        default=tempfile.gettempdir())
    parser.add_argument("--index_fname",
                        help="A YAML file that contains indices for training & test samples",
                        required=True)
    parser.add_argument("--config_fname",
                        help="Where the config file (.yml) is located")
    parser.add_argument("--data_hdf5_fname",
                        help="Where the HDF5 input file is located",
                        required=True)
    parser.add_argument("--local_rank",
                        help="The local rank in a distributed setting",
                        default=0,
                        type=int)
    args = parser.parse_args()
    assert args.mode == "train" or os.path.exists(args.model_dir), "Model directory has to exist in test mode!"
    assert args.mode == "test" or os.path.exists(args.tmp_dir), "Temporary directory does not exist!"
    assert os.path.exists(args.index_fname), "Index file does not exist!"
    assert args.config_fname is not None or args.model_dir is not None, "Cannot obtain config!"
    assert os.path.exists(args.data_hdf5_fname), "Input HDF5 file does not exist!"
    args.tmp_dir = pathlib.Path(args.model_dir).parent if args.model_dir is not None else pathlib.Path(args.tmp_dir)
    args.index_fname = pathlib.Path(args.index_fname)
    if args.model_dir is not None:
        args.config_fname = pathlib.Path(args.model_dir).joinpath("config.yml")
        args.model_dir = pathlib.Path(args.model_dir)
    else:
        args.config_fname = pathlib.Path(args.config_fname)
    args.config_fname = pathlib.Path(args.config_fname) if args.mode == "train" else pathlib.Path(args.model_dir).joinpath("config.yml")
    args.data_hdf5_fname = pathlib.Path(args.data_hdf5_fname)
    with open(args.config_fname, "r") as config_f:
        config = yaml.load(config_f, Loader=yaml.FullLoader)
    world_size = int(os.environ.get('WORLD_SIZE', '1'))
    if world_size > 1:
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = '12356'
        torch.cuda.device(args.local_rank)
        torch.distributed.init_process_group(backend='nccl',
                                     init_method='env://',
                                     world_size=world_size,
                                     rank=args.local_rank)
    if torch.cuda.is_available():
        device = torch.device(f"cuda:{2 * args.local_rank}")
        secondary_device = torch.device(f"cuda:{2 * args.local_rank + 1}")
        print("Using devices: ", device, secondary_device)
    else:
        print("No CUDA available!")
        device = torch.device("cpu")
        secondary_device = device
    restore_checkpoint = args.model_dir is not None
    if not restore_checkpoint and args.local_rank == 0:
        data_fmt = datetime.now().strftime("%Y-%m-%d_%H-%M-%S_C_ANON_GAN")
        args.model_dir = args.tmp_dir.joinpath(data_fmt)
        args.model_dir.mkdir()
        args.model_dir.joinpath("checkpoints").mkdir()
        args.model_dir.joinpath("sample_pngs").mkdir()
        args.model_dir.joinpath("sample_gifs").mkdir()
        args.model_dir.joinpath("sample_arrays").mkdir()
        shutil.copyfile(src=args.config_fname, dst=args.model_dir.joinpath("config.yml"))
        shutil.copytree(".", args.model_dir.joinpath("Code"))
        print("Model directory:", args.model_dir)
        print(args.config_fname.read_text().strip(), "\n---")
    gan = GAN(mode=args.mode,
              model_dir=args.model_dir,
              data_fname=args.data_hdf5_fname,
              index_fname=args.index_fname,
              config=config,
              local_rank=args.local_rank,
              world_size=world_size,
              device=device,
              secondary_device=secondary_device)
    global_step = 0
    if restore_checkpoint:
        global_step = gan.restore_checkpoint()
    if args.mode == "train":
        gan.train(global_step)
    else:
        gan.test()
if __name__ == "__main__":
    main()
