""" Bischoff-Grethe et al. """
from argparse import ArgumentParser
import pathlib
from tempfile import gettempdir
import gzip
import shutil
import subprocess
import shlex
from functools import partial
import stat
from joblib import delayed, Parallel
from tqdm import tqdm
MRI_DEFACE_DIR = pathlib.Path().cwd().joinpath("traditional_anonymization_methods", "deface_bin")
MRI_DEFACE_BIN_PATH = MRI_DEFACE_DIR.joinpath("mri_deface_linux")
MRI_DEFACE_FACE_PATH = MRI_DEFACE_DIR.joinpath("face.gca.gz")
MRI_DEFACE_TALAIRACH_PATH = MRI_DEFACE_DIR.joinpath("talairach_mixed_with_skull.gca.gz")
MRI_DEFACE_BIN_PATH.chmod(MRI_DEFACE_BIN_PATH.stat().st_mode | stat.S_IEXEC)
def gunzip_and_move_to_tmp_dir(fname: pathlib.Path, tmp_dir: pathlib.Path):
    output_fname = tmp_dir.joinpath(fname.stem)
    with gzip.open(fname, "rb") as f_zipped, open(output_fname, "wb") as f_gunzipped:
        shutil.copyfileobj(f_zipped, f_gunzipped)
    return output_fname
def execute_deface(nifti_fname: pathlib.Path,
                   output_dir: pathlib.Path,
                   mri_deface_talairach_path: pathlib.Path,
                   mri_deface_face_path: pathlib.Path):
    output_fname = output_dir.joinpath(nifti_fname.name)
    if output_fname.exists():
        print(f"File {output_dir} exists already. Skipping.")
        return
    try:
        subprocess.check_output(shlex.split(f"{MRI_DEFACE_BIN_PATH} {nifti_fname} {mri_deface_talairach_path} {mri_deface_face_path} {output_fname}"))
    except subprocess.CalledProcessError as e:
        print(e.output)
def main():
    parser = ArgumentParser()
    parser.add_argument("--nifti_dir",
                        help="Where the NifTi input files reside",
                        required=True)
    parser.add_argument("--tmp_dir",
                        help="Where the auxiliary files will be unzipped",
                        default=gettempdir())
    parser.add_argument("--num_jobs",
                        type=int,
                        default=8)
    args = parser.parse_args()
    args.nifti_dir = pathlib.Path(args.nifti_dir)
    args.tmp_dir = pathlib.Path(args.tmp_dir)
    args.nifti_dir.exists(), "Input NifTi directory does not exist!"
    mri_deface_face_path = gunzip_and_move_to_tmp_dir(fname=MRI_DEFACE_FACE_PATH, tmp_dir=args.tmp_dir)
    mri_deface_talairach_path = gunzip_and_move_to_tmp_dir(fname=MRI_DEFACE_TALAIRACH_PATH, tmp_dir=args.tmp_dir)
    output_dir = args.nifti_dir.with_name(f"{args.nifti_dir.name}__deface__")
    output_dir.mkdir(exist_ok=True)
    nifti_fnames = [nifti_fname for nifti_fname in args.nifti_dir.glob("*.nii.gz")]
    _execute_tmp_deface = partial(execute_deface,
                                  mri_deface_talairach_path=mri_deface_talairach_path,
                                  mri_deface_face_path=mri_deface_face_path)
    Parallel(n_jobs=args.num_jobs)(delayed(_execute_tmp_deface)(nifti_fname=nifti_fname,
                                                                output_dir=output_dir) for nifti_fname in tqdm(nifti_fnames))
if __name__ == "__main__":
    main()
