from argparse import ArgumentParser
import pathlib
import operator
import functools
import numpy as np
import re
import pandas as pd
from typing import Dict, List
from utils import get_real_stem
from tqdm import tqdm
import nibabel as nib
from skimage.metrics import structural_similarity as ssim
from joblib import delayed, Parallel
extract_type = re.compile(r"__([a-z_-]*)__")
def calculate_ssim(example_id: str,
                   reference_dir: pathlib.Path,
                   anonymized_dirs: Dict[str, pathlib.Path],
                   win_sizes: List[int]):
    reference_img = nib.load(reference_dir.joinpath(example_id).with_suffix(".nii.gz")).get_fdata(dtype=np.float32)
    results = [dict(win_size=win_size) for win_size in win_sizes]
    for anonymized_dir, anonymized_dir_path in anonymized_dirs.items():
        img = nib.load(anonymized_dir_path.joinpath(example_id).with_suffix(".nii.gz")).get_fdata(dtype=np.float32)
        for i, win_size in enumerate(win_sizes):
            ssim_score = ssim(reference_img, img, win_size=win_size)
            results[i].update({anonymized_dir: ssim_score})
    return results
def main():
    parser = ArgumentParser()
    parser.add_argument("--reference_dir",
                        help="Where the original NifTi files resides",
                        required=True)
    parser.add_argument("--anonymized_dirs",
                        nargs="+",
                        required=True)
    parser.add_argument("--test_indices",
                        required=True)
    parser.add_argument("--win_sizes",
                        help="Size of the window",
                        nargs="+",
                        default=(3, 5, 7, 9, 11, 13),
                        type=int)
    parser.add_argument("--num_jobs",
                        default=8,
                        type=int)
    args = parser.parse_args()
    args.reference_dir = pathlib.Path(args.reference_dir)
    args.anonymized_dirs = list(map(pathlib.Path, args.anonymized_dirs))
    args.test_indices = pathlib.Path(args.test_indices)
    assert args.reference_dir.exists(), "Reference directory does not exist!"
    assert all(anonymized_dir.exists() for anonymized_dir in args.anonymized_dirs), "Not all anonymized directories exist!"
    assert args.test_indices.exists(), "Test indices file does not exist!"
    ix = set(args.test_indices.read_text().strip().split("\n"))
    example_ix = {dir_path.name: set(map(get_real_stem, dir_path.glob("*"))) for dir_path in args.anonymized_dirs}
    common_example_ix = set(functools.reduce(operator.and_, example_ix.values()))
    test_ix = list(ix & common_example_ix)
    args.anonymized_dirs = {extract_type.search(anonymized_dir.name).group(1): anonymized_dir for anonymized_dir in args.anonymized_dirs}
    ssim_scores = Parallel(n_jobs=args.num_jobs)(delayed(calculate_ssim)(example_id=example_id,
                                                                         reference_dir=args.reference_dir,
                                                                         anonymized_dirs=args.anonymized_dirs,
                                                                         win_sizes=args.win_sizes) for example_id in
                                                 tqdm(test_ix))
    ssim_scores = [item for sublist in ssim_scores for item in sublist]
    data = pd.DataFrame(ssim_scores)
    data.to_csv(args.reference_dir.parent.joinpath("SSIM_scores.csv"))
if __name__ == "__main__":
    main()
