from argparse import ArgumentParser
import pathlib
import operator
import itertools
import functools
import operator
import numpy as np
import re
from joblib import delayed, Parallel
from tqdm import tqdm
import pandas as pd
import nibabel as nib
extract_table_header = re.compile(r"^([a-zA-Z]+)\s+([a-zA-Z]+)\s+([a-zA-Z-]+)$", re.MULTILINE)
extract_table_row = re.compile(r"([a-zA-Z]+)\s+([0-9]+[\.[0-9]*]?)\s+([0-9]+[\.[0-9]*]?)")
SEGMENTATION_TYPES = ["I", "I_brain", "I_brain_mask", "I_brain_skull", "I_periph_render", "I_render", "I_stdmask",
                      "I_stdmask_segperiph", "I_stdmask_segvent", "I_stdmaskbrain", "I_stdmaskbrain_mixeltype",
                      "I_stdmaskbrain_pve_0", "I_stdmaskbrain_pve_0_segvent", "I_stdmaskbrain_pve_1",
                      "I_stdmaskbrain_pve_1_segperiph", "I_stdmaskbrain_pve_2", "I_stdmaskbrain_pveseg",
                      "I_stdmaskbrain_seg", "I_stdmaskbrain_seg_0", "I_stdmaskbrain_seg_1", "I_stdmaskbrain_seg_2"]
def extract_data_from_sienax_dir(dir_path: pathlib.Path, data_type: str):
    subject_id = dir_path.name.replace("_sienax", "")
    assert dir_path.exists(), "SIENAX directory does not exist!"
    report_data = dir_path.joinpath("report.sienax").read_text().strip()
    if "Cannot read" in report_data:
        return None
    table_header = extract_table_header.search(report_data).groups()
    table_rows = extract_table_row.findall(report_data)
    df = pd.DataFrame(table_rows, columns=table_header)
    df.insert(loc=0, column="data_type", value=data_type)
    df.insert(loc=0, column="subject_id", value=subject_id)
    return df
def dice_score(a: np.ndarray, b: np.ndarray) -> float:
    return 2 * np.divide(np.sum(a * b), np.sum(a) + np.sum(b))
def compute_pairwise_dice_scores(segmentation_types, **sienax_dirs):
    subject_id = sienax_dirs["original"].name.replace("_sienax", "")
    results = []
    try:
        for segmentation_type in segmentation_types:
            segmentations = dict()
            for data_type, sienax_dir in sienax_dirs.items():
                segmentation_fname = sienax_dir.joinpath(f"{segmentation_type}.nii.gz")
                segmentations[data_type] = nib.load(segmentation_fname).get_fdata(dtype=np.float32)
            dice_scores = {data_type: dice_score(a=segmentations["alt"], b=other_segmentation) for data_type, other_segmentation in segmentations.items()}
            dice_scores["subject_id"] = subject_id
            dice_scores["segmentation_type"] = segmentation_type
            results.append(dice_scores)
    except FileNotFoundError:
        print(f"Error with subject {subject_id}")
    finally:
        return results
def main():
    parser = ArgumentParser()
    parser.add_argument("--data_root_dir",
                        help="The dataset (e.g. 'OASIS-3')",
                        required=True)
    parser.add_argument("--base_modality",
                        help="The modality that the anonymization algorithms ran on (e.g. 'IXI_NIFTI_PT')",
                        required=True)
    parser.add_argument("--data_types",
                        help="The data types (i.e. 'face_mask', 'mri_watershed' etc.) that should be evaluated upon",
                        nargs="+",
                        default=('deface', 'face_mask', 'mri_watershed', 'quickshear', 'cp-gan', 'alt'))
    parser.add_argument("--segmentation_types",
                        help="For which files the DICE scores should be computed for",
                        nargs="+",
                        default=('I_stdmask', 'I_stdmaskbrain_seg_1',
                                 'I_stdmaskbrain_seg_2', 'I_stdmask_segvent'))
    parser.add_argument("--resolution",
                        help="The resolution that was used with SIENAX",
                        default=128)
    parser.add_argument("--num_jobs",
                        type=int,
                        default=10)
    args = parser.parse_args()
    args.data_root_dir = pathlib.Path(args.data_root_dir)
    args.data_dirs = {data_type: args.data_root_dir.joinpath(f"{args.base_modality}_{args.resolution}__{data_type}___FslSienax") for data_type in args.data_types}
    args.data_dirs["original"] = args.data_root_dir.joinpath(f"{args.base_modality}_{args.resolution}_FslSienax")
    assert all(map(operator.methodcaller("exists"), args.data_dirs.values())), "Not all data directories exist!"
    assert all(seg_type in SEGMENTATION_TYPES for seg_type in args.segmentation_types), "Segmentation types are not specified correctly!"
    sienax_subjects = {data_type: set(map(operator.attrgetter('name'), dir_path.glob("*_sienax"))) for data_type, dir_path in args.data_dirs.items()}
    sienax_common_subjects = functools.reduce(operator.and_, sienax_subjects.values())
    sienax_subjects = {data_type: subjects & sienax_common_subjects for data_type, subjects in sienax_subjects.items()}
    data_sienax_dirs = {data_type: [args.data_dirs[data_type].joinpath(sienax_subject) for sienax_subject in sorted(sienax_subjects)] for data_type, sienax_subjects in sienax_subjects.items()}
    def _non_empty_concat(xs, **kwargs):
        return pd.concat([x for x in xs if x is not None], **kwargs)
    evaluation_volumes = [_non_empty_concat([extract_data_from_sienax_dir(dir_path=sienax_dir, data_type=data_type) for sienax_dir in sienax_dirs], axis=0, ignore_index=True) for data_type, sienax_dirs in data_sienax_dirs.items()]
    evaluation_volumes = pd.concat(evaluation_volumes, axis=0, ignore_index=True)
    paired_fnames = [{data_type: data_dir.joinpath(sienax_subject) for data_type, data_dir in args.data_dirs.items()} for sienax_subject in sienax_common_subjects]
    paired_results = Parallel(n_jobs=args.num_jobs)(delayed(compute_pairwise_dice_scores)(segmentation_types=args.segmentation_types, **sienax_dirs) for sienax_dirs in tqdm(paired_fnames))
    evaluation_dice = pd.DataFrame.from_records(itertools.chain(*paired_results))
    evaluation_volumes.to_csv(args.data_root_dir.joinpath(f"{args.base_modality}_{args.resolution}_FslSienax_VOLUMES.csv"), index=False)
    evaluation_dice.to_csv(args.data_root_dir.joinpath(f"{args.base_modality}_{args.resolution}_FslSienax_DICE.csv"), index=False)
if __name__ == "__main__":
    main()
