from argparse import ArgumentParser
import pathlib
from joblib import delayed, Parallel
import numpy as np
import nibabel as nib
from tqdm import tqdm
WIDTH = 5
PADDING = 0
def process_file(input_fname: pathlib.Path, output_dir: pathlib.Path):
    output_fname = output_dir.joinpath(input_fname.name)
    if output_fname.exists():
        return
    img = nib.load(str(input_fname))
    arr = img.get_fdata(dtype=np.float32)
    fill_v = np.min(arr)
    arr_n = (arr - arr.min())/(arr.max() - arr.min())
    noise_edge = []
    for axis in (0, 1, 2):
        for i in (0, -1):
            axis_slice = slice(0, WIDTH) if i == 0 else slice(-WIDTH, None)
            axes_slice = tuple([slice(None) if j != axis else axis_slice for j in (0, 1, 2)])
            noise_edge.append(arr_n[axes_slice])
    noise_level = np.median(np.concatenate(noise_edge, axis=None))
    for axis in (0, 1, 2):
        intensity_line = np.max(arr_n, axis=tuple(k for k in (0, 1, 2) if k != axis))
        above_noise_threshold = intensity_line > noise_level
        binary_code = "".join(above_noise_threshold.astype(np.uint8).astype(str))
        subseqs = binary_code.split("0")
        ix = np.argmax(list(map(len, subseqs)))
        longest_subseq = subseqs[ix]
        start_ix = binary_code.index(longest_subseq)    
        end_ix = start_ix + len(longest_subseq)     
        start_ix = start_ix - PADDING
        end_ix = end_ix + PADDING
        arr = arr.take(indices=range(max(start_ix, 0), min(end_ix, len(binary_code))), axis=axis)
        pad_width = [(0, 0) for _ in (0, 1, 2)]
        pad_width[axis] = (-start_ix if start_ix < 0 else 0, end_ix - len(binary_code) if end_ix > len(binary_code) else 0)
        arr = np.pad(arr,
                     pad_width=pad_width,
                     mode="constant",
                     constant_values=fill_v)
    img = nib.Nifti1Image(arr, img.affine, img.header)
    nib.save(img=img, filename=str(output_fname))
def main():
    parser = ArgumentParser()
    parser.add_argument("--input_dir",
                        help="Where the input NifTi files are stored",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the trimmed NifTi files are being stored",
                        required=True)
    parser.add_argument("--num_jobs",
                        type=int,
                        default=1)
    args = parser.parse_args()
    args.input_dir = pathlib.Path(args.input_dir)
    args.output_dir = pathlib.Path(args.output_dir)
    args.output_dir.mkdir(exist_ok=True)
    assert args.input_dir, "Input directory does not exist!"
    Parallel(n_jobs=args.num_jobs)(delayed(process_file)(input_fname=input_fname,
                                                         output_dir=args.output_dir) for input_fname in tqdm(args.input_dir.glob("*.nii.gz")))
if __name__ == "__main__":
    main()
