import os
import pathlib
from argparse import ArgumentParser
import boto3
import pandas as pd
from datetime import datetime
from user_study.upload_images_to_s3_bucket import get_random_string
ENDPOINT_URL = 'https://mturk-requester-sandbox.us-east-1.amazonaws.com'
MARKETPLACE_URL = 'https://workersandbox.mturk.com/mturk/preview?groupId='
S3_PATH = "https://{}.s3.amazonaws.com"
NUM_HITS = 800
LIFETIME_HOURS = 24
ASSIGNMENT_DURATION_MINUTES = 10
AUTO_APPROVAL_DURATION_HOURS = 12
REWARD = '0.02'
MAX_ASSIGNMENTS = 5
RANDOM_ID = get_random_string()
def one_vs_many(mturk, regime, subject_id_to_obfuscated_id, bucket_name, export_as_html, output_fname):
    print("Estimated costs: ", NUM_HITS * MAX_ASSIGNMENTS * float(REWARD), "USD")
    template_html = open('template_1vK.htm', mode='r').read().strip()
    regime['HITGroupId'] = ''
    regime['HITId'] = ''
    for hit_id in range(NUM_HITS):
        hit_def = regime.iloc[hit_id, :]
        anon_type = hit_def['anon_type']
        original_pos = hit_def['original_pos']
        participant_ids = hit_def[['original_id'] + [col for col in hit_def.keys() if col.startswith("alternative_id")]].tolist()
        participant_ids.insert(original_pos + 1, participant_ids[0])
        hit_source_code = template_html
        for i, participant_id in enumerate(participant_ids):
            img_type = "original" if i == 0 else anon_type
            hit_source_code = hit_source_code.replace("{{IMG_" + str(i) + "_SRC}}",
                                                      os.path.join(S3_PATH.format(bucket_name), subject_id_to_obfuscated_id[participant_id], f"{img_type}.png"))
        if export_as_html:
            with open(pathlib.Path.cwd().joinpath("sample_html").joinpath(f"{hit_id}.htm"), "w") as f:
                f.write(hit_source_code)
        else:
            template_xml = open('template_base.xml', mode='r').read()
            hit_source_code = template_xml.format(hit_source_code)
            hit = mturk.create_hit(
                Title='Identification of patients on anonymized MRI scans ' + f"[ver: {RANDOM_ID}]",
                Description='We are interesting in quantifying how well certain anonymization methods anonymize MRI scans. The task '
                            'is to match those anonymized images to one original (unanonymized) image. Keyboard shortcuts are built-in'
                            'to make the selection process quicker.',
                Keywords='images, quick, matching',
                Reward=REWARD,
                MaxAssignments=MAX_ASSIGNMENTS,
                LifetimeInSeconds=LIFETIME_HOURS * 60 * 60,
                AssignmentDurationInSeconds=ASSIGNMENT_DURATION_MINUTES * 60,
                AutoApprovalDelayInSeconds=AUTO_APPROVAL_DURATION_HOURS * 60 * 60,
                Question=hit_source_code,
            )
            regime.at[hit_id, 'HITId'] = hit['HIT']['HITId']
            regime.at[hit_id, 'HITGroupId'] = hit['HIT']['HITGroupId']
            print(hit_id, hit['HIT']['HITId'], MARKETPLACE_URL + hit['HIT']['HITGroupId'])
    regime.to_csv(output_fname, index=False)
def many_vs_many(mturk, regime):
    raise NotImplementedError()
def main():
    parser = ArgumentParser()
    parser.add_argument("--regime_fname",
                        help="The CSV file that defines the sampled HITs",
                        required=True)
    parser.add_argument("--bucket_obfuscation_table_fname",
                        required=True)
    parser.add_argument("--mode",
                        help="The matching mode",
                        choices=("1vK", "KvK"),
                        default="1vK")
    parser.add_argument("--export_as_html",
                        default=False,
                        action="store_true")
    args = parser.parse_args()
    args.bucket_obfuscation_table_fname = pathlib.Path(args.bucket_obfuscation_table_fname)
    args.regime_fname = pathlib.Path(args.regime_fname)
    args.output_fname = args.regime_fname.parent.joinpath(f"{args.regime_fname.stem}_{datetime.today().isoformat()}.csv")
    assert args.bucket_obfuscation_table_fname.exists(), "Bucket obfuscation table not found!"
    assert args.regime_fname.exists(), "Regime file does not exist!"
    if args.export_as_html:
        pathlib.Path.cwd().joinpath("sample_html").mkdir(exist_ok=True)
    mturk = boto3.client( 'mturk', endpoint_url=ENDPOINT_URL)
    print("Account Balance:", mturk.get_account_balance()['AvailableBalance'])
    regime = pd.read_csv(args.regime_fname)
    subject_id_to_obfuscated_id = dict()
    bucket_name = args.bucket_obfuscation_table_fname.stem.split("_")[0]
    with open(args.bucket_obfuscation_table_fname, "r") as f:
        for line in f:
            obfuscated_subject_id, subject_id = line.strip().split("\t")
            subject_id_to_obfuscated_id[subject_id] = obfuscated_subject_id
    if args.mode == "1vK":
        one_vs_many(mturk=mturk,
                    regime=regime,
                    subject_id_to_obfuscated_id=subject_id_to_obfuscated_id,
                    bucket_name=bucket_name,
                    export_as_html=args.export_as_html,
                    output_fname=args.output_fname)
    elif args.mode == "KvK":
        many_vs_many(mturk, regime, subject_id_to_obfuscated_id, bucket_name)
    else:
        raise SystemError()
if __name__ == "__main__":
    main()
