from argparse import ArgumentParser
import pathlib
import xmltodict
import boto3
import pandas as pd
ENDPOINT_URL = 'https://mturk-requester.us-east-1.amazonaws.com'
def extract_integer_choice(s: str) -> int:
    if s == "choice-not-possible":
        return -1
    return int(s.split("-")[1]) - 1
def main():
    parser = ArgumentParser()
    parser.add_argument("--regime_hit_fname",
                        help="The CSV file that contains the regime and the HIT IDs",
                        required=True)
    args = parser.parse_args()
    args.regime_hit_fname = pathlib.Path(args.regime_hit_fname)
    args.output_fname = args.regime_hit_fname.parent.joinpath(f"{args.regime_hit_fname.stem}_eval.csv")
    assert args.regime_hit_fname.exists(), "Regime CSV does not exist!"
    regime = pd.read_csv(args.regime_hit_fname)
    regime = regime[~pd.isna(regime['HITId'])]
    anon_types = regime['anon_type'].unique().tolist()
    mturk = boto3.client( 'mturk', endpoint_url=ENDPOINT_URL)
    anon_counts = {anon_type: 0 for anon_type in anon_types}
    anon_correct_count = {anon_type: 0 for anon_type in anon_types}
    anon_unsure_count = {anon_type: 0 for anon_type in anon_types}
    results = []
    for i, row in regime.iterrows():
        original_id = row['original_id']
        anon_type = row['anon_type']
        original_pos = row['original_pos']
        hit_group_id = row['HITGroupId']
        hit_id = row['HITId']
        alternative_ids = row[[k for k in row.keys() if k.startswith('alternative_id')]].tolist()
        alternative_ids.insert(original_pos, original_id)
        worker_results = mturk.list_assignments_for_hit(HITId=hit_id)
        print(f"Found {worker_results['NumResults']} completed assignment(s) for HIT-ID = {hit_id} [{anon_type}]")
        for j, assignment in enumerate(worker_results['Assignments']):
            assignment_id = assignment['AssignmentId']
            worker_id = assignment['WorkerId']
            print(f"\t[{j}] Assignment-ID = {assignment_id}")
            print(f"\t[{j}] Worker-ID = {worker_id}")
            answer = xmltodict.parse(assignment['Answer'])['QuestionFormAnswers']['Answer']
            worker_choice = extract_integer_choice(answer['FreeText'])
            is_correct = worker_choice == original_pos
            print(f"\t[{j}] Worker-Choice = {worker_choice + 1}, [correctly identified = {'yes' if is_correct else 'no'}]")
            anon_counts[anon_type] += 1
            if is_correct:
                anon_correct_count[anon_type] += 1
            elif worker_choice == -1:
                anon_unsure_count[anon_type] += 1
            results.append({
                'hit_group_id': hit_group_id,
                'hit_id': hit_id,
                'assignment_id': assignment_id,
                'worker_id': worker_id,
                'anon_type': anon_type,
                'worker_choice': worker_choice,
                'correct_choice': original_pos
            })
    results = pd.DataFrame(results)
    results.to_csv(args.output_fname, index=False)
    print("Final evaluation:")
    for anon_type in anon_types:
        correct_percentage = 100 * anon_correct_count[anon_type]/anon_counts[anon_type]
        unsure_percentage = 100 * anon_unsure_count[anon_type]/anon_counts[anon_type]
        print(f"\t{anon_type}: 
if __name__ == "__main__":
    main()
