import pathlib
from argparse import ArgumentParser
import boto3
from botocore.exceptions import ClientError
import logging
from tqdm import tqdm
import string
import random
def get_random_string(n: int = 32):
    return "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(n))
def create_bucket(bucket_name, region=None):
    """Create an S3 bucket in a specified region
    If a region is not specified, the bucket is created in the S3 default
    region (us-east-1).
    :param bucket_name: Bucket to create
    :param region: String region to create bucket in, e.g., 'us-west-2'
    :return: True if bucket created, else False
    """
    try:
        if region is None:
            s3_client = boto3.client('s3')
            s3_client.create_bucket(Bucket=bucket_name)
        else:
            s3_client = boto3.client('s3', region_name=region)
            location = {'LocationConstraint': region}
            s3_client.create_bucket(Bucket=bucket_name,
                                    CreateBucketConfiguration=location)
    except ClientError as e:
        logging.error(e)
        return False
    return True
def main():
    parser = ArgumentParser()
    parser.add_argument("--user_study_dir",
                        help="Where the user study directory is located",
                        required=True)
    parser.add_argument("--make_public",
                        help="Whether to make the uploaded data public",
                        action="store_true",
                        default=False)
    args = parser.parse_args()
    args.user_study_dir = pathlib.Path(args.user_study_dir)
    args.subjects_dir = args.user_study_dir.joinpath("renderings")
    assert args.user_study_dir.exists(), "User study directory does not exist!"
    assert args.subjects_dir.exists(), "Subjects directory does not exist!"
    client = boto3.client('s3')
    bucket_name = "images" + get_random_string()
    create_bucket(bucket_name=bucket_name)
    s3 = boto3.resource('s3')
    if args.make_public:
        bucket = s3.Bucket(bucket_name)
        bucket.Acl().put(ACL='public-read')
    response = client.list_buckets()
    print('Existing buckets:')
    for bucket in response['Buckets']:
        print(f'  {bucket["Name"]}')
    obfuscation_table = dict()
    for subject_dir in tqdm(args.subjects_dir.iterdir()):
        subject_id = subject_dir.name
        if not subject_dir.is_dir():
            continue
        obfuscated_subject_id = get_random_string()
        obfuscation_table[subject_id] = obfuscated_subject_id
        client.put_object(Bucket=bucket_name, Body='', Key=obfuscated_subject_id + '/')
        for image_fname in subject_dir.glob("*.png"):
            item_path = f"{obfuscated_subject_id}/{image_fname.name}"
            client.upload_file(str(image_fname), bucket_name, item_path)
            if args.make_public:
                object_acl = s3.ObjectAcl(bucket_name, item_path)
                object_acl.put(ACL='public-read')
    obfuscation_table_fname = args.user_study_dir.joinpath(f"{bucket_name}_obfuscation_table.dat")
    with open(obfuscation_table_fname, "w") as f:
        for subject_id, obfuscated_subject_id in obfuscation_table.items():
            f.write(f"{obfuscated_subject_id}\t{subject_id}\n")
    obfuscation_table_fname.chmod(0o444)
if __name__ == "__main__":
    main()
