from typing import Dict
import io
import tarfile
import lzma
import pathlib
import itertools
import numpy as np
LZMA_PRESET = lzma.PRESET_DEFAULT
def get_real_stem(fname: pathlib.Path) -> str:
    return fname.name.split(".")[0]
def glob_ndarrays(fname: pathlib.Path):
    return itertools.chain(fname.glob("*.npy"), fname.glob("*.npy.xz"))
def read_ndarray(fname: pathlib.Path) -> np.ndarray:
    file_extension = get_file_extension(fname)
    assert file_extension in (".npy", ".npy.xz"), "File needs to have '.npy' | '.npy.xz' extension!"
    if file_extension == ".npy":
        return np.load(fname)
    elif file_extension == ".npy.xz":
        return read_lzma_ndarray(fname)
    else:
        raise SystemError()
def save_ndarray(arr: np.ndarray, output_fname: pathlib.Path, compress: bool) -> None:
    if compress:
        save_lzma_ndarray(arr=arr, output_fname=output_fname)
    else:
        file_extension = get_file_extension(output_fname)
        assert file_extension == ".npy", "File needs to have '.npy' extension!"
        np.save(file=output_fname, arr=arr)
def get_file_extension(fname: pathlib.Path) -> str:
    return "".join(fname.suffixes).lower()
def read_lzma_ndarray(fname: pathlib.Path) -> np.ndarray:
    assert get_file_extension(fname) == ".npy.xz", "File extension has to be 'npy.tar.xz'!"
    with lzma.LZMAFile(fname, "rb") as f:
        return np.load(f)
def save_lzma_ndarray(arr: np.ndarray, output_fname: pathlib.Path, preset = LZMA_PRESET) -> None:
    file_extension = get_file_extension(output_fname)
    assert file_extension in (".npy", ".npy.xz"), "Output file needs to have '.npy' | '.npy.xz' extension!"
    output_fname = output_fname.with_suffix(".npy.xz") if file_extension == ".npy" else output_fname
    with lzma.LZMAFile(output_fname, mode="wb", preset=preset) as f:
        np.save(file=f, arr=arr)
def read_lzma_ndarray_archive(fname: pathlib.Path) -> Dict[str, np.ndarray]:
    assert get_file_extension(fname) == ".npy.tar.xz", "File extension has to be 'npy.tar.xz'!"
    arr_dict = dict()
    with tarfile.open(fname, "r:xz") as tar_f:
        for tar_info in tar_f:
            fobj = tar_f.extractfile(tar_info)
            buf = io.BytesIO(fobj.read())
            arr_dict[tar_info.name] = np.load(buf)
    return arr_dict
def save_lzma_ndarray_archive(output_fname: pathlib.Path, preset=LZMA_PRESET, **arr_dict) -> None:
    file_extension = get_file_extension(output_fname)
    assert file_extension in (".npy", ".npy.tar.xz"), "Output file needs to have '.npy' | '.npy.tar.xz' extension!"
    output_fname = output_fname.with_suffix(".npy.tar.xz") if file_extension == ".npy" else output_fname
    with tarfile.open(output_fname, "w:xz", preset=preset) as tar_f:
        for key, arr in arr_dict.items():
            buf = io.BytesIO()
            np.save(file=buf, arr=arr)
            tar_info = tarfile.TarInfo(name=key)
            tar_info.size = len(buf.getbuffer())
            buf.seek(0)
            tar_f.addfile(tarinfo=tar_info, fileobj=buf)
