import numpy as np
def fast_trilinear_interpolation(img, x, y, z, fill_value=0.):
    """
    Trilinearly interpolates points (x_i, y_i, z_i) on image 'img'
    :param img: An ndarray of shape [m, n, j]
    :param x: An ndarray of shape [k]
    :param y: An ndarray of shape [k]
    :param z: An ndarray of shape [k]
    :param fill_value: Which value do use for out-of-bounds interpolations
    :return: An ndarray of interpolated values of shape [k]
    """
    m, n, j = img.shape
    g = np.zeros(shape=(m-1, n-1, j-1, 8), dtype=img.dtype)     
    g[..., 0] = img[:m-1, :n-1, :j-1]
    g[..., 1] = img[1:m, :n-1, :j-1]
    g[..., 2] = img[:m-1, :n-1, 1:j]
    g[..., 3] = img[1:m, :n-1, 1:j]
    g[..., 4] = img[:m-1, 1:n, 1:j]
    g[..., 5] = img[1:m, 1:n, :j-1]
    g[..., 6] = img[:m-1, 1:n, 1:j]
    g[..., 7] = img[1:m, 1:n, 1:j]
    x_0 = np.floor(x).astype(np.int32)      
    y_0 = np.floor(y).astype(np.int32)      
    z_0 = np.floor(z).astype(np.int32)      
    x_ib = (0 <= x_0) & (x_0 < n - 1)   
    y_ib = (0 <= y_0) & (y_0 < m - 1)   
    z_ib = (0 <= z_0) & (z_0 < j - 1)   
    oob = ~(x_ib & y_ib & z_ib)    
    x_0[oob] = 0
    y_0[oob] = 0
    z_0[oob] = 0
    x = x - x_0.astype(img.dtype)     
    y = y - y_0.astype(img.dtype)     
    z = z - z_0.astype(img.dtype)     
    q = g[y_0, x_0, z_0, :]     
    coeffs = np.stack([
        (1 - x) * (1 - y) * (1 - z),
        (1 - x) * y * (1 - z),
        (1 - x) * (1 - y) * z,
        (1 - x) * y * z,
        x * (1 - y) * (1 - z),
        x * y * (1 - z),
        x * (1 - y) * z,
        x * y * z
    ], axis=1)  
    f = np.sum(np.multiply(coeffs, q), axis=1)  
    f[oob] = fill_value
    return f
