from argparse import ArgumentParser
import pathlib
import operator
import functools
import random
import shutil
import numpy as np
from utils import read_lzma_ndarray, get_real_stem, read_lzma_ndarray_archive
class ExamplePath(pathlib.Path):
    _flavour = pathlib._posix_flavour
    def get_real_stem(self) -> str:
        return self.name.split(".")[0]
    def __eq__(self, other):
        return self.get_real_stem() == other.get_real_stem()
    def __hash__(self):
        return hash(self.get_real_stem())
def main():
    parser = ArgumentParser()
    parser.add_argument("--root_dir",
                        help="The root directory containing the desired sub-directories",
                        required=True)
    parser.add_argument("--dir_names",
                        help="The names of the directories to sample from",
                        nargs="+",
                        required=True)
    parser.add_argument("--output_dir",
                        help="Where the assembled batch should be stored",
                        required=True)
    parser.add_argument("--batch_size",
                        type=int,
                        default=16)
    args = parser.parse_args()
    args.root_dir = ExamplePath(args.root_dir)
    args.dir_paths = [args.root_dir.joinpath(dir_name) for dir_name in args.dir_names]
    assert all(map(operator.methodcaller("exists"), args.dir_paths)), "At least one of the directories does not exist!"
    assert all(map(operator.methodcaller("is_dir"), args.dir_paths)), "Not all `dir_names` refer to directories!"
    args.output_dir = pathlib.Path(args.output_dir)
    args.output_dir.mkdir(exist_ok=True)
    output_dirs = dict()
    for dir_path in args.dir_paths:
        name = dir_path.name
        output_dirs[name] = args.output_dir.joinpath(name)
        output_dirs[name].mkdir(exist_ok=True)
    fnames = {dir_path.name: set(dir_path.glob("*")) for dir_path in args.dir_paths}
    common_example_ids = list(map(operator.methodcaller("get_real_stem"),
                                  functools.reduce(operator.and_, fnames.values())))
    batch_ids = random.sample(common_example_ids, k=args.batch_size)
    for dir_path in args.dir_paths:
        for fname in fnames[dir_path.name]:
            if fname.get_real_stem() in batch_ids:
                name = dir_path.name
                if "NIFTI" in dir_path.name:
                    shutil.copy(src=fname, dst=output_dirs[name])
                elif "NDARRAY" in dir_path.name:
                    if ".npy.xz" in fname.name:
                        arr = read_lzma_ndarray(fname)
                    else:
                        archive = read_lzma_ndarray_archive(fname)
                        arr = archive['convex_hull']
                    np.save(output_dirs[name].joinpath(get_real_stem(fname)), arr=arr)
                print(f"Copied {fname} to {output_dirs[name]}")
if __name__ == "__main__":
    main()
